""" Build 'contour' polygons of specified runoff limits - annual, multi-annual ...

Development and testing was undertaken in explore_runoff_contours.ipynb
(which relied on a previous version of this script for input data, so will
no longer run correctly).

AT October 2020
"""

import pandas as pd
import os
from shapely.geometry import Polygon
import geopandas as gpd
import numpy as np
from scipy.spatial import cKDTree
import rasterio.mask
import argparse

import pdb

from load_env import *


def generate_locs_df(boxes):
    """
    Generate a complete set of box-slice indices (i.e. entire ice sheet, with no
    regard to data availability).

    """

    ix_boxes = []
    ix_slices = []
    for ix, box in boxes.iterrows():
        slices = gpd.read_file(os.path.join(SLICES_SHPF_ROOT, 
                                'boxes_fid{id}.shp'.format(id=box.FID)))
        if box.rev_subs == 1:
            ix_slices.extend(slices.FID[::-1])
        else:
            ix_slices.extend(slices.FID)
        ix_boxes.extend([box.FID]*len(slices))

    ix_complete = pd.MultiIndex.from_arrays(
        (np.array(ix_boxes), np.array(ix_slices)), 
        names=('box_id', 'slice_id'))

    # Generate a 'dummy' dataframe which we can left-join the incomplete dataset to.
    locations = pd.DataFrame({
        'status':0, 
        'i':np.arange(0, len(ix_boxes), 1).astype(int) }, 
        index=ix_complete)

    return locations



def identify_segments(df):
    """
    Accepts dataframe with (box_id, slice_id) index and a continuous integer column i.

    returns: segments. List of tuples. (seg_start_i, seg_end_i, (seg_start_box_slice, seg_end_box_slice))

    """

    # Identify contiguous segments of missing runoff limit retrievals.
    segments = []
    pointer = -1
    first_good = df[df.x.notna()].iloc[0]['i']
    for ix, sl in df[int(first_good):-1].iterrows():
        # In the middle of a sequence, iterate out of it
        if sl['i'] < pointer:
            continue
        
        # Checking the next sequence
        remaining_vals = df.iloc[int(sl['i'])+1:].dropna(subset=('x','y'))

        try:
            next_good = remaining_vals.iloc[0]
        except IndexError:
            #print('IndexError. Probably means there are just NaNs left in dataset, so wrapping to find next good from start of sequence.')
            #print(ix, sl)
            remaining_vals = df.iloc[0:].dropna(subset=('x','y'))
            next_good = remaining_vals.iloc[0]
            segments.append((sl['i'], next_good['i']))
            break
        
        # Start of sequence
        if next_good['i'] - sl['i'] > 1:
            segments.append((sl['i'], next_good['i'], (ix, next_good.name)))
            pointer = next_good['i']

        # Case of contiguous slices
        elif (next_good['i'] - sl['i']) == 1:
            continue

        else:
            print('reached else')
            #print(ix)
            #print(sl)

    return segments



def interpolate_segments(segments, df, min_seg_length=20, max_seg_length=150,
        interpolate_every=10,
        elev_tol_hi=20,
        elev_tol_lo=50,
        dist_bound=30000): #CHECK!!
    """
    Generate x,y values at every xth location in each missing segment, as a
    function of elevation between the segment start and end points.

    min_seg_length : number of missing slices needed before a segment gets filled
    interpolate_every : generate an interpolated value every x slices.

    returns: list, pts, containing tuples:
        (here_boxid, here_sliceid, found_x, found_y, found_e, distances)

    """

    pts = []

    # Iterate through all the segments found above to be missing data
    for seg in segments:
        
        seglength = abs(seg[1] - seg[0])
        if (seglength < min_seg_length) or (seglength > max_seg_length):
            continue    
        
        # Special case of wrapping back around the southern tip, where indexes reset to 0.
        # Continue the indexes sequentially so as to enable the interpolation below, but flag it with wrap=True.
        if seg[1] - seg[0] < 0:
            wrap = True
            sl_arr = np.array((seg[0], seg[1]+len(df)))
        # The usual case
        else:
            wrap = False
            sl_arr = np.array(seg[0:2])
            
        # Get information about the start and end points of the segment
        st = df.iloc[int(seg[0])]
        en = df.iloc[int(seg[1])]
        
        # Linearly interpolate elevation from the start to the end of the segment, as a function of 'kms' (slice)
        elevs = np.array((st.elev, en.elev))
        func = np.polyfit(sl_arr, elevs, 1)
        sl_i = np.arange(sl_arr[0]+interpolate_every, sl_arr[1], interpolate_every)
        elevs_i = func[0] * sl_i + func[1]
        
        # Get a copy of last 'good' X,Y position
        last_x = st.x
        last_y = st.y
        
        # Skip: used to note n. skipped boxes so that search window can be widened progressively
        skip = 1
        
        # Proceed through the positions to be interpolated at
        for sl, elev in zip(sl_i, elevs_i):
            
            # Deal with special wrapping case
            sl_here = sl
            if wrap:
                if sl > len(df):
                    sl_here = sl - len(df)
            
            # Identify the box and slice IDs to be interpolated for
            try:
                here_boxid = df.iloc[int(sl_here)].name[0]
                here_sliceid = df.iloc[int(sl_here)].name[1]
            except IndexError:
                print('Index error at sl %s, continuing' %sl)
                continue
            
            # Read in the slices shapefile correpsonding to the box
            box_slices = gpd.read_file(os.path.join(SLICES_SHPF_ROOT, 
                            'boxes_fid{n}.shp'.format(n=here_boxid)))
            # Get just the slice we are interested in
            this_slice = box_slices.loc[here_sliceid]
            
            # Read the DEM within the slice
            with rasterio.open(DEM_FH, 'r') as ch:
                out_im, aff = rasterio.mask.mask(ch, [this_slice.geometry], crop=True)
            # Mask it to ensure off-Ice Sheet points do not get retrieved accidentally
            with rasterio.open(MSK_FH, 'r') as ch:
                out_im_msk, aff = rasterio.mask.mask(ch, [this_slice.geometry], crop=True)
            out_im = out_im * out_im_msk
            
            # Create coordinates for DEM
            out_im_flat = out_im.ravel()
            x_coords = aff[2] + (aff[0] * np.arange(out_im.shape[2]))
            y_coords = aff[5] + (aff[4] * np.arange(out_im.shape[1]))
            X, Y = np.meshgrid(x_coords, y_coords)
            X = X.ravel()
            Y = Y.ravel()
            
            # Create lookup DF
            comb = pd.DataFrame({'elev':out_im_flat, 'x':X, 'y':Y})
            comb = comb[comb.elev > 0]

            # Retain just elevations within a tolerance
            comb = comb[(comb.elev < elev+elev_tol_hi) & (comb.elev > elev-elev_tol_lo)]
            
            # If for some reason there are no points
            if len(comb) < 1:
                skip += 1
                continue
                
            # Convert the retained DEM points to a lookup tree
            tree = cKDTree(np.array((comb.x, comb.y)).swapaxes(0,1))

            # Query the tree using the last known point
            distances, indices = tree.query((last_x, last_y),
                k=1,
                distance_upper_bound=(dist_bound*(skip*10)))
            
            if not np.isfinite(distances):
                skip += 1
                continue
            else:
                skip = 1
            
            # Save the results
            found_x = comb.iloc[indices-1].x
            found_y = comb.iloc[indices-1].y
            found_e = comb.iloc[indices-1].elev
            
            pts.append((here_boxid, here_sliceid, found_x, found_y, found_e, distances))
            
            # Update the last known point for the next tree search
            last_x = found_x
            last_y = found_y      

    return pts



def insert_interpolated_pts(pts, df):
    """ 
    Put the interpolated points into the provided data frame. Sets their
    row `status` to 1.

    pts : list of interpolated points
    df : dataframe to insert them into

    """

    for pt in pts:
        df.loc[(int(pt[0]), int(pt[1])), 'x'] = pt[2]
        df.loc[(int(pt[0]), int(pt[1])), 'y'] = pt[3]
        df.loc[(int(pt[0]), int(pt[1])), 'elev'] = pt[4]
        df.loc[(int(pt[0]), int(pt[1])), 'status'] = 1

    return df



def smooth_xy(df, win_size, min_periods, win_type):
    """ Smooths x and y columns, returns same df with new columns xs, ys. """

    df.loc[:,'xs'] = df['x'].rolling(win_size, center=True, min_periods=min_periods, 
        win_type=win_type).mean()
    df.loc[:,'ys'] = df['y'].rolling(win_size, center=True, min_periods=min_periods, 
        win_type=win_type).mean()

    return df



def generate_contour_poly(df, contour_name,
        x='x', y='y',
        crs=3413, save_path=None):
    """
    Create a polygon of the runoff limit points - essentially a contour.

    Optionally, save this single contour to a shapefile.

    """

    df_here = df.dropna(subset=(x, y))
    poly = Polygon([(row[x], row[y]) for ix, row in df_here.iterrows()])
    shp = gpd.GeoDataFrame({'id':[contour_name,]}, geometry=[poly], crs=crs)

    # Cautious cleaning
    shp['geometry'] = shp.buffer(0.1)
    
    if save_path is not None:
        shp.to_file(save_path)

    return shp


def restrict_to_regions(gis, poly, field_id='FID'):
    """ This forces the generated polygon to conform to the ice sheet margins in regions of 'impossible' detection and
    in all boxes that are not specified as part of a region."""
    conformed = gis.conform_bad_regions_to_margin()
    mod_poly = poly.append(conformed)
    mod_poly[field_id] = 0
    print(mod_poly)
    mod_poly = mod_poly.dissolve(by=field_id)
    print(mod_poly)
    return mod_poly


def _minmax_reducer(grp, method):
    if method == 'max':
        idx = grp.elev.argmax() 
    elif method == 'min':
        idx = grp.elev.argmin()
    else:
        raise ValueError('Unknown reduction method')
    return grp.iloc[idx]

def _median_reducer(grp):
    med_elev = grp.elev.median()
    grp_sorted = grp.sort_values('elev')
    row = grp_sorted[grp_sorted.elev >= med_elev].iloc[0]
    return row

if __name__ == '__main__':

    parser = argparse.ArgumentParser(description='Runoff limit polygon/contour generation. Generate a single contour.')
    parser.add_argument('year_range_earliest', type=int, 
        help='int, earliest year to use as input to contour generation.')
    parser.add_argument('year_range_latest', type=int, 
        help='int, latest year to use as input to contour generation.')
    parser.add_argument('-agg', dest='agg', default=False, action='store_true',
        help='bool. If False, produce one contour per year. If True, aggregate data to produce a single contour.')
    parser.add_argument('-aggm', dest='aggm', default='median', 
        help='Defines aggregation method to use if multiple values per slice are available when using -agg option. One of median, min, max.')
    parser.add_argument('-restrict', dest='restrict_to_regions', action='store_true',
        help='bool. If True, load_env.gis.regions_boxes and impossible_runoff_zones.shp are used to draw the runoff limit only where valid. Where invalid, the polygon will follow the ice margin.')
    parser.add_argument('-interpolate', dest='interpolate', action='store_true',
        help='bool. If True, use the interpolation routine to fill between missing slices.')
    parser.add_argument('-xyt', dest='xyt_csv', default=None, type=str, 
        help='string. Path of the xyt file to use, otherwise defaults to that specified in load_env.py.')
    args = parser.parse_args()

    """
    Other choices currently remain hard-coded:

    drop : ids of boxes to drop
    dist_std_tol: 0-3
    outlier_dist : in combination with outlier_std. Distances > outlier_dist AND dist. std < outlier_std. are removed.
    outlier_std : see outlier_dist.
    smooth : none or nslices over which to smooth. min_periods, box type etc are hardcoded.

    """


    ## Inputs
    SLICES_SHPF_ROOT = os.path.join(os.environ['L0lib'],'100k_boxes/')


    DEM_FH = os.path.join(os.environ['L0data'], 'GIMPDEM/gimpdem_90m_v01.1_EPSG3413.tif')
    # 'Special' mask - only Greenland Ice Sheet, other areas masked out.
    # gdalwarp -cutline ../gris_only_outline/greenland_icesheet_simplify5000_buffer0.shp GimpIceMask_90m_v1.1_epsg3413.tif GimpIceMask_90m_v1.1_epsg3413_gris.tif
    MSK_FH = os.path.join(os.environ['L0lib'], 'GIMPMASK/GimpIceMask_90m_v1.1_epsg3413_gris.tif')


    ############################################################################
    ## Loading

    gis = gisloader()
    if args.xyt_csv is not None:
        xytpd = pd.read_csv(args.xyt_csv, index_col=0)
        print('Loaded {f}.'.format(f=args.xyt_csv))
        if ('z' in xytpd.columns) and ('elev' not in xytpd.columns):
            xytpd = xytpd.rename({'z':'elev'}, axis='columns')
        xytpd = xytpd[(xytpd.x != 0) & (xytpd.y != 0)]
    else:
        rlims = rlimsloader()
        xytpd = rlims.xytpd

    # Do this once, as if there are multiple years to process it'll get used multiple times.
    locs_df = generate_locs_df(gis.boxes.sort_index())

    ############################################################################
    ## Extract data to be converted to contour.

    raw_store = {}
    if args.agg:
        to_agg = xytpd[(xytpd.year >= args.year_range_earliest) & 
                                 (xytpd.year <= args.year_range_latest)]
        grps = to_agg.groupby(to_agg.loc_id)
        if args.aggm == 'median':
            res = grps.apply(lambda x:_median_reducer(x))
        elif args.aggm == 'min':
            res =  grps.apply(lambda x:_minmax_reducer(x,'min'))
        elif args.aggm == 'max': 
            res = grps.apply(lambda x:_minmax_reducer(x,'max'))
        else:
            raise ValueError('Aggregation method not known.')
        contour_name = '{s}_{e}_{m}'.format(s=args.year_range_earliest, e=args.year_range_latest, m=args.aggm)
        raw_store[contour_name] = res

    else:
        years = np.arange(args.year_range_earliest, args.year_range_latest+1, 1).astype(int)
        for year in years:
            raw_store[year] = xytpd[xytpd.year == year]
    
    # Sort data and proceed to running.
    print('{n} polygons to build.'.format(n=len(raw_store)))
    for item in raw_store:

        print(item) 

        td = raw_store[item]

        # Make a multi-index of box_id, slice_id and apply it to the dataset
        td.index = pd.MultiIndex.from_frame(td.filter(items=('box_id', 'slice_id')))

        # Create a dataframe with all possible box, slices. 
        # Implicitly joins on the multiindexes.
        # Sort on column 'i' to preserve the clockwise order of points.
        td_cont = locs_df.join(td).sort_values('i')


        # Generate statistics for culling outliers
        td_cont['dist_fwd'] = np.sqrt( np.abs(td_cont.x.shift(1) - td_cont.x)**2 + 
                                       np.abs(td_cont.y.shift(1) - td_cont.y)**2 )
        td_cont['dist_bck'] = td_cont['dist_fwd'].shift(1)

        td_cont['dist_fwd_std'] = td_cont.dist_fwd.rolling(25, center=True).std() 

        # Do clean-up
        outliers = td_cont.where((td_cont.dist_fwd > td_cont.dist_fwd_std*2) | 
                           (td_cont.dist_bck > td_cont.dist_fwd_std*2) | 
                           ((td_cont.dist_fwd > 5000) & (td_cont.dist_fwd_std < 3000)))

        i = td_cont['i']
        cleaned = td_cont.where(outliers.isna())
        # Give the i column back after it's been wiped by isna()!
        cleaned['i'] = i 

        # If restricted, remove data from bad regions already
        if args.restrict_to_regions:
            bad_regions = gis.conform_bad_regions_to_margin()
            bad_regions['geometry'] = bad_regions.buffer(1000)
            cleaned = gpd.GeoDataFrame(cleaned, 
                geometry=gpd.points_from_xy(cleaned['x'], cleaned['y']),
                crs=3413)
            #cleaned.loc[:,'within'] = cleaned.within(bad_regions)
            print('Points before restriction:', len(cleaned))
            cleaned = gpd.sjoin(cleaned, bad_regions, how='left', op='within')
            # The left-join means that any overlapping region geometries can cause points to get two memberships - get rid.
            cleaned = cleaned[~cleaned.index.duplicated()]
            # Points that don't fall within a bad region will get a NaN for the region index
            cleaned = cleaned[cleaned.index_right.isna()]
            print('Points after  restriction:',len(cleaned))

        # Identify remaining segments and interpolate them
        # Might need to check this in the future - esp. max_length of segments now that bad points can be cleared above.
        if args.interpolate:
            print('Interpolating')
            segments = identify_segments(cleaned)
            interpolated_pts = interpolate_segments(segments, cleaned)
            complete_ds = insert_interpolated_pts(interpolated_pts, cleaned)
        else:
            complete_ds = cleaned

        # Smooth points to produce nice polygon
        smoothed = smooth_xy(complete_ds, 25, 1, 'triang') #SHOULD BE MIN_POINTS=3, CHANGED TO =1 FOR EMAX

        if args.restrict_to_regions:
            # Add the manual east coast points to allow geometries to dissolve properly
            sub_points = gpd.read_file(os.path.join(os.environ['L0lib'], '100k_boxes', 'boxes_high_points.shp'))
            sub_points['xs'] = sub_points.geometry.x
            sub_points['ys'] = sub_points.geometry.y
            smoothed = pd.concat([smoothed, sub_points], axis=0)

        # Some final cleanup:
        # There are problems with the rapid changes in angle in box 53, especially 
        # given that it doesn't automatically wrap back around to box 1 for smoothing. 
        # It's a small area, and not especially important to the runoff limit, so get rid of it.
        # But, dealt with earlier if restrict_to_regions is True.
        if not args.restrict_to_regions:
            smoothed = smoothed[smoothed.index.get_level_values('box_id') < 53]

        # Draw a polygon
        poly = generate_contour_poly(smoothed, item, x='xs', y='ys')

        # Finally, restrict to only certain regions if requested.
        # This is computationally rubbish, as we've already tried to find 
        # a runoff limit in these areas, but it's only really at this point
        # that we have access to a polygon for vector operations as opposed to 
        # all the x,y points that we're trying to draw the polygon from.
        if args.restrict_to_regions:
            poly = restrict_to_regions(gis, poly)

        # Save polygon as shapefile
        if args.xyt_csv is not None:
            n = os.path.split(args.xyt_csv)[-1][:-4]
            fn = 'poly_{i}_{n}.shp'.format(i=item, n=n)
        else:
            fn = 'poly_{i}.shp'.format(i=item)

        poly.to_file(os.path.join(os.environ['ANALYSIS_OUT'], 'runoff_limit_polys/{f}'.format(f=fn)))

        # Save underlying Point data as CSV (N.B. NOT RESTRICTED!!)
        fn = fn[:-3] + 'csv'
        complete_ds.to_csv(os.path.join(os.environ['ANALYSIS_OUT'], 'runoff_limit_polys/{f}'.format(f=fn)))

