# ---
# jupyter:
#   jupytext:
#     cell_metadata_filter: all
#     formats: ipynb,py:percent
#     notebook_metadata_filter: all,-language_info
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.7.1
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# %% [markdown]
# # MacFerrin excess melting calculations

# %% trusted=true
from load_env import *
gis = gisloader()


# %% trusted=true
# #%run ~/scripts/paper_rlim_detection_repo/calculate_mar_stats_by_box.py

# %% trusted=true
cluster = SGECluster()
cluster.scale(jobs=4)
client = Client(cluster)

# %% trusted=true
mar_partial = os.path.join(gis.MAR_DS_ROOT, '*.nc')

melt = marutils.open_dataset(mar_partial, transform_func=lambda x: x.ME)
rainwater = marutils.open_dataset(mar_partial, transform_func=lambda x: x.RF)
accumulation = marutils.open_dataset(mar_partial, transform_func=lambda x: x.SF) # all mmwe/day

"""
mmwe is equivalent to kg (m-2):

    mm-->m : / 1000
    kgm3 = m.w.e. * 1000 (density of water, kgm3)
    
    e.g 500 mm w.e. --> 0.5 m w.e. --> * 1000 --> 500 kg m-2.
"""

def _res(xa):
    """ Resampling logic """
    return xa.resample(time='1A-AUG').sum(dim='time') 
melt = _res(melt)
rainwater = _res(rainwater)
accumulation = _res(accumulation)


t_air = marutils.open_dataset(mar_partial, transform_func=lambda x: x.TT)
t_avg = t_air.resample(time='1A-AUG').mean(dim='time')

# %% [markdown]
# See lines ~400 onwards of https://github.com/mmacferrin/Greenland_Ice_Slabs/blob/master/src/RCM_Manager.py

# %% trusted=true
pi = .873 # Refrozen ice density, from Machguth 2016
# Snow density parameterization from Langen, et al. (2017)
a_rho = 328.35
b_rho = -0.049376
c_rho = 1.0427
d_rho = -0.11186
ps = (a_rho + b_rho*gis.mar_ds.SH + c_rho*gis.mar_ds.LAT + d_rho*gis.mar_ds.LON) / 1000.0

Z = (pi - ps)/ps
c = 2.108 # Heat capacity of ice, kJ/(kg *C)
L = 333.55 # Latent heat of fusion of ice, kJ/kg
# The t_annual value is being used to calculate firn/snow temperatures.
# For this calculation, null-out any average temperatures above zero (should be none for the ice sheet).
t_avg = t_avg.where(t_avg < 0)



# %% trusted=true
# %matplotlib widget
Z.plot()

# %% trusted=true
# %matplotlib widget
t_avg.sel(time='2000').plot()


# %% trusted=true
Q = (c/L) * -1.0*t_avg # Expressed in *positive* degrees below zero.

MC_threshold = (Z + Q) * (1 + Z)**-1

# To be run with annual values

M_e = ( 
        ((melt+rainwater)/accumulation) - MC_threshold
      ) * accumulation

# %% trusted=true
import xarray as xr
M_e_full = xr.Dataset({'M_e':M_e}) #, 'M_e_10y':M_e_avg})

M_e_full.to_netcdf('/flash/tedstona/analysis/MAR_10km_excess_melt_mbyear.nc')
