# ---
# jupyter:
#   jupytext:
#     cell_metadata_filter: all
#     formats: ipynb,py:percent
#     notebook_metadata_filter: all,-language_info
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.7.1
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# %% [markdown]
# # RACMO - MacFerrin excess melting calculations

# %% trusted=true
from load_env import *
gis = gisloader()


# %% trusted=true
cluster = SGECluster(queue='huge.q@node56', interface='ib0', cores=1, memory='16GB', processes=2, job_extra=['-pe smp 1'], python='ionice -c 3 /home/tedstona/miniconda3/envs/geospa/bin/python')

cluster.scale(jobs=2)
client = Client(cluster)

# %% trusted=true
cluster.close()

# %% trusted=true
from glob import glob 
def read_netcdfs(files, dim, transform_func=None, decode_times=False):
    def process_one_path(path):
        # use a context manager, to ensure the file gets closed after use
        with xr.open_dataset(path, decode_times=decode_times) as ds:
            # transform_func should do some sort of selection or
            # aggregation
            if transform_func is not None:
                ds = transform_func(ds)
            # load all data from the transformed dataset, to ensure we can
            # use it after closing each original file
            if ('lat' or 'LAT') in ds.coords:
                ds = ds.rename({'lat':'y', 'lon':'x'})
            ds.load()
            
            if "1990" in path:
                if len(ds['time']) > 12:
                    print('WARNING: clipping time bounds of 1990 file')
                    ds = ds.isel(time=slice(0,12))
            
            return ds

    paths = sorted(glob(files))
    datasets = [process_one_path(p) for p in paths]
    combined = xr.concat(datasets, dim)
    new_ts = pd.date_range('1980-01-01', '2020-12-31', freq='M')
    combined['time'] = new_ts
    return combined



"""
mmwe is equivalent to kg (m-2):

    mm-->m : / 1000
    kgm3 = m.w.e. * 1000 (density of water, kgm3)
    
    e.g 500 mm w.e. --> 0.5 m w.e. --> * 1000 --> 500 kg m-2.
"""




# %% trusted=true
melt = read_netcdfs(os.path.join(gis.RACMO_DS_ROOT, '1km', 'monthly', 'snowmelt', '*.nc'), 
                    dim='time',
                    transform_func=lambda ds: ds.snowmeltcorr)
snowfall = read_netcdfs(os.path.join(gis.RACMO_DS_ROOT, '1km', 'monthly', 'snowfall', '*.nc'), 
                    dim='time',
                    transform_func=lambda ds: ds.snowfallcorr)
rainwater = read_netcdfs(os.path.join(gis.RACMO_DS_ROOT, '1km', 'monthly', 'precip', '*.nc'), 
                    dim='time',
                    transform_func=lambda ds: ds.precipcorr)
t2m = read_netcdfs(os.path.join(gis.RACMO_DS_ROOT, '1km', 'monthly', 't2m', '*.nc'), 
                    dim='time',
                    transform_func=lambda ds: ds.t2mcorr)


# %% trusted=true
def _res(xa):
    """ Resampling logic """
    return xa.resample(time='1A-AUG').sum(dim='time') 

melt_a = _res(melt)
rain_a = _res(rainwater)
accum_a = _res(snowfall)
t_avg = t2m.resample(time='1A-AUG').mean(dim='time')

# %% [markdown]
# See lines ~400 onwards of https://github.com/mmacferrin/Greenland_Ice_Slabs/blob/master/src/RCM_Manager.py

# %% trusted=true
t_avg = t2m.resample(time='1A-AUG').mean(dim='time')

# %% trusted=true
# %matplotlib widget
t2m.sel(time='2000-06').plot()

# %% trusted=true
pi = .873 # Refrozen ice density, from Machguth 2016
# Snow density parameterization from Langen, et al. (2017)
a_rho = 328.35
b_rho = -0.049376
c_rho = 1.0427
d_rho = -0.11186
ps = (a_rho + b_rho*gis.racmo_ds.Topography + c_rho*gis.racmo_ds.LAT + d_rho*gis.racmo_ds.LON) / 1000.0

Z = (pi - ps)/ps
c = 2.108 # Heat capacity of ice, kJ/(kg *C)
L = 333.55 # Latent heat of fusion of ice, kJ/kg
# The t_annual value is being used to calculate firn/snow temperatures.
# For this calculation, null-out any average temperatures above zero (should be none for the ice sheet).
t_avg_clean = t_avg - 273.15
t_avg_clean = t_avg_clean.where(t_avg_clean < 0)



# %% trusted=true
# %matplotlib widget
t_avg.mean(dim=('x', 'y')).plot()

# %% trusted=true
# %matplotlib widget
Z.plot(vmin=0, vmax=5)

# %% trusted=true
t_avg.sel(time='2000')

# %% trusted=true
# %matplotlib widget
t_avg_clean.sel(time='2000').plot(vmin=-50)


# %% trusted=true
Q = (c/L) * -1.0*t_avg_clean # Expressed in *positive* degrees below zero.

MC_threshold = (Z + Q) * (1 + Z)**-1

# To be run with annual values

M_e = ( 
        ((melt_a+rain_a)/accum_a) - MC_threshold
      ) * accum_a

# %% trusted=true
import xarray as xr
M_e_full = xr.Dataset({'M_e':M_e}) #, 'M_e_10y':M_e_avg})

M_e_full.to_netcdf('/flash/tedstona/analysis/RACMO_1km_excess_melt_mbyear.nc')

# %% trusted=true
M_e_full

# %% trusted=true
# %matplotlib widget
M_e_full.M_e.plot(col='time', col_wrap=5)

# %% trusted=true
M_e_full.M_e.sel(time='2019').where(gis.regions_as_racmo_masks.sel(region='SW')).plot()

# %%
