# ---
# jupyter:
#   jupytext:
#     cell_metadata_filter: all
#     formats: ipynb,py:percent
#     notebook_metadata_filter: all,-language_info
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.7.1
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Check spatial coverage by elevation, annually

# %% [markdown]
# Data lineage and pre-requisites.
#
# ### 1. GIMPDEM corresponding to 1km reduced grid.
#
# Grid info: see section 2, below.
#
#     $ gdalwarp -tr 1000 1000 gimp_90m.dataset gimp_1km.tif
#
#
# ### 2. Annual maps of count coverage
#
#     $ python river_detection/create_master_maps.py merge_sw.template -yearly <year> -cct <maxm. cloud threshold, likely 50>
#
# This produces 30 m resolution tiles of annual cloud-free observations of dimensions according to those set in the quadrant merging template. The tiles are linked together into annual quadrant maps as VRTs. All tiffs and VRTs are in OUTPUT_DIR/master_maps/merge_<quad>/.
#
# The per-quadrant VRTs are linked together to form GrIS-wide VRT files, which are stored in ANALYSIS_OUT/pixel_counts/.
#
# Then, these VRTs are aggregated/transformed to 1 km resolution and matching GIMP DEM extent.
#
# ```bash
# gdalwarp -tr 1000 1000 -te <take from GIMP> vrt_in.vrt master_<year>_count_matchgimp_1km.tif
# ```
#
# This produces ANALYSIS_OUT/pixel_counts/master_<year>_count_matchgimp_1km.tif 
#
# Then, these files are used by this script Pt. 1, which generates spatial_count_1km_results_{y}.csv.
#
#     
# N.b. see also Dropbox -> how-tos -> modify_vrt_links.md (I updated the paths in the GrIS-wide VRT count files to be absolute).
#     
#     
# ### Old approaches
#     
# Noting for reference...Deprecated scripts, which can still be found in repository history but have been deleted from the HEAD, consisted of:
#     
# * spatial_elev_coverage.py
# * stats_annual_counts_by_elev.py
#
#

# %% trusted=true
from load_env import *
gis = gisloader()
rlims = rlimsloader()

### Step 1: Compute the annual counts by elevation

# %% trusted=true
# Set to True to (re)generate the counts - this takes a long time
compute_counts = False

# %% trusted=true
if compute_counts:
    cluster = SGECluster()
    cluster.scale(jobs=4)
    client = Client(cluster)

# %% trusted=true
if compute_counts:
    import rasterio as rio

    chunks = {'x':1248, 'y':1350}

    def load_year(year):
        asxr = rioxarray.open_rasterio(
                os.path.join(os.environ['ANALYSIS_OUT'],
                             'pixel_counts/master_{y}_count_matchgimp_1km.tif'.format(y=year)), 
                chunks=chunks)
        return asxr

    def load_box(xarr, bounds, geom):
        asxr = xarr.sel(x=slice(bounds[0], bounds[2], 10), y=slice(bounds[3], bounds[1], 10))
        asxr = xarr.rio.clip([geom], drop=False)
        return asxr

    def compute_count_for_bin(data, dem, elev_bin, year, fid):
        data['x'] = dem['x']
        data['y'] = dem['y']
        cc = data.where((dem >= elev_bin) & (dem < elev_bin+100)).mean().load()
        try:
            v = cc.values
        except IndexError:
            v = np.nan
        return {'year':year, 'elev':elev_bin, 'count':v, 'fid':fid}

    def merge_year_results(year, incoming):
        merged = pd.DataFrame(incoming)
        merged.to_csv(
            os.path.join(os.environ['ANALYSIS_OUT'],
                         'pixel_counts/spatial_count_1km_results_{y}.csv'.format(y=year))
        )
        return merged

    def concat_all(incoming):
        return pd.concat(incoming, axis=0)

    dem = rioxarray.open_rasterio(
                os.path.join(os.environ['ANALYSIS_OUT'],
                             'pixel_counts/gimpdem_90m_v01.1_EPSG3413_1km.tif'),
                chunks=chunks)


    L = {}
    for year in range(1989, 2021): 

        print(year)
        data_year = load_year(year)

        LL = []
        for ix, row in gis.boxes.iterrows():
            geometry = row.geometry
            bounds = geometry.bounds

            window = rio.windows.from_bounds(*bounds, dem.rio.transform())
            dst_dim = (int(np.ceil(window.width)), int(np.ceil(window.height)))

            data_box = load_box(data_year, bounds, geometry)
            dem_box = load_box(dem, bounds, geometry)

            for eb in np.arange(800, 2400, 100):
                counts = compute_count_for_bin(data_box, dem_box, eb, year, row.FID)
                LL.append(counts)

        counts_for_box = merge_year_results(year, LL)

        L[year] = counts_for_box

        data_year.close()

# %% [markdown]
# ### Step 2: Plotting

# %% trusted=true
# Load all the points in to memory
store = {}
for year in range(1985,2020):
    data = pd.read_csv(os.path.join(os.environ['ANALYSIS_OUT'],
                                    'pixel_counts/spatial_count_1km_results_{y}.csv'.format(y=year)), 
                       index_col=0)
    store[year] = data
all_counts = pd.concat(store, axis=0)

# %% trusted=true
fig = plt.figure(figsize=(7, 5.5))
cax = plt.subplot(4,2,8)

letgen = letters(start=0)

n = 1 
rlim_used = {}
for region in gis.regions_boxes:
    
    ax = plt.subplot(4,2,n)
    
    # Generate region-wide spatial counts
    bmin, bmax = gis.regions_boxes[region]
    just_region = all_counts[(all_counts.fid >= bmin) & (all_counts.fid <= bmax)]
    res = just_region.groupby(by=['year', 'elev']).mean()
    region_map = res['count'].unstack().T
    
    # Plot them
    if n == 1:
        cb = True
    else:
        cb = False
    sns.heatmap(region_map, vmin=0, vmax=10, cmap="Blues", ax=ax, 
                xticklabels=10, yticklabels=4, 
                linecolor='white',linewidths=0.3,
                cbar=cb, cbar_ax=cax, cbar_kws={'label':'N', 'extend':'max', 'extendfrac':0.2})
    # Turn around the tick labels
    ax.tick_params(axis='y', rotation=0)
    
    # Plot the runoff limit
    rlim = rlims.rlims_plot.filter(items=list(range(bmin, bmax+1)),axis=0).mean()
    rlim.name = 'elev'
    rlim = rlim.to_frame()
    # The heatmap above uses simple 0...n coordinates for x and y, so need to convert
    # runoff limits to these coordinates.
    rlim['year_int'] = np.arange(0, len(rlim.index))
    # As the heatmap is only binned, convert our coordinates to bins too
    rlim['bin'] = pd.cut(rlim.elev, pd.interval_range(start=800,end=2300,freq=100))
    rlim['bin_int'] = rlim['bin'].apply(lambda x: ((x/100)-8))
    # The +0.5 centres the value vertically in the bins
    rlim['bin_int'] = rlim.bin_int.astype('category').cat.codes + 0.5
    rlim = rlim.dropna(subset=['bin'])
    plt.plot(rlim.year_int+0.5, rlim.bin_int, 'orange', marker='+', linestyle='none') 

    # Reverse y-axis so that elevation goes up the axis
    ax.set_ylim(ax.get_ylim()[::-1])
    
    ax.tick_params(axis='both', length=2)
    
    plt.xlabel('')
    plt.ylabel('')
    
    ax.annotate(next(letgen), fontweight='bold', xy=(0,1.01), xycoords='axes fraction',
           horizontalalignment='left', verticalalignment='bottom')
    
    ax.annotate(region, xy=(1,1.01), xycoords='axes fraction',
           horizontalalignment='right', verticalalignment='bottom', 
               color=regions_colors[region])   
    
    rlim_used[region] = rlim
    
    n += 1

plt.subplots_adjust(hspace=0.6)

# Adjust colourbar axis
pos = cax.get_position()
pos.x0 = pos.x0
pos.x1 = pos.x1 - 0.33
cax.set_position(pos)

fig.text(0.03, 0.5, r'$\grave\upsilon$ (m asl)', rotation=90, verticalalignment='center', fontsize=9)

plt.savefig(os.path.join(os.environ['FIGS_OUT'],'suppl/suppl_fig_spatial_cover.pdf'),)

# %%
