# ---
# jupyter:
#   jupytext:
#     cell_metadata_filter: all
#     formats: ipynb,py:percent
#     notebook_metadata_filter: all,-language_info
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.7.1
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Hypsometry

# %% [markdown]
# ## Proceed by raster approach

# %% trusted=true
from load_env import *
gis = gisloader()
rlims = rlimsloader()

# %%
# %matplotlib widget
gis.racmo_ds.Promicemask.plot()

# %%
# %matplotlib widget
gis.racmo_ds.Topography.where(gis.regions_as_racmo_masks.sel(region='S')).where(gis.racmo_ds.Promicemask == 3).plot()

# %% trusted=true
from scipy.optimize import curve_fit
from scipy import integrate
import matplotlib.pyplot as plt


# %% trusted=true
def get_region_dem(region):
     return gis.racmo_ds.Topography.where(gis.regions_as_racmo_masks.sel(region=region)).where(gis.racmo_ds.Promicemask == 3)

def get_region_area(region, elow=500, ehigh=2200):
    dem = get_region_dem(region)
    total = dem.where(dem >= elow).where(dem <= ehigh).count().values
    return total

def region_hypsometry(region):
    region_dem = get_region_dem(region)
    e_bins = np.arange(100, 2300, 100)
    dem_binned = region_dem.groupby_bins(region_dem, bins=e_bins).count().to_pandas()
    dem_binned.index = dem_binned.index.left + 50
    return dem_binned
    
def region_rlims(region):
    bmin, bmax = gis.regions_boxes[region]
    boxes_in_region = np.arange(bmin, bmax+1)
    region_all = rlims.rlims.filter(items=boxes_in_region, axis=0)
    region_annual = region_all.mean()
    
    early_lim = region_annual.loc[1985:1992].mean()
    late_lim = region_annual.loc[2013:2020].mean()
    diff = late_lim - early_lim
    
    return early_lim, late_lim, diff

def _f(x,a,b):
    y = a * x**b
    return y

def hypsometry_to_curve(binned):
    popt, pcov = curve_fit(_f, binned.index, binned)
    return popt

# Integrating over data already binned into 100 m bands is completely wrong, do not do this!
def elev_to_area(popt, e1, e2):
    area,_ = integrate.quad(_f, e1, e2, args=tuple(popt))
    return area

def plot_region_hyp(region, binned, popt, early, late):
    plt.figure()
    binned.plot(color='k', marker='o')
    xx = np.arange(500, 2210, 10)
    yy = _f(xx, popt[0], popt[1])
    plt.plot(xx, yy)
    plt.axvline(early, color='b')
    plt.axvline(late, color='r')
    plt.xlim(0, 2500)
    
    plt.title(region)



# %%
# %matplotlib widget
store = []
bins_store = {}
for region in gis.regions_boxes:
    rlim_early, rlim_late, rlim_diff = region_rlims(region)
    
    dem_bins = region_hypsometry(region)
    dem_bins = dem_bins.loc[500:2200]
    popt = hypsometry_to_curve(dem_bins)
    
    a_8592_1320 = elev_to_area(popt, rlim_early, rlim_late)
    a_1320_plus = elev_to_area(popt, rlim_late, rlim_late+rlim_diff)
    a_8592_pre = elev_to_area(popt, rlim_early-rlim_diff, rlim_early)
    
    store.append(dict(r=region, a8520=a_8592_1320, a1320p=a_1320_plus, a8592b=a_8592_pre, p1=popt[0], p2=popt[1]))
    plot_region_hyp(region, dem_bins, popt, rlim_early, rlim_late)
    
    bins_store[region] = dem_bins
    

# %% trusted=true
hyp = pd.DataFrame(store)
hyp.index = hyp.r

# %% [markdown]
# ### Areas
#
# * a8520: the absolute new runoff area between 85/92 and 13/20
# * a1320p: the additional area that would contribute to runoff for the same rise again in the runoff elevation
# * a8592b: the area bounded by (85/92 minus the elevation change 85-20), (85-92 limit) 
# * p1, p2: power law parameters
#

# %%
hyp

# %% [markdown]
# ### % increase in area if runoff elevation were to rise by the same amount again as 85-92 -> 13-20

# %%
100 / hyp.a8520 * hyp.a1320p

# %%
100 / hyp.a8592b * hyp.a8520

# %%
# %matplotlib widget
plt.figure(figsize=(5,3.5))
incr = 1. / 7
start = 0
for region in gis.regions_boxes:
    cs = bins_store[region].cumsum()
    kws = dict(label=region, color=regions_colors[region], marker='o')
    (100 / cs.iloc[-1] * cs).plot(**kws)
    #(100 / cs.iloc[-1] * bins_store[region]).plot(**kws)
    #lo, hi, _ = region_rlims(region)
    #plt.axvspan(lo, hi, color=regions_colors[region], alpha=0.5, ymin=start, ymax=start+incr)
    #start += incr
plt.legend()
plt.xlabel('Elevation (m)')
plt.ylabel('% of total area')
sns.despine()

# %%
region_rlims('NE')

# %%
for region in gis.regions_boxes:
    rlim_early, rlim_late, rlim_diff = region_rlims(region)
    tot_area = bins_store[region].cumsum().iloc[-1]
    at = elev_to_area((hyp.loc[region].p1, hyp.loc[region].p2), 500, rlim_late)
    ab = elev_to_area((hyp.loc[region].p1, hyp.loc[region].p2), 500, rlim_early)
    perc_e = (100 / bins_store[region].index[-1] * rlim_early)
    perc_a = (100 / tot_area * ab)
    perc_b = (100 / tot_area * at)
    #print(region, np.round(perc_e), np.round(perc_a), np.round(perc_b))
    
    a1 = get_region_area(region)
    a2 = get_region_area(region, elow=rlim_early, ehigh=rlim_late)
    a3 = get_region_area(region, elow=rlim_late, ehigh=rlim_late+rlim_diff)
    a4 = get_region_area(region, elow=rlim_early-rlim_diff, ehigh=rlim_early)
    print(region, rlim_diff, np.round(perc_e), np.round(100 / a1 * a4), np.round(100 / a1 * a2), np.round(100 / a1 * a3))

# %%
a1

# %%
a3

# %%
rlim_diff

# %%
get_region_area('NE', elow=rlim_late, ehigh=rlim_late+(3*rlim_diff))

# %%
bins_store[region].index[-1]

# %%
at

# %%
tot_area

# %%
rlim_late

# %%
bins_store[region].cumsum()

# %%
region

# %%
elev_to_area((hyp.loc['NE'].p1, hyp.loc['NE'].p2), 500, 2150)

# %% trusted=true
