"""

Generate annual maximum runoff limit positions at 1 km posting, and summary
statistics for 100 km boxes.
* X-Y-Time frame of 1 km runoff limit locations.
* Summary CSV of per-box median annual runoff limits.

Originally written May/June 2020.
1st major refactoring autumn 2020.
2nd major refactoring July/August 2021 (converted to use clustering approach).
    - Renamed from investigate_inter_annual.py to find_annual_rlims.py.
    - Initial development of this approach is in investigate_clustering_for_uncertainty.ipynb.

"""

import pandas as pd
import os
from math import trunc
import geopandas as gpd
from multiprocessing import Pool
from tqdm import tqdm
from sklearn.cluster import KMeans
from sklearn.preprocessing import StandardScaler
import math
import argparse

from sqlalchemy import create_engine
import osmnx as ox

import warnings
warnings.filterwarnings('ignore')

import pdb

db_href = "postgresql://tedstona@localhost/rlims"
boxes_shpf = '/home/geoscience/nobackup_cassandra/L0data/100k_boxes/boxes.shp'


def matches_in_slice(sl_here, spatial_index, df):
    # To speed up compute times, 'internally' split the slice into cut bins
    # Find the runoff points within this slice
    # https://github.com/gboeing/urban-data-science/blob/master/19-Spatial-Analysis-and-Cartography/rtree-spatial-indexing.ipynb
    geometry_cut = ox.utils_geo._quadrat_cut_geometry(sl_here.geometry, 20000)
    matches = []
    for poly in geometry_cut:
        
        possible_matches_index = list(spatial_index.intersection(poly.bounds))
        possible_matches = df.iloc[possible_matches_index]
        precise_matches = possible_matches[possible_matches.within(poly)]
        matches.append(precise_matches)

    # Gather all the cut bins composing the slice together
    matches = pd.concat(matches, axis=0)

    return matches


def do_clustering(df):
    # Calculate the number of clusters to be used each year.
    # Rationale = allow one cluster per 7 days / 1 week.
    # On a per-box basis !!!
    product_dates = df.groupby('product_id').date.first()
    p_d = pd.Series(product_dates.index, index=pd.DatetimeIndex(product_dates.values))
    counts = p_d.resample('1W').count()
    n_clusters_per_year = (counts > 0).groupby(counts.index.year).sum()

    # At least this many runoff limit observations must be available in the box in a given year,
    # otherwise that year is skipped.
    min_features_in_box_per_year = 100
    # Minimum number of points required to keep a cluster.
    min_pts_in_cluster = 30 # 100 km = 10000 m, / 30 = 333 obs per product if product complete. 

    cluster_assignments = []

    for year in df.year.unique():
        matches_year = df[df.year == year]
        
        # Extract features destined for the clustering algorithm
        features = matches_year.filter(items=['date','elev'], axis='columns')
        features.loc[:,'date'] = pd.to_numeric(pd.DatetimeIndex(features.date))
        # Scale those features
        scaler = StandardScaler()
        scaled_features = scaler.fit_transform(features.to_numpy())
        
        if len(scaled_features) < min_features_in_box_per_year:
            continue
        
        n_clusters = n_clusters_per_year.loc[year]
        kmeans = KMeans(n_clusters=int(n_clusters))
        kmeans.fit(scaled_features)

        # Apply labels back to dataframe so that we can do some calculations
        matches_year.loc[:,'label'] = kmeans.labels_
        
        label_elev = matches_year.groupby('label').elev
        meds = label_elev.median()
        counts = label_elev.count()
        
        meds_above_thresh = meds[counts > min_pts_in_cluster]
        if len(meds_above_thresh) == 0:
            continue
        
        # Can do everything with this once it has been merged on to the master box dataframe.
        cluster_assignments.append(matches_year.filter(items=['label'], axis='columns'))
    if len(cluster_assignments) > 0:
        ca = pd.concat(cluster_assignments)
        return ca
    else:
        return False


def generate_clusters_info(df_labelled):
    grpd = df_labelled.groupby(['year', 'label'])

    cluster_medians = grpd.elev.median()
    cluster_medians.name = 'cluster_median'

    cluster_mads = grpd.elev.mad()
    cluster_mads.name = 'cluster_mad'

    min_date = grpd.date.min()
    max_date = grpd.date.max()
    duration = pd.Series(pd.TimedeltaIndex(max_date - min_date).days, index=max_date.index)
    duration.name = 'cluster_duration_days'

    n_products = grpd.product_id.nunique()
    n_products.name = 'cluster_n_products'

    n_pts = grpd.elev.count()
    n_pts.name = 'cluster_n_pts'

    clusters_info = pd.concat([cluster_medians, cluster_mads, duration, n_products, n_pts], axis=1)

    return clusters_info 


def query_box(fid, engine):
    # Note here: I am removing September data asare they  too unreliable with clouds + low illumination.
    with engine.connect() as conn:
        df = gpd.read_postgis("""
            SELECT x, y, date_part('year',date) as year, elev, date, geometry, jid, PRODUCT_ID 
            FROM rlims
            JOIN boxes ON ST_Contains(boxes.geom, rlims.geometry) 
            WHERE boxes.fid={n} AND date_part('month', date) < 9
            AND date_part('year', date) = {y}
            """.format(n=fid, y=YEAR), conn, geom_col='geometry')
    return df


def control_extraction(fid):

    def _cluster_reducer(df):
        # Find the median of each cluster of retrievals in this slice, then sort by elevation.
        slice_cluster_medians = df.groupby('label').elev.median().sort_values(ascending=False)
        # Use the highest cluster.
        cluster_ix = slice_cluster_medians.index[0]
        # Isolate the retrievals of the picked cluster.
        preferred_cluster = df[df.label == cluster_ix]
        preferred_cluster = preferred_cluster.sort_values(by='elev')
        # Find the integer-location of the cluster's median value.
        preferred_med = preferred_cluster.iloc[int(len(preferred_cluster)/2)].elev
        # Get all the details of the median value.
        m1 = preferred_cluster[preferred_cluster.elev >= preferred_med].iloc[0] 
        return m1

    # Try to guard against multiprocessing race conditions
    engine.dispose()

    # Read in the 1 km slices for this box
    slices = gpd.read_file('/home/geoscience/nobackup_cassandra/L0data/100k_boxes/boxes_fid{:d}.shp'.format(fid))

    # Get all the points for the box
    df = query_box(fid, engine)

    # Execute box-level clustering.
    cluster_assignments = do_clustering(df)
    if type(cluster_assignments) is not gpd.GeoDataFrame:
        print('No cluster assignments for {fid}'.format(fid=fid))
        return False

    df_labelled = pd.merge(df, cluster_assignments, 
        left_index=True, right_index=True, how='inner')
    # Generate the RTree index for this box.
    spatial_index = df_labelled.sindex

    clusters_info = generate_clusters_info(df_labelled)

    #v = {}
    xyt_store = []
    # Look at each 1 km slice in turn. 
    for s_ix, sl in slices.iterrows():

        # Find retrievals matching slice, from the clusters-labelled dataframe.
        matches = matches_in_slice(sl, spatial_index, df_labelled)
        
        # Reduce data base on clusters info.
        m1 = matches.groupby('year').apply(_cluster_reducer)
        if len(m1) < 1:
            print('\t No result: FID {fid} S {s}'.format(fid=fid, s=s_ix))
            continue

        # Convert index to x,y,z
        #elevs = m1.filter(items=['elev', 'year'])
        #elevs.index = elevs.year
        #elevs = elevs.drop(labels=['year'], axis=1)
        xyt = m1.filter(items=['date', 'x', 'y', 'elev', 'label'])
        xyt.loc[:,'box_id'] = fid
        xyt.loc[:,'slice_id'] = s_ix
        xyt.loc[:,'loc_id'] = float('{b}.{s}'.format(b=fid,s=str(s_ix).zfill(2))) # Note hard-coded z fill!!

        # Save annual results for this slice
        #v[s_ix] = elevs
        xyt_store.append(xyt)

    xyt = pd.concat(xyt_store, axis=0)
    xyt = xyt.reset_index()

    # Label each retained point with median and MAD of parent cluster.
    xyt_with_clusters = pd.merge(
        xyt, 
        clusters_info.filter(items=['cluster_median', 'cluster_mad'], axis='columns'), 
        how='left', left_on=['year','label'], right_index=True)

    # Save clusters info to a file already.
    clusters_info.to_csv(os.path.join(os.environ['ANALYSIS_OUT'], 'rlim_annual_maxm', 
        'clusters_info_FID{n}_{y}.csv'.format(n=fid, y=YEAR)))
    
    slices = None

    print('Box FID{n} complete.'.format(n=fid))

    #return (fid, this_box, this_box.median(axis=0), xyt_store) 
    r = (fid, xyt_with_clusters, clusters_info)
    #results.append(r)
    return r



## -----------------------------------------------------------------------------
## Pool Execution

if __name__ == '__main__':

    YEAR = 2021
    print('ATTENTION: processing one year only: %s' %YEAR)

    debug = False

    # Open database engine (from which connections are spawned by each process)
    engine = create_engine(db_href)

    # Open boxes
    boxes_gpd = gpd.read_file(boxes_shpf)

    results = []

    # Do extraction box-by-box
    try:
        nworkers = int(os.environ['EOLEARN_WORKERS'])
    except KeyError:
        print('WARNING: EOLEARN_WORKERS env-var not set, defaulting to one worker.')
        nworkers = 1

    if nworkers > 1:
        pool = Pool(processes=nworkers,
                  maxtasksperchild=1)
        for res in tqdm(pool.imap_unordered(control_extraction, boxes_gpd.FID), total=len(boxes_gpd)):
            results.append(res)
        pool.close()
    else:
        for fid,row in boxes_gpd.iterrows():
            results.append(control_extraction(row.FID))
            if debug:
                break

    
    # Re-organise results returned by the Pool
    xyt_all = []
    
    # Box-by-box : ...
    for result in results:

        if result == False:
            continue

        # Slice-level runoff limit retrievals for the box
        xyt = result[1]
        # Clusters info for the box (not currently needed here)
        # cinfo = result[2]

        # By this point year has appeared in the index - get rid of it.
        xyt = xyt.reset_index()
    
        # The boxes have had a flag set manually (in QGIS) that denotes whether
        # their constituent 1 km slices need reversing.
        # This is to ensure correct Point ordering of each contour Polygon.
        reverse = boxes_gpd[boxes_gpd.FID == result[0]].rev_subs
        if reverse.iloc[0] == 1:
            print('reversing slices in FID{n}'.format(n=result[0]))
            xyt = xyt[::-1]

        xyt_all.append(xyt)

    ## Main output 1: X-Y-Time frame of 1 km runoff limit locations.
    # At the moment this contains all years of data
    xytpd = pd.concat(xyt_all, axis=0)
    xytpd.to_csv(os.path.join(os.environ['ANALYSIS_OUT'], 'rlim_annual_maxm', 'xytpd_%s.csv' %YEAR))  

    ## Main output 2: Median annual runoff limit for each 100 km box, through time.
    medspd = xytpd.groupby(['box_id', 'year']).elev.median().unstack()
    medspd = medspd.reindex(sorted(medspd.columns),axis=1)
    # Sort rows so that north is at top of dataframe (and hence heatmap)
    medspd = medspd.reindex(sorted(medspd.index, reverse=True), axis=0)
    medspd.to_csv(os.path.join(os.environ['ANALYSIS_OUT'], 'rlim_annual_maxm', 
        '100km_box_medians_%s.csv' %YEAR))

    #madspd = xyt_all.groupby(['box_id', 'year']).elev.mad()
    #box_meds_mads = pd.merge(medspd, madspd, left_index=True, right_index=True)







