"""
Set up common environment for landsat rlim detection over decades paper analysis.

Relies on setup_rivers.sh having first been run in environment.

"""

## Main packages and compute

import xarray as xr
import marutils
from marutils import masking
import geopandas as gpd
import rioxarray
import numpy as np
import argparse
import os
from rasterio import features
import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
import datetime as dt
import pyproj

from dask_jobqueue import SLURMCluster as MyCluster
from dask.distributed import Client

## Plotting
sns.set_context('paper', rc={'font.family':'arial', 'font.size':7})
sns.set_style('ticks')

regions_colors = {
         'S' : 'tab:gray',
       'SSW' : 'tab:red',
        'SW' : 'tab:blue',
        'NW' : 'tab:orange',
       'NNW' : 'tab:cyan',
         'N' : 'tab:green',
        'NE' : 'tab:purple'
    }

db_href = "postgresql://tedstona@localhost/rlims"

## Data, masks
class gisloader:

    # Numbers are inclusive !
    regions_boxes = {
         'S' : ( 2,  3), # 2
       'SSW' : ( 4,  7), # 4
        'SW' : ( 8, 11), # 4
        'NW' : (12, 15), # 4
        'NNW': (16, 20), # 5
         'N' : (22, 27), # 6
        'NE' : (28, 32)  # 5
        # Note that SE is not analysable, so not included here.
    }

    # Manual identification of boxes with lots of aquifers
    aquifer_boxes = [1, 3, 5, 6, 19, 20]   # 44, 45, 46, 47, 48, 49, 50]
    impossible_boxes = [21, 45]

    MAR_DS_ROOT = '/flash/tedstona/MAR-v3.11.5-ERA5-10km/'
    fn_roff_as_mar_masks = os.path.join(os.environ['ANALYSIS_OUT'],'runoff_limit_polys/polys_as_mar_rasters_annual_10k.nc')
    fn_regions_as_mar_masks = os.path.join(os.environ['ANALYSIS_OUT'],'regions_as_mar_rasters_10k.nc')

    RACMO_DS_ROOT = '/flash/tedstona/RACMO/'
    racmo_proj = pyproj.Proj(3413)
    fn_roff_as_racmo_masks = os.path.join(os.environ['ANALYSIS_OUT'],'runoff_limit_polys/polys_as_racmo_rasters_annual_1k.nc')
    fn_regions_as_racmo_masks = os.path.join(os.environ['ANALYSIS_OUT'],'regions_as_racmo_rasters_1k.nc')


    def __init__(self):      
        # Open a sample MAR dataset, this is used mainly for accessing georeferencing
        #self.mar_ds = marutils.open_dataset(os.path.join(self.MAR_DS_ROOT, 'MARv3.11.2-6km-daily-ERA5-2019.nc'))
        self.mar_ds = marutils.open_dataset(os.path.join(self.MAR_DS_ROOT, 'MARv3.11.5-10km-daily-ERA5-2020.nc')) #ICE.2020.01-12.a69.nc
        self.gris_mask = masking.gris_mask(self.mar_ds)

        # Sample RACMO dataset
        self.racmo_ds = xr.open_dataset(os.path.join(self.RACMO_DS_ROOT, '1km', 'Icemask_Topo_Iceclasses_lon_lat_average_1km_GrIS.nc'))
        self.racmo_ds.rio.write_crs(self.racmo_proj.srs, inplace=True)

        # Open the 100 km boxes and generate nice labels for them based on the regions specified above.
        self.boxes = gpd.read_file(os.path.join(os.environ['L0lib'], '100k_boxes/boxes.shp'))
        self.boxes.index = self.boxes.FID
        self.boxes['nice_label'] = 'none'
        self.boxes['region'] = 'none'
        for region in self.regions_boxes:
            bmin, bmax = self.regions_boxes[region]
            nn = 1
            for b in range(bmin, bmax+1):
                self.boxes.loc[b, 'nice_label'] = '{r}-{n}'.format(r=region, n=nn)
                self.boxes.loc[b, 'region'] = region
                nn += 1
        # Version of boxes geometry in MAR CRS
        self.boxes_mar = self.boxes.to_crs(self.mar_ds.rio.crs)
        # And in RACMO
        self.boxes_racmo = self.boxes.to_crs(self.racmo_ds.rio.crs)

        # Use GIMPDEM@MAR res, not MAR DEM - the rest of the processing pipeline uses GIMPDEM, be consistent.
        gimpdem_f = os.path.join(os.environ['L0lib'], 'GIMPDEM/gimpdem_90m_v01.1.MAR10km.tif')
        self.gimpdem = xr.open_rasterio(gimpdem_f)
        self.gimpdem['x'] = self.mar_ds.x
        self.gimpdem['y'] = self.mar_ds.y[::-1]

        # Ice sheet outline
        # _fix refers to the geometry in this file having been fixed of illegal geometries using geopandas buffer(0.1).
        outline_shpf = os.path.join(os.environ['L0lib'],'gris_only_outline/greenland_icesheet_fix.shp')
        self.gris_outline = gpd.read_file(outline_shpf)
        self.gris_outline = self.gris_outline.to_crs(3413)

        # Areas in which retrieval of the runoff limit is not possible
        self.nozones = gpd.read_file(os.path.join(os.environ['ANALYSIS_OUT'], 'impossible_rlim_zones/impossible_rlim_zones.shp')).to_crs(3413)
        # A version of the 'nozones' shapefile in which the 'box' polygons have been adjusted to conform to the GrIS geometry.
        # nozones_conform shpfile is generated by uncommenting relevant lines of plot_map_decadal_change.py
        self.nozones_conform = gpd.read_file(os.path.join(os.environ['ANALYSIS_OUT'], 'impossible_rlim_zones/impossible_rlim_zones_conform.shp'))

        # Do MAR mask checks
        if os.path.exists(self.fn_roff_as_mar_masks):
            self.roff_as_mar_masks = xr.open_dataarray(self.fn_roff_as_mar_masks)
        else:
            self.roff_as_mar_masks = None
            print('load_env.py Warning: no rasterised annual masks for MAR exist.')

        if os.path.exists(self.fn_regions_as_mar_masks):
            self.regions_as_mar_masks = xr.open_dataarray(self.fn_regions_as_mar_masks)
        else:
            self.regions_as_mar_masks = None
            print('load_env.py Warning: no rasterised masks of regions for MAR exist.')

        # Do RACMO mask checks
        if os.path.exists(self.fn_roff_as_racmo_masks):
            self.roff_as_racmo_masks = xr.open_dataarray(self.fn_roff_as_racmo_masks)
        else:
            self.roff_as_racmo_masks = None
            print('load_env.py Warning: no rasterised annual masks for RACMO exist.')

        if os.path.exists(self.fn_regions_as_racmo_masks):
            self.regions_as_racmo_masks = xr.open_dataarray(self.fn_regions_as_racmo_masks)
        else:
            self.regions_as_racmo_masks = None
            print('load_env.py Warning: no rasterised masks of regions for RACMO exist.')


    def _get_region_boxes_ids(self):
        valid_boxes = []
        for region in self.regions_boxes:
            bmin, bmax = self.regions_boxes[region]
            valid_boxes.extend(range(bmin, bmax+1))  #regions_boxes values are inclusive
        return valid_boxes

    def conform_bad_regions_to_margin(self):
        valid_boxes = self._get_region_boxes_ids()
        only_bad = self.boxes.drop(valid_boxes)
        
        only_bad['geometry'] = only_bad.buffer(10000)
        nozones_aq = pd.concat((self.nozones_conform, only_bad), axis=0)
        nozones_conform_aq = gpd.overlay(nozones_aq, self.gris_outline, how='intersection')

        return nozones_conform_aq

    def filter_regions_boxes(self):
        return self.boxes.filter(items=self._get_region_boxes_ids(), axis=0)



class emloader:

    def __init__(self, gis):
        # Generated by calculate_racmo_excess_melt.py
        self.excess_melt = xr.open_dataset(os.path.join(os.environ['ANALYSIS_OUT'],'RACMO_1km_excess_melt_mbyear.nc'))
        self.excess_melt.rio.write_crs(gis.racmo_ds.rio.crs)


class rlimsloader:

    def __init__(self):
        # Various 'reduced' annual runoff limit measurements.
        # Interface with the main 'rlims' database which contains all raw points is not addressed here.
        self.xytpd = pd.read_csv(os.path.join(os.environ['ANALYSIS_OUT'], 'rlim_annual_maxm/xytpd.csv'), 
            index_col=0, parse_dates=['date'])
        self.rlims = pd.read_csv(os.path.join(os.environ['ANALYSIS_OUT'], 'rlim_annual_maxm/100km_box_medians.csv'), index_col=0)
        self.rlims.columns = pd.to_numeric(self.rlims.columns).astype(int)

        self.rlims_plot = self.rlims.sort_index(ascending=True)
        self.rlims_plot.loc[:,1996] = np.nan
        self.rlims_plot.loc[28:32,2001:2002] = np.nan
        self.rlims_plot.loc[1:8, :1991] = np.nan

    def get_annual_polys(self, poly_type='median'):
        """ Load annual polygons of specified type/variety into a geodataframe. 

        poly_type : str, one of median, plus_mad, minus_mad, emax

        """

        filename_pattern = 'poly_{y}'
        if poly_type != 'median':
            filename_pattern += '_xytpd_{m}'.format(m=poly_type)
        filename_pattern += '.shp'

        store = []
        for year in range(1985, 2021):
            c = gpd.read_file(os.path.join(os.environ['ANALYSIS_OUT'], 'runoff_limit_polys', filename_pattern.format(y=year)))
            c['year'] = year
            store.append(c)
        annual_rlims = pd.concat(store, axis=0)

        return annual_rlims






## -----------------------------------------------------------------------------
## Ancillary functions

# Define the MAR dimensions to rasterize each box to

def polygon_to_mar_mask(mar_ds, conform_with=None, polygon_file=None, polygons=None,
    invert=False, nc=False, all_touched=False):

    try:
        out_shape = (mar_ds.dims['y'], mar_ds.dims['x'])
    except TypeError:
        out_shape = mar_ds.shape
    
    if polygon_file is not None:
        if not os.path.exists(polygon_file):
            raise IOError
        polygons = gpd.read_file(polygon_file)
    
    # Adjust the geometry to include the no-detection areas
    if conform_with is not None:
        polygons = polygons.append(conform_with)
        polygons['FID'] = 0
        polygons['geometry'] = polys.dissolve(by='FID')
        
    # Following the dissolve, can now convert to MAR projection
    polygons = polygons.to_crs(mar_ds.rio.crs)
    polygons['geometry'] = polygons.buffer(1000)
    #poly = polys.iloc[0]

    # also need to generate box-range masks corresponding to the geographic areas of interest.
    # Proceed per year

    # Rasterize the box polygon
    year_mask = features.rasterize(
        ((poly, 1) for poly in polygons.geometry),
        out_shape=out_shape,
        transform=mar_ds.rio.transform(),
        all_touched=all_touched)
    
    year_mask = year_mask.astype(np.bool)
    if invert:
        year_mask = ~year_mask

    if nc:
        year_mask = xr.DataArray(year_mask, dims=('y', 'x'), 
            coords={'y':mar_ds.y, 'x':mar_ds.x})
    
    return year_mask


def get_box_slice_counts(rlims):
    return rlims.xytpd.groupby(['box_id', 'year']).elev.count().unstack()


def get_region_availability(gis, rlims, slices_thresh=0.2):
    """
    slices_thresh : cut-off value, 0-->1. Minimum proportion of slices within 
    the region which must have observations for the year to be retained.
    """
    box_counts = get_box_slice_counts(rlims)
    region_available = {}
    for region in gis.regions_boxes:
        bmin, bmax = gis.regions_boxes[region]
        max_slices = (bmax - bmin) * 100
        box_count = box_counts.filter(items=range(bmin, bmax), axis='index')
        i = box_count.sum(axis=0)
        i = i.where(i > slices_thresh*max_slices)
        i.index = [dt.datetime(int(y), 8, 31) for y in i.index]
        region_available[region] = i
    return pd.DataFrame.from_dict(region_available, orient='columns')


import string
def letters(start=0, letters=list(string.ascii_lowercase)):
    """ Return letters of alphabet, in order, 1 by 1.
    
    :param start: integer position in letters list at which to begin.
    
    Usage:
    gen = letters()
    next(gen)
    Out: 'a'
    
    gen = letters(start=2)
    next(gen)
    Out: 'c'
    """
    ii = 0
    for i in letters:
        if ii < start:
            ii += 1
            continue
        yield(i) 


################################################################################
# Run script

if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument('-cluster', dest='start_cluster', default=False, 
        action='store_true')
    parser.add_argument('-gis', dest='load_gis', default=False, 
        action='store_true')
    parser.add_argument('-em', dest='load_excess_melt', default=False, action='store_true')
    parser.add_argument('-rlims', dest='load_rlims', default=False, action='store_true')

    args = parser.parse_args()
    
    if args.start_cluster:
        cluster = SGECluster()
        client = Client(cluster)

    if args.load_gis:
        gis = gisloader()

    if args.load_excess_melt:
        em = emloader()

    if args.load_rlims:
        rlims = rlimsloader()