"""
Given a contour polygon of Greenland, draw boxes spanning 100 km lengths of the 
contour line and which extend out by 100s km ~perpendicular to the contour.

AT, October 2020

"""

import geopandas as gpd
import pandas as pd
import numpy as np	
import matplotlib.pyplot as plt
from shapely.geometry import Point, Polygon


# Load the contour of interest
contours = gpd.read_file('/scratch/BedMachine.v3.contours/contour.shp')
# Look up the contour ID using QGIS
c25k = contours[contours.index == 35555]
# Simplify it to reduce problems in east Greenland
c25k = c25k.simplify(tolerance=50000)
g = c25k.geometry
gg = g.values[0]

# Extract the individual x,y points that compose the contour
x, y = gg.xy
x = pd.Series(x)
y = pd.Series(y)

# Calculate distances between points
dists = np.sqrt( np.abs(x.shift(1) - x)**2 + np.abs(y.shift(1) - y)**2 )
dists[0] = 0.0

# Convert to dataframe
combo = pd.DataFrame({'x':x, 'y':y, 'dist':dists, 'dist_cum':dists.cumsum()})

# Convert cumulative distance to rounded kilometres
combo['dcr'] = np.round(combo.dist_cum / 1000)
# Remove rows with duplicated rounded kilometres
combo = combo.drop_duplicates(subset='dcr', keep='first')
# Make round.km the index
combo.index = combo.dcr
# Reindex and interpolate to ensure it is regular
regular_ix = np.arange(0, np.max(combo.index), 1)
combo = combo.reindex(regular_ix)
combo = combo.interpolate()

# We are only interested in the points at 100 km increments, find them and filter them
max_val = np.ceil(combo.dcr.iloc[-1] / 100) * 100
keep_dcrs = np.arange(0, max_val, 100)
coords = combo.filter(items=keep_dcrs, axis='index')

# Calculate angle between these points
coords['angle'] = np.arctan2(
	(coords.y.shift(-1, fill_value=coords.iloc[0].y) - coords.y), 
	(coords.x.shift(-1, fill_value=coords.iloc[0].x) - coords.x) )

## Calculate coordinates of the polygons

pt1 = [(x,y) for x,y in zip(coords.x, coords.y)]

pt2_x = coords.x + (3.5e5 * np.cos(coords.angle + np.deg2rad(90) ))
pt2_y = coords.y + (3.5e5 * np.sin(coords.angle + np.deg2rad(90) ))
pt2 = [(x,y) for x,y in zip(pt2_x, pt2_y)]

# Either make rectangles...
# pt3_x = pt2_x + (1e5 * np.cos(coords.angle) )
# pt3_y = pt2_y + (1e5 * np.sin(coords.angle) )
# pt3 = [(x,y) for x,y in zip(pt3_x, pt3_y)]
# ... or polygons with no gaps between them.
pt3_x = pt2_x.shift(-1, fill_value=pt2_x.iloc[0])
pt3_y = pt2_y.shift(-1, fill_value=pt2_y.iloc[0])
pt3 = [(x,y) for x,y in zip(pt3_x, pt3_y)]

pt4_x = coords.x.shift(-1, fill_value=coords.iloc[0].x)
pt4_y = coords.y.shift(-1, fill_value=coords.iloc[0].y)
pt4_x.iloc[-1] = coords.x.iloc[0]
pt4_y.iloc[-1] = coords.y.iloc[0]
pt4 = [(x,y) for x,y in zip(pt4_x, pt4_y)]

geom = [Polygon((p1, p2, p3, p4)) for p1, p2, p3, p4 in zip(pt1, pt2, pt3, pt4)]
ids = np.arange(0,len(geom), 1)
geoms = gpd.GeoDataFrame({'geometry':geom}, crs=3413 )

geoms.to_file('boxes.shp')

