"""

Given a shapefile of ~rectangular Polygons covering the ice sheet margins, 
divide each polygon into strips which are 1 km wide along the highest side of the
rectangle and which continue to lower elevations, where they may be more or less
than 1 km wide depending on the angles of the 'rectangle'. Output each set of
sub boxes to its own shapefile named after the field ID of the originating box.

The main boxes/polygons were originally generated by make_100k_bins.py. They
were then modified manually in QGIS, particularly along the south-east coast.

Every box must have 4 (5) points! (Point 5 is the closing point). This script
will not work with polygons with > 4 corners.

Uses the GIMP DEM to find the 2 high points of each rectangle.

(The lower elevation line cannot be estimated based on angle relative to high
elevation line as they are not true boxes. Instead needs to be node positions.)

AT, October 2020

"""


import geopandas as gpd
import rasterio
from scipy.interpolate import interp1d
import pandas as pd
import numpy as np
from shapely.geometry import Polygon
import os
from math import trunc
import pdb

boxes = gpd.read_file('/home/geoscience/nobackup_cassandra/L0data/100k_boxes/boxes.shp')

dem = '/flash/tedstona/L0data/GIMPDEM/gimpdem_30m_3413.vrt'
icemsk = '/flash/tedstona/L0data/GIMPMASK/gimpmask_30m_3413.vrt'
out_dir = '/home/geoscience/nobackup_cassandra/L0data/100k_boxes/'

# Width of each sub box at high elevation, metres
SUB_SIZE_M = 1000


for ix, b in boxes.iterrows():

	print(b.FID)

	x, y = b.geometry.exterior.coords.xy[0], b.geometry.exterior.coords.xy[1]

	# Find the 'inception points' by using the DEM to extract elevation of all 
	# points making up polygon
	with rasterio.open(dem, 'r') as d:
		with rasterio.open(icemsk, 'r') as m:

			# For a single box
			store = []
			for xx, yy in zip(x, y):
				py, px = d.index(xx, yy)
				window = rasterio.windows.Window(px-1//2, py-1//2, 1, 1)	
				clip = d.read(window=window, fill_value=0, boundless=True)

				py, px = m.index(xx, yy)
				window = rasterio.windows.Window(px-1//2, py-1//2, 1, 1)
				clip_m = m.read(window=window, fill_value=0, boundless=True)
				if float(clip_m.squeeze()) < 1:
					v = 0
				else:
					v = float(clip.squeeze())
				store.append(v)

	# Temporary data frame relating x,y to elevation
	sm = pd.DataFrame({'x':x, 'y':y, 'el':store})
	sm = sm.drop_duplicates(subset=('x','y'))

	## Begin to space out the 1 km bins by using the polygon edge corresponding to the 'home' high contour
	hipts = sm.drop_duplicates().sort_values(by='el', ascending=False).iloc[0:2]
	dist = np.sqrt( np.abs(hipts.x.iloc[1] - hipts.x.iloc[0])**2 + np.abs(hipts.y.iloc[1] - hipts.y.iloc[0])**2 )
	dist = np.array([0, dist])
	fx = interp1d(dist, hipts.x)
	fy = interp1d(dist, hipts.y)
	alpha = np.arange(0, dist[1], SUB_SIZE_M)
	x_reg, y_reg = fx(alpha), fy(alpha)


	## Now figure out the equivalent spacing on the downstream end of each 1km sliver
	# Ignore ix4 (i.e. position 5) as this is the same as ix0
	lpts = sm.drop_duplicates().sort_values(by='el', ascending=False).iloc[2:4]

#	pdb.set_trace()

	# Need to propagate the points in the same direction as we did for the high side,
	# so search for the distance of each low point from 'home' then use the
	# nearest one to commence propagation.
	# dist_from_home = np.sqrt( np.abs(hipts.x.iloc[0] - lpts.x.iloc[0])**2
	# 	+ np.abs(hipts.y.iloc[0] - lpts.y.iloc[0])**2 )
	# dist_from_home2 = np.sqrt( np.abs(hipts.x.iloc[0] - lpts.x.iloc[1])**2 
	# 	+ np.abs(hipts.y.iloc[0] - lpts.y.iloc[1])**2 )
	# if dist_from_home < dist_from_home2:
	# 	lx = lpts.x
	# 	ly = lpts.y
	# else:
	# 	lx = lpts.x[::-1]
	# 	ly = lpts.y[::-1]

	if (hipts.iloc[0].name == 0) and (hipts.iloc[1].name == 3):
		from_pt = 1
		to_pt = 2
	elif (hipts.iloc[0].name == 3) and (hipts.iloc[1].name == 0):
		from_pt = 2
		to_pt = 1
	elif (hipts.iloc[0].name == 0) and (hipts.iloc[1].name == 1):
		from_pt = 3
		to_pt = 2
	elif (hipts.iloc[0].name == 1) and (hipts.iloc[1].name == 0):
		from_pt = 2
		to_pt = 3
	elif (hipts.iloc[0].name == 3) and (hipts.iloc[1].name == 2):
		from_pt = 0
		to_pt = 1
	elif (hipts.iloc[0].name == 2) and (hipts.iloc[1].name == 3):
		from_pt = 1
		to_pt = 0
	else:
		raise ValueError('Could not deduce draw direction.')


	dist = np.sqrt(
		np.abs(sm.iloc[to_pt].x - sm.iloc[from_pt].x)**2
		+ np.abs(sm.iloc[to_pt].y - sm.iloc[from_pt].y)**2 )


	# dist = np.sqrt( 
	# 	np.abs(lpts.x.iloc[1] - lpts.x.iloc[0])**2 
	# 	+ np.abs(lpts.y.iloc[1] - lpts.y.iloc[0])**2 )
	dist = np.array([0, dist])
	fx2 = interp1d(dist, [sm.iloc[from_pt].x, sm.iloc[to_pt].x])
	fy2 = interp1d(dist, [sm.iloc[from_pt].y, sm.iloc[to_pt].y])
	alpha = np.linspace(start=0, stop=dist[1], num=len(x_reg))
	x2_reg, y2_reg = fx2(alpha), fy2(alpha)


	## Construct the polygons

	# Points 1 and 4 are on the high elevation side.
	# Point(4) corresponds to Point(1) of Polygon(n+1), hence shift()
	# Point(3) corresponds to Point(2) of Polygon(n+1), hence shift()
	#
	#      3------------------------4
	# x km |                        | 1 km
	#      2------------------------1

	x_reg = pd.Series(x_reg)
	y_reg = pd.Series(y_reg)
	x2_reg = pd.Series(x2_reg)
	y2_reg = pd.Series(y2_reg)

	xy1 = [(x,y) for x,y in zip(x_reg, y_reg)]
	xy2 = [(x,y) for x,y in zip(x2_reg, y2_reg)]

	x3 = x2_reg.shift(-1, fill_value=x2_reg.iloc[0])
	y3 = y2_reg.shift(-1, fill_value=y2_reg.iloc[0])
	xy3 = [(x,y) for x,y in zip(x3, y3)]

	x4 = x_reg.shift(-1, fill_value=x_reg.iloc[0])
	y4 = y_reg.shift(-1, fill_value=y_reg.iloc[0])
	xy4 = [(x,y) for x,y in zip(x4, y4)]

	geoms = [Polygon((p1,p2,p3,p4)) for p1, p2, p3, p4 in zip(xy1,xy2,xy3,xy4)]

	# Remove the last polygon generated, this corresponds to the full bounding
	# box (i.e. the original 100 km box, due to fill_value in shift() wrap-around)
	geoms = geoms[:-1]

	geomspd = gpd.GeoDataFrame({'geometry':geoms}, crs=3413)

	fn = 'boxes_fid{:d}.shp'.format(trunc(b.FID))
	fn = os.path.join(out_dir, fn)
	geomspd.to_file(fn)

