# ---
# jupyter:
#   jupytext:
#     cell_metadata_filter: all
#     formats: ipynb,py:percent
#     notebook_metadata_filter: all,-language_info
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.7.1
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Integrate manual common period runoff limits into the automatic ones

# %% [markdown]
# This script updates the automatic 1985-92 and 2013-20 runoff limits to include manually-delineated runoff areas.
#
# The manual delineation was done as follows:
#
# 1. Google Earth Engine -> rlims-landsat-mosaics, to create 7-year mosaics for the NW and NE areas of interest, and export as assets
# 2. Google Earth Engine -> rlims-landsat-mosaics-use-assets, which simply displays these mosaics. I then progressively manually digitised the areas into polygons, which this script can export to GDrive as shapefiles.
# 3. This script. Check out the notes below.

# %% trusted=true
import geopandas as gpd
import os

# %% trusted=true
from load_env import *
gis = gisloader()

# %% trusted=true
c8592 = gpd.read_file(os.path.join(os.environ['ANALYSIS_OUT'], 'runoff_limit_polys/poly_{y}.shp'.format(y='1985_1992_median')))
c8592['year'] = '1985_1992'
c8592_lo = gpd.read_file(os.path.join(os.environ['ANALYSIS_OUT'], 'runoff_limit_polys/poly_{y}.shp'.format(y='1985_1992_median_xytpd_minus_mad')))
c8592_lo['year'] = '1985_1992_lo'
c8592_hi = gpd.read_file(os.path.join(os.environ['ANALYSIS_OUT'], 'runoff_limit_polys/poly_{y}.shp'.format(y='1985_1992_median_xytpd_plus_mad')))
c8592_hi['year'] = '1985_1992_hi'
c8592_emax = gpd.read_file(os.path.join(os.environ['ANALYSIS_OUT'], 'runoff_limit_polys/poly_{y}.shp'.format(y='1985_1992_median_xytpd_region_emax')))
c8592_emax['year'] = '1985_1992_emax'

c1320 = gpd.read_file(os.path.join(os.environ['ANALYSIS_OUT'], 'runoff_limit_polys/poly_{y}.shp'.format(y='2013_2020_median')))
c1320['year'] = '2013_2020'
c1320_lo = gpd.read_file(os.path.join(os.environ['ANALYSIS_OUT'], 'runoff_limit_polys/poly_{y}.shp'.format(y='2013_2020_median_xytpd_minus_mad')))
c1320_lo['year'] = '2013_2020_lo'
c1320_hi = gpd.read_file(os.path.join(os.environ['ANALYSIS_OUT'], 'runoff_limit_polys/poly_{y}.shp'.format(y='2013_2020_median_xytpd_plus_mad')))
c1320_hi['year'] = '2013_2020_hi'
c1320_emax = gpd.read_file(os.path.join(os.environ['ANALYSIS_OUT'], 'runoff_limit_polys/poly_{y}.shp'.format(y='2013_2020_median_xytpd_region_emax')))
c1320_emax['year'] = '2013_2020_emax'

# %% trusted=true
# North-west manual limits
nw_early = gpd.read_file(os.path.join(os.environ['ANALYSIS_OUT'], 'gee_manual_rlims','NW_1985_1992_rlim.shp'))
nw_late = gpd.read_file(os.path.join(os.environ['ANALYSIS_OUT'], 'gee_manual_rlims','NW_2013_2020_rlim.shp'))
nw_early = nw_early.to_crs(3413)
nw_late = nw_late.to_crs(3413)

# %% trusted=true
# North-east manual limits
ne_early = gpd.read_file(os.path.join(os.environ['ANALYSIS_OUT'], 'gee_manual_rlims','NE_1985_1992_rlim.shp'))
ne_late = gpd.read_file(os.path.join(os.environ['ANALYSIS_OUT'], 'gee_manual_rlims','NE_2013_2020_rlim.shp'))
ne_early = ne_early.to_crs(3413)
ne_late = ne_late.to_crs(3413)

# %% trusted=true
## Fix the auto-polygonize problems in the NW corner
# First set up the ice margin geometry to substitute in to this area
conform_nw_area = gis.boxes.filter(items=[22], axis=0)
conform_nw_area['merge_on'] = 1
conform_nw_area = conform_nw_area.dissolve(by='merge_on')
conform_nw = gpd.overlay(conform_nw_area, gis.gris_outline, how='intersection')

# Now do the substitution
def _conform_nw(shp):
    shp_mod = shp.append(conform_nw)
    shp_mod['merge_on'] = 1
    shp_mod = shp_mod.dissolve(by='merge_on')
    return shp_mod

c8592_mod = _conform_nw(c8592)
c8592_lo_mod = _conform_nw(c8592_lo)
c8592_hi_mod = _conform_nw(c8592_hi)
c8592_emax_mod = _conform_nw(c8592_emax)

c1320_mod = _conform_nw(c1320)
c1320_lo_mod = _conform_nw(c1320_lo)
c1320_hi_mod = _conform_nw(c1320_hi)
c1320_emax_mod = _conform_nw(c1320_emax)

# %% trusted=true
## Do the manual rlim substitutions for early period, both NW and NE
for ix, row in nw_early.iterrows():
    c8592_mod = c8592_mod.difference(row.geometry)
    c8592_lo_mod = c8592_lo_mod.difference(row.geometry)
    c8592_hi_mod = c8592_hi_mod.difference(row.geometry)
    c8592_emax_mod = c8592_emax_mod.difference(row.geometry)

c8592_mod = c8592_mod.difference(ne_early.iloc[0].geometry)
c8592_lo_mod = c8592_lo_mod.difference(ne_early.iloc[0].geometry)
c8592_hi_mod = c8592_hi_mod.difference(ne_early.iloc[0].geometry)
c8592_emax_mod = c8592_emax_mod.difference(ne_early.iloc[0].geometry)


# %% trusted=true
## Same for the late period...
for ix, row in nw_late.iterrows():
    c1320_mod = c1320_mod - row.geometry
    c1320_lo_mod = c1320_lo_mod - row.geometry
    c1320_hi_mod = c1320_hi_mod - row.geometry
    c1320_emax_mod = c1320_emax_mod - row.geometry
    
# ... except, note that we have to subtract ne_early as well.
# This is because there is no 'duplicate geometry' option in GEE. So rather than re-draw the 
# 1985-92 limits with only a small modification for 2013-20, I've instead for 2013-20 just 
# digitised the change in area compared with 1985-92.
c1320_mod = c1320_mod - ne_late.iloc[0].geometry - ne_early.iloc[0].geometry
c1320_lo_mod = c1320_lo_mod - ne_late.iloc[0].geometry - ne_early.iloc[0].geometry
c1320_hi_mod = c1320_hi_mod - ne_late.iloc[0].geometry - ne_early.iloc[0].geometry
c1320_emax_mod = c1320_emax_mod - ne_late.iloc[0].geometry - ne_early.iloc[0].geometry

# %% trusted=true
## Plot to check the results.
# %matplotlib widget
import matplotlib.pyplot as plt
plt.figure()
ax = plt.subplot(111)
gis.gris_outline.plot(ax=ax, color='#ececec')
#nw_early.plot(ax=ax)
c8592.plot(ax=ax, edgecolor='lightblue', color='none')
c8592_mod.plot(ax=ax, edgecolor='black', color='none')
c8592_lo_mod.plot(ax=ax, edgecolor='black', color='none', linestyle=':')
c8592_hi_mod.plot(ax=ax, edgecolor='black', color='none', linestyle=':')
#c8592_emax.plot(ax=ax, edgecolor='red', color='none', linestyle=':', zorder=1000)
c8592_emax_mod.plot(ax=ax, edgecolor='orange', color='none', linestyle='--')

plt.figure()
ax = plt.subplot(111)
gis.gris_outline.plot(ax=ax, color='#ececec')
c1320.plot(ax=ax, edgecolor='lightblue', color='none')
c1320_mod.plot(ax=ax, edgecolor='black', color='none')
c1320_lo_mod.plot(ax=ax, edgecolor='black', color='none', linestyle=':')
c1320_hi_mod.plot(ax=ax, edgecolor='black', color='none', linestyle=':')
c1320_emax.plot(ax=ax, edgecolor='red', color='none', linestyle=':', linewidth=1)
c1320_emax_mod.plot(ax=ax, edgecolor='orange', color='none', linestyle='-')
pts = pd.read_csv('/flash/tedstona/analysis/runoff_limit_polys/poly_2013_2020_median_xytpd_region_emax.csv')
plt.plot(pts.x, pts.y, '.b')
plt.show()
#nozones_conform_aq.plot(ax=ax, color='red', alpha=0.2)

# %% trusted=true
# Export the results.
c8592_mod.to_file(os.path.join(os.environ['ANALYSIS_OUT'],'runoff_limit_polys', 'poly_1985_1992_median_edited.shp'))
c8592_lo_mod.to_file(os.path.join(os.environ['ANALYSIS_OUT'],'runoff_limit_polys', 'poly_1985_1992_median_xytpd_minus_mad_edited.shp'))
c8592_hi_mod.to_file(os.path.join(os.environ['ANALYSIS_OUT'],'runoff_limit_polys', 'poly_1985_1992_median_xytpd_plus_mad_edited.shp'))
c8592_emax_mod.to_file(os.path.join(os.environ['ANALYSIS_OUT'],'runoff_limit_polys', 'poly_1985_1992_median_xytpd_region_emax_edited.shp'))

c1320_mod.to_file(os.path.join(os.environ['ANALYSIS_OUT'],'runoff_limit_polys', 'poly_2013_2020_median_edited.shp'))
c1320_lo_mod.to_file(os.path.join(os.environ['ANALYSIS_OUT'],'runoff_limit_polys', 'poly_2013_2020_median_xytpd_minus_mad_edited.shp'))
c1320_hi_mod.to_file(os.path.join(os.environ['ANALYSIS_OUT'],'runoff_limit_polys', 'poly_2013_2020_median_xytpd_plus_mad_edited.shp'))
c1320_emax_mod.to_file(os.path.join(os.environ['ANALYSIS_OUT'],'runoff_limit_polys', 'poly_2013_2020_median_xytpd_region_emax_edited.shp'))

# %% trusted=true
c8592_emax_mod.to_file(os.path.join(os.environ['ANALYSIS_OUT'],'runoff_limit_polys', 'poly_1985_1992_median_xytpd_region_emax_edited.shp'))
c1320_emax_mod.to_file(os.path.join(os.environ['ANALYSIS_OUT'],'runoff_limit_polys', 'poly_2013_2020_median_xytpd_region_emax_edited.shp'))
