# ---
# jupyter:
#   jupytext:
#     cell_metadata_filter: all
#     formats: ipynb,py:percent
#     notebook_metadata_filter: all,-language_info
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.7.1
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Plot map of early versus late runoff limits

# %% trusted=true
from load_env import *
gis = gisloader()


# %% [markdown]
# ## Ancillary data loading

# %% trusted=true
# World land areas
greenland = gpd.read_file(os.path.join(os.environ['L0lib'], 'NaturalEarth/ne_10m_land/ne_10m_land.shp'))
# Crop world land areas to Greenland and surrounding areas
bbox = gpd.read_file(os.path.join(os.environ['L0lib'], 'greenland_region_bbox/greenland_area_bbox.shp')).to_crs(3413)
just_greenland = gpd.clip(greenland.to_crs(3413), bbox)

# Manually isolate contiguous Greenland polygon from the main multi-polygon.
jg = just_greenland.filter(items=[0], axis=0)
jgg = jg.loc[0].geometry
jgg_poly = jgg[9]
jgg_gdf = gpd.GeoDataFrame({'ix':[1,]}, geometry=[jgg_poly], crs=3413)

# %% trusted=true
# Aquifer locations
aq_store = []
for y in range(2010, 2015):
    aq = gpd.read_file(os.path.join(
        os.environ['ANALYSIS_OUT'], 
        'miege_aquifers/MiegeFirnAquiferDetections{y}.shp'.format(y=y))
                                ).to_crs(3413)
    aq['year'] = y
    aq_store.append(aq)
aqs = pd.concat(aq_store, axis=0)

# %% trusted=true
# Surface contours
# shp : greenland_icesheet.shp from Horst, run through geopandas simplify and buffer operations
# gdalwarp -cutline ../gris_only_outline/greenland_icesheet_simplify5000_buffer0.shp GimpIceMask_90m_v1.1_epsg3413.tif GimpIceMask_90m_v1.1_epsg3413_gris.tif
# gdal_calc.py -A gimpdem_90m_v01.1_EPSG3413.tif -B ../GIMPMASK/GimpIceMask_90m_v1.1_epsg3413_gris.tif --calc=A*B --outfile=gimpdem_90m_v01.1_EPSG3413_grisonly.tif
# gdalwarp -tr 2000 2000 gimpdem_90m_v01.1_EPSG3413_grisonly_filled.tif gimpdem_90m_v01.1_EPSG3413_grisonly_filled_2km.tif
# gdal_contour -i 500 gimpdem_90m_v01.1_EPSG3413_grisonly_filled_2km.tif gimpdem_90m_v01.1_EPSG3413_grisonly_contours_2km_i500 -f 'ESRI Shapefile' -a elev
surf_contours = gpd.read_file(os.path.join(os.environ['L0lib'], 'GIMPDEM', 'gimpdem_90m_v01.1_EPSG3413_grisonly_contours_2km_i500', 'contour.shp')).to_crs(3413)

# %% trusted=true
# Manually delineated region divides
# These need to be updated` manually if region definitions change!
region_divides = gpd.read_file(os.path.join(os.environ['ANALYSIS_OUT'], 'manual_region_divides/manual_region_divides.shp'))

# %% [markdown]
# ## Contours

# %% trusted=true
# Contours
c_early = gpd.read_file(os.path.join(os.environ['ANALYSIS_OUT'], 'runoff_limit_polys/poly_1985_1992_median_edited.shp'))

c_early_lo = gpd.read_file(os.path.join(os.environ['ANALYSIS_OUT'], 'runoff_limit_polys/poly_{y}_edited.shp'.format(y='1985_1992_median_xytpd_minus_mad')))
c_early_hi = gpd.read_file(os.path.join(os.environ['ANALYSIS_OUT'], 'runoff_limit_polys/poly_{y}_edited.shp'.format(y='1985_1992_median_xytpd_plus_mad')))
c_early_emax = gpd.read_file(os.path.join(os.environ['ANALYSIS_OUT'], 'runoff_limit_polys/poly_{y}_edited.shp'.format(y='1985_1992_median_xytpd_region_emax')))

c_late = gpd.read_file(os.path.join(os.environ['ANALYSIS_OUT'], 'runoff_limit_polys/poly_2013_2020_median_edited.shp'))
c_late_lo = gpd.read_file(os.path.join(os.environ['ANALYSIS_OUT'], 'runoff_limit_polys/poly_{y}_edited.shp'.format(y='2013_2020_median_xytpd_minus_mad')))
c_late_hi = gpd.read_file(os.path.join(os.environ['ANALYSIS_OUT'], 'runoff_limit_polys/poly_{y}_edited.shp'.format(y='2013_2020_median_xytpd_plus_mad')))
c_late_emax = gpd.read_file(os.path.join(os.environ['ANALYSIS_OUT'], 'runoff_limit_polys/poly_{y}_edited.shp'.format(y='2013_2020_median_xytpd_region_emax')))

# %% [markdown]
# ## Transform no-coverage zones to hug the ice sheet margins

# %% [markdown]
# The 'impossible runoff limit zones' have been defined manually in QGIS. They are polygons which extend over the area to exclude.

# %% trusted=true
# Uncomment to regenerate and save the conformed no-zones.
#nozones_conform = gpd.overlay(nozones, gris_outline, how='intersection')
#nozones_conform.to_file(os.path.join(os.environ['ANALYSIS_OUT'], 'impossible_rlim_zones/impossible_rlim_zones_conform.shp'))


# %% [markdown]
# ## Examination & Plotting


# %% trusted=true
import scalebar
import cartopy.crs as ccrs
# %matplotlib inline

# %% trusted=true
# Colours
col_early_area = '#0c4969'
col_late_area = '#df778e'


# %% trusted=true
fig = plt.figure(figsize=(8,12))

# Define the CartoPy CRS object.
crs = ccrs.NorthPolarStereo(central_longitude=-45., true_scale_latitude=70.)
ax = plt.axes(projection=crs)

# This can be converted into a `proj4` string/dict compatible with GeoPandas
crs_proj4 = crs.proj4_init

ax.set_facecolor('white')
jgg_gdf.plot(ax=ax, color='#CDB380', zorder=1)

gis.gris_outline.plot(ax=ax, color=col_early_area, edgecolor='none', alpha=1, zorder=10)

c_early.plot(ax=ax, color=col_late_area, zorder=30, edgecolor='none') #,
c_late.plot(ax=ax, color='whitesmoke', edgecolor='none', zorder=40)

surf_contours.plot(ax=ax, edgecolor='#c2c2c2', linewidth=0.7, zorder=100, alpha=0.5) ##E8DDCB

plt.plot([v.geometry.x for i,v in aqs.iterrows()], [v.geometry.y for i,v in aqs.iterrows()], 
         '.', markersize=2, mfc='#7cbc9c', mec='none', alpha=0.2, zorder=101)

region_divides.plot(ax=ax, color='k', alpha=1, linewidth=1, zorder=102)

# x0, x1, y0, y1
ax.set_extent([-692338, 916954, -3392187, -627732], crs=crs)
ax.gridlines(draw_labels=True, xlocs=[-50, -35], ylocs=[65, 75], x_inline=False, y_inline=False, zorder=120)

scalebar.scale_bar(ax, (0.65, 0.06), 300)

plt.box(on=None)

plt.savefig(os.path.join(os.environ['FIGS_OUT'],'main/fig1_map.pdf'), dpi=300)

# %% [markdown]
# ## Metrics

# %% trusted=true
# Not actually interested in the area of each runoff polygon/contour. Rather, we're interested in the 
# outside the polygon but inside the GrIS margins.

area_early = (gis.gris_outline.area - c_early.area).iloc[0] *1e-6
area_early_lo = (gis.gris_outline.area - c_early_lo.area).iloc[0] *1e-6
area_early_hi = (gis.gris_outline.area - c_early_hi.area).iloc[0] *1e-6
area_early_emax = (gis.gris_outline.area - c_early_emax.area).iloc[0] *1e-6

area_late = (gis.gris_outline.area - c_late.area).iloc[0] *1e-6
area_late_lo = (gis.gris_outline.area - c_late_lo.area).iloc[0] *1e-6
area_late_hi = (gis.gris_outline.area - c_late_hi.area).iloc[0] *1e-6
area_late_emax = (gis.gris_outline.area - c_late_emax.area).iloc[0] *1e-6

early_emax_perc = 100 / area_early * area_early_emax
late_emax_perc = 100 / area_late * area_late_emax

area_change = area_late - area_early
perc_change = ((100 / area_early) * area_late) - 100

area_change_small = area_late_lo - area_early_hi
perc_change_small = ((100 / area_early_hi) * area_late_lo) - 100
area_change_large = area_late_hi - area_early_lo
perc_change_large = ((100 / area_early_lo) * area_late_hi) - 100

perc_minusmad = ((100 / area_early) * area_late_lo) - 100
perc_plusmad = ((100 / area_early) * area_late_hi) - 100

print('Area, early period: {lo} - {v} - {hi} ({emax}) km2'.format(lo=area_early_lo, v=area_early, hi=area_early_hi, emax=area_early_emax))
print('Area, early emax % to median: {v}'.format(v=early_emax_perc))
print('Area, late period: {lo} - {v} - {hi} ({emax}) km2'.format(lo=area_late_lo, v=area_late, hi=area_late_hi, emax=area_late_emax))
print('Area, late emax % to median: {v}'.format(v=late_emax_perc))
print('Area change: {v} km2'.format(v=area_change))
print('Area change %: {v} '.format(v=perc_change))
print('Smallest area change %: {v}'.format(v=perc_change_small))
print('Largest area change %: {v}'.format(v=perc_change_large))

print('2013-20 -1MAD compared to 1985-92 MED %: {v}'.format(v=perc_minusmad))
print('2013-20 +1MAD compared to 1985-92 MED %: {v}'.format(v=perc_plusmad))

# %% trusted=true
area_max = gis.gris_outline.area - c_max.area
print(area_max)
print(((100/area_early) * area_max) -100)

# %% trusted=true
# %matplotlib widget
plt.figure(figsize=(3,1.7))
ax = plt.subplot(111)
plt.hlines(2, area_early_lo, area_early_hi, linewidth=12, color=col_early_area, alpha=0.3)
plt.vlines(area_early, 1.7, 2.3, color=col_early_area, linewidth=2, alpha=1)
plt.plot(area_early_emax, 2, color=col_early_area, linewidth=1, alpha=0.9, marker='s')

plt.hlines(1, area_late_lo, area_late_hi, linewidth=12, color=col_late_area, alpha=0.3)
plt.vlines(area_late, 0.7, 1.3, color=col_late_area, linewidth=2, alpha=1)
plt.plot(area_late_emax, 1, color=col_late_area, linewidth=1, alpha=0.9, marker='s')

plt.ylim(0.4,2.6)
plt.xlabel(r'|$\acute\upsilon$| (km$^2$)')

ax_perc = ax.twiny()

import matplotlib.ticker as ticker
#ps = (100 / area_early)
ticks_as_perc = np.arange(.90, 1.4, .1)
ticks_as_data = ticks_as_perc * area_early
#ticks = np.arange(ps*area_early_lo, ps*area_late_hi, 10)
#ticks = np.linspace(area_early_lo, area_late_hi, 10)
ax_perc.set_xticks(ticks=ticks_as_data)
ax_perc.set_xlim(ax.get_xlim()[0], ax.get_xlim()[-1])
ax_perc.xaxis.set_major_formatter(ticker.PercentFormatter(xmax=area_early, decimals=0))

plt.tight_layout()
sns.despine(left=True, top=False)
ax.yaxis.set_visible(False)

plt.savefig(os.path.join(os.environ['FIGS_OUT'], 'main', 'fig1_area_change.pdf'))

# %% [markdown]
# ## Supplementary figure: coincidence with MacFerrin ice slabs

# %% trusted=true
oib_slabs = gpd.read_file(os.path.join(os.environ['ANALYSIS_OUT'], 'macferrin_slabs_obs/IceBridgeArea_Shape.shp'))
oib_slabs = oib_slabs.to_crs(3413)

# %% trusted=true
# IceBridge flight lines
h5 = pd.read_hdf(os.path.join(os.environ['ANALYSIS_OUT'],'macferrin_slabs_obs/IceBridgeDB.h5'), key='Accumulation/Coordinates')
h5 = h5[h5.OnGreenlandIce]
# Coarsen
h5 = h5.iloc[::500]
h5_gdf = gpd.GeoDataFrame(h5, geometry=gpd.points_from_xy(h5.Longitude, h5.Latitude), crs=4326)
h5_gdf = h5_gdf.to_crs(3413)

# %% trusted=true
# %matplotlib inline
fig = plt.figure(figsize=(8,12))

# Define the CartoPy CRS object.
crs = ccrs.NorthPolarStereo(central_longitude=-45., true_scale_latitude=70.)
ax = plt.axes(projection=crs)

# This can be converted into a `proj4` string/dict compatible with GeoPandas
crs_proj4 = crs.proj4_init

ax.set_facecolor('white')
just_greenland.plot(ax=ax, color='#e3dfba', zorder=1)

gis.gris_outline.plot(ax=ax, color='powderblue', edgecolor='none', alpha=1, zorder=10)

c_early.plot(ax=ax, color='#df778e', zorder=30, edgecolor='none') #,
c_late.plot(ax=ax, color='whitesmoke', edgecolor='none', zorder=40)

h5_gdf.plot(ax=ax, marker='o', markersize=1, zorder=45, color='grey', alpha=0.3)
oib_slabs.plot(ax=ax, color='none', edgecolor='black', linestyle='-', linewidth=1.5, zorder=50)

surf_contours.plot(ax=ax, edgecolor='#E8DDCB', linewidth=0.5, zorder=100, alpha=0.5)

# x0, x1, y0, y1
ax.set_extent([-692338, 916954, -3392187, -627732], crs=crs)
ax.gridlines(draw_labels=True, xlocs=[-50, -35], ylocs=[65, 75], x_inline=False, y_inline=False, zorder=120)

scalebar.scale_bar(ax, (0.65, 0.06), 300, zorder=200)

plt.box(on=None)

plt.savefig(os.path.join(os.environ['FIGS_OUT'],'suppl/sf_ice_slabs.pdf'), dpi=300)

# %%
