# ---
# jupyter:
#   jupytext:
#     cell_metadata_filter: all
#     formats: ipynb,py:percent
#     notebook_metadata_filter: all,-language_info
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.7.1
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Main figure: temporal heatmap(s)

# %% [markdown]
# Uses outputs from `investigate_inter_annual.py`

# %% trusted=true
from load_env import *
gis = gisloader()
rlims = rlimsloader()

# %% trusted=true
# Exclude certain boxes
for v in gis.impossible_boxes:
    rlims.rlims.loc[v] = np.nan

# %% trusted=true
# get count of available slices per box per year - functionality is in load_env.py.
box_slice_counts = get_box_slice_counts(rlims)

# %% trusted=true
rlims_labels = rlims.rlims.join(gis.boxes, how='outer')
rlims_labels.head()
rlims_labels = rlims_labels.sort_index(ascending=False)


# %% trusted=true
def label_aquifers(axis):
    ylabels = axis.yaxis.get_ticklabels()
    new_vals = []
    for v in ylabels:
        vv = int(v.get_text())
        if vv in gis.aquifer_boxes:
            new_vals.append('* {v}'.format(v=vv))
        else:
            new_vals.append(vv)
    axis.yaxis.set_ticklabels(new_vals)
    
   
    
from copy import deepcopy
def hatch_impossible(rlims_here, impossible_list):
    rlims_here = deepcopy(rlims_here)
    rlims_here.loc[:,:] = np.nan
    for v in impossible_list:
        rlims_here.loc[v,:] = 1 
    x = np.arange(len(rlims_here.columns)+1)
    y = np.arange(len(rlims_here.index)+1)
    plt.pcolor(x, y, rlims_here, hatch='//////', alpha=0.)

    
def hatch_bad(counts, impossible_list, thresh):
    counts_here = deepcopy(counts)
    counts_here.loc[:,:] = np.where(counts_here < thresh, 1, np.nan)
    for v in impossible_list:
        counts_here.loc[v,:] = np.nan 
    x = np.arange(len(counts_here.columns)+1)
    y = np.arange(len(counts_here.index)+1)
    plt.pcolor(x, y, counts_here, hatch='....', alpha=0.)

    
def finish_formatting(ax):
    ax.set_facecolor('#f0f0f0')
    plt.yticks(rotation=0)
    

    
thresh = 25

box_slice_counts = box_slice_counts.sort_index(ascending=False)

# %matplotlib inline
plt.figure(figsize=(8,11))

nn = len(gis.regions_boxes)
first = True
for region in gis.regions_boxes:
    if first:
        ax1 = plt.subplot(len(gis.regions_boxes), 1, nn)
        ax_here = ax1
    else:
        axn = plt.subplot(len(gis.regions_boxes), 1, nn) # , sharex=ax1
        ax_here = axn
    
    bmin, bmax = gis.regions_boxes[region]
    
    if first or nn == 1:
        xticks = 5 #'auto'
    else:
        xticks = []
        
    # min, max
    vmin = int(rlims.rlims.loc[bmax:bmin].min().min()) + 5
    vmax = int(rlims.rlims.loc[bmax:bmin].max().max())
    print(vmin, vmax)
    sns.heatmap(rlims.rlims.loc[bmax:bmin], yticklabels=rlims_labels.loc[bmax:bmin, 'nice_label'], 
                 cmap='flare_r', #vmin=1000, vmax=2100,
                ax=ax_here, square=True, 
                xticklabels=xticks,
                cbar_kws={'shrink':0.4, 'aspect':5, 'ticks':[vmin, vmax], 'pad':0.01})

    finish_formatting(ax_here)
    hatch_bad(box_slice_counts.loc[bmax:bmin], gis.impossible_boxes, thresh)
    
    if nn == 1:
        plt.tick_params(axis='x', which='both', 
                labelbottom=False, bottom=False, labeltop=True, top=True)
        
    if first:
        first = False
        plt.tick_params(axis='x', which='both', 
                labelbottom=True, bottom=True)
    
    nn -= 1

plt.savefig(os.path.join(os.environ['FIGS_OUT'],'main/fig1_regions_heatmaps.pdf'))


# %% [markdown]
# ## Region-by-region time series

# %% trusted=true
import statsmodels.api as sm

# Delete periods where there is insufficient spatial coverage of a known-to-be hetereogeneous area.
rlims_plot = rlims.rlims_plot.sort_index(ascending=True)

# %matplotlib inline
plt.figure(figsize=(6.5,4))
ax = plt.subplot(111)

regions_store = {}
for region in gis.regions_boxes:
    
    bmin, bmax = gis.regions_boxes[region]
    boxes_in_region = np.arange(bmin, bmax+1)
    
    # Prepare region data
    region_all = rlims_plot.filter(items=boxes_in_region, axis=0)
    region_all_mean = region_all.mean() #.rolling(5, min_periods=3).mean()
    region_all_mean.name = region
    region_all_mean.index = pd.date_range(start=str(region_all_mean.index.min()), end=str(region_all_mean.index.max()+1), freq='1A-AUG')
    
    # Region MAD ...
    # Calculate this as the annual median of the cluster MADs associated with all
    # slice values underlying the region estimate.
    for_mad = rlims.xytpd[(rlims.xytpd.box_id >= bmin) & (rlims.xytpd.box_id <= bmax)]
    for_mad = for_mad.groupby('year').cluster_mad.median()
    for_mad.name = 'mad'
    for_mad.index = [dt.datetime(int(y),8,31) for y in for_mad.index]
    
    combo = pd.merge(region_all_mean, for_mad, left_index=True, right_index=True, how='left')

    regions_store[region] = region_all_mean
    
    # Plot data
    plt.fill_between(combo.index, combo[region]-combo['mad'], combo[region]+combo['mad'],
             alpha=0.22, color=regions_colors[region]) 
    plt.plot(region_all_mean.index, region_all_mean, 
             linewidth=1, alpha=1, color=regions_colors[region]) 
    
    
    # Regress and plot results
    use_to_fit = combo.dropna()
    use_to_fit.loc[:,'year'] = use_to_fit.index.year
    #res = sm.OLS(use_to_fit, sm.add_constant(np.arange(0, len(use_to_fit)))).fit()
    
    res = sm.WLS(use_to_fit[region], 
                 sm.add_constant(use_to_fit['year']),
                 weights=use_to_fit['mad']).fit()
    print(res.summary())

    fitted = res.params['year'] * np.arange(1985, 2021, 1) + res.params['const']
    fitted = pd.Series(fitted, index=pd.date_range(start='1985-01-01', end='2020-01-01', freq='1AS'))
    #fitted.plot(color=regions_colors[region], linestyle='--', label='_{r}'.format(r=region))
    plt.plot(fitted.index, fitted, linestyle='--', label='_{r}'.format(r=region), color=regions_colors[region])
    
    plt.text(fitted.index[-1], fitted.iloc[-1], 
             r'  %s: %.1f $m~a^{-1}$, R$^2$ %.2f' %(region, res.params['year'], res.rsquared), 
             color=regions_colors[region], va='center')
    
sns.despine()  
plt.ylim(800, 2100)
plt.yticks(np.arange(800, 2400,400), np.arange(800, 2400, 400))
ax.tick_params(which='minor', bottom=False)

from matplotlib import dates
years = dates.YearLocator(5) 
ax.xaxis.set_major_locator(years)
ax.xaxis.set_major_formatter(dates.DateFormatter('%Y'))

plt.savefig(os.path.join(os.environ['FIGS_OUT'],'main/fig1_regions_ts.pdf'))

regions_ts = pd.DataFrame.from_dict(regions_store, orient='index')


# %% trusted=true
regions_ts

# %% trusted=true
regions_ts.to_csv(os.path.join(os.environ['FIGS_OUT'],'main/fig1d_data.csv'))

# %% trusted=true
fitted

# %% trusted=true
use_to_fit

# %% [markdown]
# ## Considering box-by-box stats

# %% [markdown]
# need some criteria of minimum point density required for plotting to be "allowed" or trusted with some symbol.
# - minimum number of slices returning estimates?
# - what information is critical at sub-slice level?
# - precise temporal data is provided by xytpd.csv : 1 km runoff limit locations, with date of each choice, and elevation. In a sense, annual_slice_values are just a more simply rendered version of this file.

# %% [markdown]
# beyond the 'number of slices' criterion applied above, what about the timing of the each year's maximum runoff limit?
# This will require xytpd

# %% [markdown]
# ## Stats for manuscript

# %% [markdown]
# ### Overall change in elevation

# %% trusted=true
early = rlims.rlims_plot.loc[:,1985:1992].mean().mean()
late = rlims.rlims_plot.loc[:,2013:2020].mean().mean()
print(late-early)

# %% trusted=true
means = regions_ts.loc[:,'2013':'2020'].mean(axis=1) - regions_ts.loc[:,'1985':'1992'].mean(axis=1)
print(means)

# %% trusted=true
print('West:', means.loc['S':'NNW'].mean())
print('North:', means.loc['N':'NE'].mean())


# %% trusted=true

# %% [markdown]
# ## Old stuff

# %% trusted=true
## Supplementary figure: standard deviation of each 100 km box

plt.figure(figsize=(8,11))
ax = plt.subplot(111)

items = gis.filter_regions_boxes()
    
stddev_plot = stddev.filter(items=items.index, axis='index')
stddev_plot.columns = [int(float(y)) for y in stddev_plot.columns]
stddev_plot.loc[:,1996] = np.nan
stddev_plot.loc[32:28,2001:2002] = np.nan
stddev_plot.loc[8:1, :1991] = np.nan
stddev_plot = stddev_plot.reindex(sorted(stddev_plot.columns), axis=1)
stddev_labels = rlims_labels.filter(items=items.index, axis='index').sort_index(ascending=False)
stddev_plot = stddev_plot.sort_index(ascending=False)

sns.heatmap(stddev_plot, yticklabels=stddev_labels.loc[:, 'nice_label'], 
                 cmap='flare_r',
                square=True, 
                xticklabels=xticks,
                cbar_kws={'shrink':0.2, 'aspect':5, 'pad':0.01, 'label':r'$\sigma~\grave\upsilon$ (m asl)'})

finish_formatting(ax)
#hatch_bad(box_slice_counts, gis.impossible_boxes, thresh)

plt.tick_params(axis='x', which='both', 
            labelbottom=True, bottom=True)

plt.savefig(os.path.join(os.environ['FIGS_OUT'],'suppl/sf_rlim_stddev.pdf'))

# %%
