# %% [markdown]
# # Figure 3: relationships between excess melting and runoff limit elevation

# %% trusted=true
# #cd ..

# %% trusted=true
import statsmodels.api as sm

# %% [markdown]
"""
plot_rlim_vs_excess_melt.py

Script examining relationship between runoff limit elevation and the amount
of excess melt. Excess melt can be considered at regional scale, for 
individual boxes, only within the runoff area which expanded between 85-92 and
13-20, ....

This script replaces analyse_excess_melt.py, following corruption/failed saving
of the associated Jupyter notebook on 15 December 2020.

It is however still written with the expectation of being run in a Notebook 
environment.
"""

# %% trusted=true
from scipy.stats import ttest_ind
from scipy.stats import normaltest, mannwhitneyu

from load_env import *
gis = gisloader()
rlims = rlimsloader()
em = emloader(gis)

## Define the excess melt area that we are interested in as the runoff area
#  that expanded between 85-92 and 13-20.
ra8592 = polygon_to_mar_mask(gis.racmo_ds,
                polygon_file=os.path.join(os.environ['ANALYSIS_OUT'], 
                    'runoff_limit_polys/poly_{y}.shp'.format(y='1985_1992_median_edited')),
                nc=True)
ra1320 = polygon_to_mar_mask(gis.racmo_ds,
                polygon_file=os.path.join(os.environ['ANALYSIS_OUT'], 
                    'runoff_limit_polys/poly_{y}.shp'.format(y='2013_2020_median_edited')),
                nc=True)
new_area = ra8592.astype(int) - ra1320.astype(int)



# %% trusted=true
cluster = MyCluster()
cluster.scale(jobs=4)
client = Client(cluster)

# %% trusted=true
plt.figure()
new_area.plot()

# %% trusted=true
############
preprep = em.excess_melt.M_e.squeeze().where(new_area > 0)
regions_em = {}
for region in gis.regions_boxes:
    region_em = preprep.where(gis.regions_as_racmo_masks.sel(region=region)) \
        .mean(dim=('x','y')).to_pandas()
    region_em.name = region
    regions_em[region] = region_em
regions_em = pd.DataFrame.from_dict(regions_em, orient='columns')
################



# %% trusted=true
# %matplotlib widget
gis.racmo_ds.Promicemask.plot(cmap='Greys_r')
preprep.sel(time='2019').where(gis.regions_as_racmo_masks.sel(region='SW')).plot()

# %% trusted=true
preprep.sel(time='2019').where(gis.regions_as_racmo_masks.sel(region='SW')).mean()

# %% trusted=true
em.excess_melt.M_e.sel(time='2019').where(gis.regions_as_racmo_masks.sel(region='SW')).mean()

# %% trusted=true
regions_em

# %% trusted=true
regions_em_elev = {}
for region in gis.regions_boxes:
    region_em = preprep.where(gis.regions_as_racmo_masks.sel(region=region)) \
        .stack(dim=('x', 'y')).to_pandas()
    region_em.name = region
    regions_em_elev[region] = region_em
#regions_em = pd.DataFrame.from_dict(regions_em, orient='columns')



# %% trusted=true
region_em.mean()

# %% trusted=true
# %matplotlib widget
regions_em.rolling(10).mean().plot()


# %% trusted=true
#################
def lighten_color(color, amount=0.5):
    """
    Lightens the given color by multiplying (1-luminosity) by the given amount.
    Input can be matplotlib color string, hex string, or RGB tuple.

    Examples:
    >> lighten_color('g', 0.3)
    >> lighten_color('#F034A3', 0.6)
    >> lighten_color((.3,.55,.1), 0.5)
    """
    import matplotlib.colors as mc
    import colorsys
    try:
        c = mc.cnames[color]
    except:
        c = color
    c = colorsys.rgb_to_hls(*mc.to_rgb(c))
    return colorsys.hls_to_rgb(c[0], 1 - amount * (1 - c[1]), c[2])


# %% [markdown]
# ### Breakpoints using T-tests

# %% trusted=true
breaks = []
for region in gis.regions_boxes:
    
    # Prepare region data
    region_all = rlims.rlims_plot.filter(items=gis.regions_boxes[region], axis=0)
    region_all_mean = region_all.mean()
    region_all_mean.name = region
    region_all_mean.index = pd.date_range(start=str(region_all_mean.index.min()), end=str(region_all_mean.index.max()+1), freq='1A-AUG')
    
    # Try each year as the breakpoint in turn
    for year in np.arange(1995,2010,1):
    
        data = region_all_mean
        data = data[data.notna()]
        data.name = 'rlim'

        data = data.to_frame()
        data['pop'] = np.where(data.index >= dt.datetime(year,1,1), 1, 0)
        stat, p = mannwhitneyu(data.rlim[data['pop'] == 0], data.rlim[data['pop'] == 1], alternative='two-sided')
        tstat, tp = ttest_ind(data.rlim[data['pop'] == 0], data.rlim[data['pop'] == 1])
        
        # T-test only works for sample sizes >= 8.
        ssize1 = len(data.rlim[data['pop'] == 0].dropna())
        ssize2 = len(data.rlim[data['pop'] == 1].dropna())
        if (ssize1 < 8) or (ssize2 < 8):
            nstat1 = np.nan
            np1 = np.nan
            nstat2 = np.nan
            np2 = np.nan
        else:
            nstat1, np1 = normaltest(data.rlim[data['pop'] == 0])
            nstat2, np2 = normaltest(data.rlim[data['pop'] == 1])
            
        breaks.append(dict(stat=stat, p=p, region=region, year=year, tstat=np.abs(tstat), tp=tp,
                          nstat1=nstat1, nstat2=nstat2, np1=np1, np2=np2))
        
bpd = pd.DataFrame(breaks)
bpd.index = pd.MultiIndex.from_frame(bpd.filter(items=('region', 'year'), axis=1))

# %% trusted=true
# Define the break points as the years with the maximum t-test statistic.
breaks = bpd.groupby(bpd.region).tstat.idxmax()

# %% trusted=true
# Plot the normality and T-test statistics for each break year
# %matplotlib inline
plt.figure(figsize=(3.5,7))

from matplotlib.lines import Line2D
legend_elements = [Line2D([0], [0], marker='o', mec='black', label='$p > 0.05$',
                          mfc='black', markersize=5, color='none'),
                   Line2D([0], [0], marker='o', mec='black', label='$p < 0.05$',
                          mfc='none', markersize=5, color='none')]

# Plot the normality test results
ax1 = plt.subplot(311)
bpd['n1_sig'] = bpd.np1 > 0.05
for region in gis.regions_boxes:  
    yr = bpd.loc[region,:].year.astype(int)
    n = bpd.loc[region,:].nstat1
    v = bpd.loc[region,:].n1_sig
    plt.plot(yr, n, '-', color=regions_colors[region], alpha=0.7, zorder=100, label=region)
    plt.plot(yr[v], n[v], 'o', color=regions_colors[region], markersize=5, alpha=0.7, zorder=110)
    plt.plot(yr[~v], n[~v], 'o', mfc='none', color=regions_colors[region], markersize=5, alpha=0.7, zorder=110)

plt.ylabel('K$^2$ early')
plt.ylim(-1,20)
plt.yticks(np.arange(0,24,4))
ax1.legend(handles=legend_elements, bbox_to_anchor=(1.35,1), frameon=False)
ax1.annotate('a', fontweight='bold', xy=(0.04,0.93), xycoords='axes fraction',
           horizontalalignment='left', verticalalignment='top')


ax2 = plt.subplot(312)
bpd['n2_sig'] = bpd.np2 > 0.05
for region in gis.regions_boxes:  
    yr = bpd.loc[region,:].year.astype(int)
    n = bpd.loc[region,:].nstat2
    v = bpd.loc[region,:].n2_sig
    plt.plot(yr, n, '-', color=regions_colors[region], alpha=0.7, zorder=100, label=region)
    plt.plot(yr[v], n[v], 'o', color=regions_colors[region], markersize=5, alpha=0.7, zorder=110)
    plt.plot(yr[~v], n[~v], 'o', mfc='none', color=regions_colors[region], markersize=5, alpha=0.7, zorder=110)

plt.ylabel('K$^2$ late')
plt.ylim(-1,20)
plt.yticks(np.arange(0,24,4))
ax2.legend(handles=legend_elements, bbox_to_anchor=(1.35,1), frameon=False)
ax2.annotate('b', fontweight='bold', xy=(0.04,0.93), xycoords='axes fraction',
           horizontalalignment='left', verticalalignment='top')


ax3 = plt.subplot(313)
bpd['t_sig'] = bpd.tp < 0.01
for region in gis.regions_boxes:
    yr = bpd.loc[region,:].year.astype(int)
    t = bpd.loc[region,:].tstat
    v = bpd.loc[region,:].t_sig
    plt.plot(yr, t, '-', color=regions_colors[region], alpha=0.7, zorder=100, label=region)
    plt.plot(yr[v], t[v], 'o', color=regions_colors[region], markersize=5, alpha=0.7, zorder=110)
    plt.plot(yr[~v], t[~v], 'o', mfc='none', color=regions_colors[region], markersize=5, alpha=0.7, zorder=110)
    bkpt = bpd.loc[region,:]
    bkpt = bkpt[bkpt.tp < 0.01].iloc[0] 
    plt.plot(breaks[region][1], t.loc[breaks[region][1]],'P',color=regions_colors[region], markersize=12, zorder=120, alpha=0.9)
    
plt.ylim(0, 10)
plt.ylabel('T-statistic')
plt.xlabel('Break year')
ax3.annotate('c', fontweight='bold', xy=(0.04,0.93), xycoords='axes fraction',
           horizontalalignment='left', verticalalignment='top')

for ax in [ax1, ax2, ax3]:
    ax.set_xlim(1994, 2009)
    yr_ticks = np.arange(1995,2009,2)
    ax.set_xticks(yr_ticks)

handles = plt.legend(bbox_to_anchor=(1,1), frameon=False).legendHandles
handles.append(Line2D([0], [0], marker='o', mec='black', label='$p < 0.01$',
                          mfc='black', markersize=5, color='none'))
handles.append(Line2D([0], [0], marker='o', mec='black', label='$p > 0.01$',
                          mfc='none', markersize=5, color='none'))
handles.append(Line2D([0], [0], marker='P', mec='black', label='Chosen break',
                          mfc='black', markersize=12, color='none'))
plt.legend(handles=handles, bbox_to_anchor=(1.01,1.4), frameon=False)


sns.despine()
plt.savefig(os.path.join(os.environ['FIGS_OUT'],'main/final/sf_breakpoints.pdf'), bbox_inches='tight')

# %% [markdown]
# ### Create main scatter plot with regressions for early and late periods

# %% trusted=true
avail = get_region_availability(gis, rlims, slices_thresh=0.2)

# %matplotlib widget
fig = plt.figure(figsize=(7,4.5)) #(7,4.5)
n=2
combo_store = {}

letgen = letters(start=0)

# Plot breakpoint years on their own axis
ax = plt.subplot(3,4,1)
nn = 1

years = np.arange(1995,2011,1).tolist()
ones = [1]*len(years)
yst = dict(zip(years,ones))
for region in gis.regions_boxes:
    y = breaks[region][1]
    plt.plot(yst[y], y, 'd', color=regions_colors[region], markersize=4)
    yst[y] += 1
    nn += 1
sns.despine(ax=ax, left=True, top=True, bottom=True, right=False)
ax.tick_params(bottom=False, labelbottom=False, top=False, labelright=True, labelleft=False, left=False)
ax.tick_params(axis='y', direction='inout', right=True)

plt.xlim(-0.5, 5)
new_xlims = list(ax.get_xlim()[::-1])
ax.set_xlim(new_xlims)

plt.ylim(1995, 2010)
ax.set_yticks([1995,2000,2005,2010])

pos = ax.get_position()
print(pos) #FIGURE space!
pos.y0 = pos.y0 + 0.03
pos.y1 = pos.y1 - 0.06
pos.x0 = pos.x0 - 0.04
pos.x1 = pos.x1 - 0.12
ax.set_position(pos)

ax.annotate('Break year', xy=(1.7,0.5), xycoords='axes fraction',
           verticalalignment='center', rotation=90)

ax.annotate(next(letgen), fontweight='bold', xy=(0.55,1.35), xycoords='axes fraction',
           horizontalalignment='left', verticalalignment='top')

store_combos = {}
for region in gis.regions_boxes:
    
    # Runoff limit data for the region
    region_all = rlims.rlims_plot.filter(items=gis.regions_boxes[region], axis=0)
    region_all_mean = region_all.mean() 
    region_all_mean.name = 'rlim'
    region_all_mean.index = pd.date_range(start=str(region_all_mean.index.min()), end=str(region_all_mean.index.max()+1), freq='1A-AUG')
    region_all_mean = region_all_mean.where(avail[region].notna())

    # Excess melting
    data_raw = regions_em[region] #.where(avail[region].notna())
    data_raw.name = 'em1y'
    data = data_raw.rolling(10, min_periods=3).mean().squeeze() #min_periods=5
    data.name = 'emwy'

    # Combine the datasets and split time series based on breakpoint found above
    combo = pd.concat((region_all_mean, data_raw, data), axis=1)
    combo['rlim'] = combo['rlim'] - combo['rlim'].mean()
    combo['year'] = combo.index.year
    combo['pop'] = np.where(combo.year >= breaks[region][1], 1, 0) 
    store_combos[region] = combo
    
    # Regression model for early period
    early = combo[combo['pop'] == 0]
    early = early.dropna(subset=('em1y', 'rlim'))
    X = early.em1y
    y = early.rlim
    X = sm.add_constant(X)
    fit_early = sm.OLS(y, X).fit()

    # Regression model for late period
    late = combo[combo['pop'] == 1] 
    late = late.dropna(subset=('em1y', 'rlim'))
    X = late.em1y
    y = late.rlim
    X = sm.add_constant(X)
    fit_late = sm.OLS(y, X).fit()    
    
    # Plotting
    ax = plt.subplot(3,4,n)
    colors = [lighten_color(regions_colors[region], amount=0.5),regions_colors[region]]
    mewy_sizes = (10,80)
    mewy_size_norm = (200, 1200)
    ret = sns.scatterplot(data=combo, x='em1y', y='rlim', marker='o',
                    size='emwy', sizes=mewy_sizes, size_norm=mewy_size_norm, hue='pop', palette=colors,
                     ax=ax, alpha=0.85, legend=False) 

    xx_early = np.arange(combo.em1y[combo['pop'] == 0].min(), combo.em1y[combo['pop'] == 0].max(), 10)
    yy_early = fit_early.params.em1y * xx_early + fit_early.params.const
    plt.plot(xx_early,yy_early, color='darkgray', linestyle='-')
    
    
    xx_late = np.arange(combo.em1y[combo['pop'] == 1].min(), combo.em1y[combo['pop'] == 1].max(), 10)
    yy_late = fit_late.params.em1y * xx_late + fit_late.params.const
    plt.plot(xx_late,yy_late, color='black', linestyle='-')
    
    # Display R2 values, with p-value denoted by text style
#     if region == 'S':
#         xy_early = (0.3, 0.2)
#     else:
    xy_early = (0.93,0.2)
    xy_late = (xy_early[0], xy_early[1]+0.15)
    xy_r2 = (xy_early[0]-0.25, xy_early[1]+0.12)
    
    if fit_early.pvalues[1] < 0.05:
        style='normal'
    else:
        style='italic'
    ax.annotate('{r}'.format(r=np.round(fit_early.rsquared,2)), xy=xy_early, xycoords='axes fraction',
           horizontalalignment='right', verticalalignment='top', color='gray', fontstyle=style)
    
    if fit_late.pvalues[1] < 0.05:
        style='normal'
    else:
        style='italic'
    ax.annotate('{r}'.format(r=np.round(fit_late.rsquared,2)), xy=xy_late, xycoords='axes fraction',
           horizontalalignment='right', verticalalignment='top', color='black', fontstyle=style)
    
    ax.annotate(r'$R^2$', xy=xy_r2, xycoords='axes fraction',
           horizontalalignment='right', verticalalignment='top', color='black', fontstyle=style)
    
    plt.xlabel('')
    plt.ylabel('')
    
    plt.xlim(0, 2000)
    plt.ylim(-400, 300)
    
    # Add subplot labels - letters and regions
    ax.annotate(next(letgen), fontweight='bold', xy=(0.08,0.95), xycoords='axes fraction',
           horizontalalignment='center', verticalalignment='top')
    ax.annotate(region, xy=(0.15, 0.95), xycoords='axes fraction', #(0.93,0.17)
           horizontalalignment='left', verticalalignment='top', color=regions_colors[region])
    
    # Remove duplicated x and y axes labels
    if n not in (5,):
        ax.tick_params(labelleft=False)
    else:
        plt.ylabel(r'$\Delta$ runoff limit (m)')
        
    if n not in (5,6,7,8):
        ax.tick_params(labelbottom=False)
        
    
    combo_store[region] = combo
    n += 1

    sns.despine(ax=ax)




## Main axis labels
fig.text(0.5, 0.3, '$M_e$ (mm w.e. y$^{-1}$)', ha='center')


## Rolling average excess melt legend
em_s = np.arange(mewy_size_norm[0],mewy_size_norm[1],300)
data = pd.DataFrame({'em':em_s, 'y':[1]*len(em_s), 'x':[1]*len(em_s)})    #np.linspace(0,1,len(em_s))[::-1]
# left, bottom, width, height
ax = fig.add_axes((0.75, 0.22, 0.15, 0.07))
ret = sns.scatterplot(data=data, x='em', y='y', marker='o',
                    size='em', sizes=mewy_sizes, size_norm=mewy_size_norm, 
                    alpha=0.9, legend=False, ax=ax, color='gray')
new_xlims = list(ax.get_xlim()) #[::-1]
new_xlims[0] = new_xlims[0] + 100
ax.set_xlim(new_xlims)
ax.set_xticks(em_s)
ax.tick_params(labeltop=False, pad=-4)
ax.tick_params(left=False, labelleft=False)
sns.despine(ax=ax, top=True, bottom=True, right=True, left=True)
ax.set_ylabel('')
ax.set_xlabel('')
ax.set_ylim(0.85,1.11)
ax.annotate('$M_e 10y$ (mm w.e. y$^{-1}$)', xy=(0.5,-1.2), xycoords='axes fraction',
           verticalalignment='bottom', rotation=0, horizontalalignment='center')

plt.savefig(os.path.join(os.environ['FIGS_OUT'],'main/final/fig2_rlim_vs_excess_melt.pdf'), dpi=300)

# %% trusted=true

# %% trusted=true
fit_early.summary()

# %% trusted=true
## Excess melt statistics: change between early and late periods
print('region, 85-92 (mm we), 13-20 (mm we), change (mm we), %')
percs = []
diffs = []
for region in gis.regions_boxes:
    r = combo_store[region]
    me_early = r['em1y'].loc['1985':'1992'].mean()
    me_late = r['em1y'].loc['2013':'2020'].mean()    
    perc = (100 / me_early * me_late) - 100
    diff = np.round(me_late-me_early)
    percs.append(perc)
    diffs.append(diff)
    print(region, np.round(me_early,1), np.round(me_late,1), diff, np.round(perc,0))

# %% trusted=true
100 / 244 * 29

# %% trusted=true
np.mean(diffs)

# %% trusted=true
for region in gis.regions_boxes:
    print(region)
    print(combo_store[region].emwy.min())
    print(combo_store[region].emwy.max())

# %% [markdown]
# ## Data exports

# %% trusted=true
with pd.ExcelWriter(os.path.join(os.environ['FIGS_OUT'],'main/final/fig2_rlim_vs_excess_melt.xlsx')) as writer:  
    for region,df in store_combos.items():
        df.to_excel(writer, sheet_name=region)

# %%
