# %% [markdown]
# # Plot runoff areas and quantities, region by region
#
# Calculating annual GrIS-wide runoff area is not appropriate because in most years at least one region is missing observations. Hence, we can only look at annual changes in runoff area by region. However, we can do ice-sheet-wide area comparisons between the two common periods.

# %% [markdown]
# ## Basic pre-requisites

# %% trusted=true
# cd ..

# %% trusted=true
from load_env import *
do_checks = True

import warnings
warnings.simplefilter('ignore')

# %% trusted=true
gis = gisloader()
rlims = rlimsloader()

# %% trusted=true
#cluster = SGECluster(queue='huge.q@node56', interface='ib0', cores=1, memory='16GB', processes=1, job_extra=['-pe smp 1'], python='ionice -c 3 /home/tedstona/miniconda3/envs/geospa/bin/python')
cluster = MyCluster()
#cluster.scale(jobs=4)
client = Client(cluster)

# %% trusted=true
cluster.scale(jobs=4)

# %% trusted=true
client

# %% trusted=true
#cluster.close()

# %% trusted=true
annual_rlims = rlims.get_annual_polys()

# %% trusted=true
# Simplify the GrIS outline a bit
gris_outline_simpl = gis.gris_outline
gris_outline_simpl['geometry'] = gris_outline_simpl.simplify(1000)

# %% trusted=true
regions_available = get_region_availability(gis, rlims)

# %% [markdown]
# ## RACMO pre-requisites

# %% trusted=true
racmo_test = xr.open_dataset('/flash/tedstona/RACMO/1km/runoff/runoff.2000_AMJ.BN_RACMO2.3p2_ERA5_3h_FGRN055.1km.DD.nc')

# %% trusted=true
racmo_test

# %% trusted=true
regenerate_racmo = False

# %% trusted=true
racmo_runoff_pth = os.path.join(os.environ['ANALYSIS_OUT'], 'racmo_interim', 
                                'runoff.BN_RACMO2.3p2_ERA5_FGRN055.1km.YEARLY_SUM.nc')

# %% trusted=true
if regenerate_racmo:
    #chunks={'x':1496/2,'y':2700/2, 'time':366}
    ru_all = xr.open_mfdataset(os.path.join(gis.RACMO_DS_ROOT, '1km', 'runoff', '*.nc'), parallel=True).runoffcorr
    ru_base = ( ru_all.resample(time='1AS').sum(dim='time') )
    ru_base.to_netcdf(racmo_runoff_pth)

# %% trusted=true
racmo_ru_base = xr.open_dataarray(racmo_runoff_pth, chunks={'time':1, 'y':2700, 'x':1496}) \
    .sel(time=slice('1985', '2020'))

# %% trusted=true
racmo_ru_base = racmo_ru_base.where(gis.racmo_ds.Promicemask == 3)


# %% trusted=true
def pstere_lat2k0(lat):
    """ Given lat in degrees, return k0 

    Implementation of eq. 1 in:
    Rollins, 2011. Computation of scale-factor and standard parallel for the 
    polar stereographic projection.
    URL: http://earth-info.nga.mil/GandG/coordsys/polar_stereographic/Polar_Stereo_phi1_from_k0_memo.pdf
    Last retrieved 2012-Dec-07.

    """

    ecc = 0.0818191908426215
    lat = np.deg2rad(lat)
    k90 = np.sqrt(np.power((1. + ecc), (1. + ecc)) * np.power((1. - ecc), (1. - ecc)))
    k0_1 = ((1. + np.sin(lat)) / 2.) 
    k0_2 = (k90 / np.sqrt(np.power(1. + (ecc * np.sin(lat)), (1.+ecc)) * np.power(1. - (ecc * np.sin(lat)), (1.-ecc))))
    k0 = k0_1 * k0_2

    return k0

## Calculate scaling factors for grid (after Bamber Interpolate_racmo.BAK)
# Calculate lat/lon coordinates for each point of projected grid
# Convert to paired columns
grid_latlon = np.vstack((np.ravel(gis.racmo_ds.LAT), np.ravel(gis.racmo_ds.LON))).T
# Convert to radians
grid_latlon_radians = grid_latlon / 57.29578
# Calculate latitudinal scaling
m = 2.0 * np.tan(45.0 / 57.29578 - (grid_latlon_radians[:, 0] / 2)) / np.cos(grid_latlon_radians[:, 0]) 
# Compute scale factor for each grid cell
# lat_0 is 70 according to Brice email 2021-08-27.
k0 = pstere_lat2k0(70)
print(k0)
racmo_scale_factors = np.power(m * k0, 2) #np.sqrt((m * k0))
# reshape scale_factors to the same dimensions as the grid
racmo_scale_factors = racmo_scale_factors.reshape(gis.racmo_ds.LAT.shape)

# %% [markdown]
# ## MAR pre-requisites

# %% trusted=true
# MAR cell areas as calculated by Xavier
mar_cell_area = xr.open_dataset(os.path.join(os.environ['L0lib'], 'MAR-GRa.nc2')).AREA
mar_cell_area = mar_cell_area.rename({'X10_153':'x', 'Y21_288':'y'})
mar_cell_area['x'] = mar_cell_area['x'] * 1000
mar_cell_area['y'] = mar_cell_area['y'] * 1000

# %% trusted=true
mar_ru_base = marutils.open_dataset(os.path.join(gis.MAR_DS_ROOT, '*.nc'), transform_func=lambda x:x.RU) \
    .sel(SECTOR=1) \
    .sel(time=slice('1985', '2020'))
mar_ru_base = mar_ru_base * gis.gris_mask
mar_ru_base = mar_ru_base.resample(time='1AS').sum(dim='time').compute()

# %% [markdown]
# ## Optional checking steps - can be skipped

# %% trusted=true
# Optional: also load the common period polygons. This is useful if wanting to use the optional
# plotting cell, below.
load_multiyear = False

if load_multiyear:
    c8592 = gpd.read_file(os.path.join(os.environ['ANALYSIS_OUT'], 'runoff_limit_polys/poly_{y}.shp'.format(y='1985_1992_median')))
    c8592['year'] = '1985_1992'
    c1320 = gpd.read_file(os.path.join(os.environ['ANALYSIS_OUT'], 'runoff_limit_polys/poly_{y}.shp'.format(y='2013_2020_median')))
    c1320['year'] = '2013_2020'
    #annual_rlims = pd.concat((annual_rlims, c8592), axis=0)
    #annual_rlims = pd.concat((annual_rlims, c1320), axis=0)

# %% trusted=true
# Optional: plots of all polygons loaded into dataframe, basically to check data density underlying the polygons.
plot_rlims = False

if plot_rlims:
    for ix,row in annual_rlims.iterrows():
        plt.figure(figsize=(8,13))
        ax = plt.subplot(111)
        try:
            if '_' in str(row.year):
                file_end = '%s_median' %row.year
            else:
                file_end = row.year
            csv = pd.read_csv(os.path.join(os.environ['ANALYSIS_OUT'], 'runoff_limit_polys/poly_%s.csv' %file_end))
        except IOError:
            print('IOError, skipping %s' %row.year)
            continue
        original = csv[csv.status == 0]
        interpolated = csv[csv.status == 1]
        gis.gris_outline.plot(ax=ax, alpha=0.2)
        plt.plot(original.x, original.y, '+', alpha=0.6, label='Data')
        plt.plot(interpolated.x, interpolated.y, 'o', label='Interpolated', mfc='none')
        #plt.plot(original['xs'], original['ys'], '^', label='Smoothed', mfc='none')
        plt.plot(*row.geometry[0].exterior.xy, color='black', alpha=0.6, label='poly_smoothed', linewidth=2)
        plt.savefig(os.path.join(os.environ['ANALYSIS_OUT'], 'runoff_limit_polys/poly_%s.png' %row.year), dpi=300)
        plt.title(row.year)
        plt.legend()
        plt.box(on=False)

        plt.close()

# %% [markdown]
# ## Calculate OBSERVED regional annual runoff areas

# %% trusted=true
# Do region-by-region area calculations
regions = {}
results = {}
for region in gis.regions_boxes:
    print(region)
    just_region = gis.boxes[gis.boxes.region == region]
    just_region.loc[:,'RID'] = 0
    just_region.loc[:,'geometry'] = just_region.buffer(500)
    just_region = just_region.dissolve(by='RID')
    gris_region = gpd.overlay(just_region, gris_outline_simpl, how='intersection')
    regions[region] = gris_region
    
    annual_polys = gris_region['geometry'].difference(annual_rlims)
    annual_areas = annual_polys.area
    annual_areas.index = [dt.datetime(y,1,1) for y in annual_rlims['year']]
    results[region] = annual_areas
    
res = pd.DataFrame.from_dict(results)

# %% [markdown]
# ## Comparison with RCM runoff area

# %% trusted=true
min_RU = 10 

# %% [markdown]
# ### MAR

# %% trusted=true
# Load in CSV which tells us about which regions are good in which years
comparison = pd.read_csv(os.path.join(os.environ['ANALYSIS_OUT'], 'compare_mar_landsat', 'landsat_mar_rlim_comparison.csv'), skiprows=4, index_col=0)
comparison.index = [dt.datetime(y,1,1) for y in comparison.index]

# %% trusted=true
comparison

# %% trusted=true
res_km = res / (1000**2)


# %% trusted=true
def compute_area_mar(threshold):
    regions_area = {}
    for region in gis.regions_boxes:

        ru_base = ( mar_ru_base * gis.gris_mask ) \
            .where(gis.regions_as_mar_masks.sel(region=region) == True) # \

        region_area = (mar_cell_area.where(ru_base > threshold).sum(dim=('x','y')) ).to_pandas()
        regions_area[region] = region_area
    return regions_area


# %% trusted=true
reg_ru_area = pd.DataFrame.from_dict(compute_area_mar(min_RU), orient='index').T


# %% [markdown]
# ### RACMO

# %% trusted=true
def compute_area_racmo(threshold):
    regions_area = {}
    for region in gis.regions_boxes:

        ru_base_reg = racmo_ru_base.where(gis.regions_as_racmo_masks.sel(region=region) == True)                       

        # At the moment, no cell weights to adjust for latitudinal scaling here.
        region_area = (ru_base_reg.where(ru_base_reg > threshold).count(dim=('x','y')) ).to_pandas()
        regions_area[region] = region_area                           
    return regions_area


# %% trusted=true
racmo_reg_ru_area = pd.DataFrame.from_dict(compute_area_racmo(min_RU), orient='index').T

# %% trusted=true
# %matplotlib widget
gis.regions_as_racmo_masks.sel(region='SW').plot()

# %% trusted=true
# %matplotlib widget
a19 = ru_base_reg.sel(time='2018')
a19.where(a19 > min_RU).count()
#a19.count()

# %% trusted=true
# %matplotlib widget
racmo_reg_ru_area.plot()

# %% trusted=true
# %matplotlib widget
racmo_reg_ru_area.plot()

# %% trusted=true
# %matplotlib widget
(racmo_reg_ru_area - reg_ru_area).plot()

# %% trusted=true
# %matplotlib widget
((100 / reg_ru_area) * racmo_reg_ru_area).plot()

# %% trusted=true
# %matplotlib widget
for region in gis.regions_boxes:
    plt.figure(figsize=(5,2.5))
    (100 / racmo_reg_ru_area[region] * res_km[region]).where(comparison['r_%s'%region] != 'N').plot(linestyle='none', marker='o', color=regions_colors[region])
    plt.title(region)
    plt.grid()
    plt.ylim(40, 120)
# As a % of average 1985-2020 runoff area size? Not sure this makes sense.
# Do need to normalise against something though, in order to be able to interpret whether these numbers matter or not
# Even though the absolute values should tell us this, they are difficult to interpret.
# Against maximum?


# %% [markdown]
# 1 kg == 1 L == 1000 g == 1000 ml == 1g of water is 1cm3 --> 10 mm w.e. == 1cm, 
#
# Let's say we care only about MAR runoff > 10 mm w.e. per annum.
# A 10 km grid box generating 10 mm w.e. per annum yields how many gigatonnes of runoff?
#
# Simplest to first calculate in kilograms.
# 10 km2 == 10,000 m2.
# Kg of melt water per 1 m2: 10 mm * (1 m * 1000) * (1 m * 1000) == 10000000 mm3
# Let's work in metres only: 0.01 * 1 * 1 = 0.01 m3 water.
#
# 1 m3 of water is equivalent to 1000 kg of water, as 
# 1cm3 = 1g
# 1m3 == 100 cm x 100 cm  * 100 cm = 1000000 cm == 1000000 g == 1000 kg.
#
# 0.01 m3 water over 10,000 m2 = 100 m3.
# As kg, 100 m3 * 1000 = 100000 kg, or 100 tonnes
# 1 Gt = 1000000000000 kg (e12), so
# 100000 / 1e12
#
# Basically, not much
#
#
# 1m3 = 1000kg/1T
# 10 mm w.e. over a 1m^2 box: 1 cm. 1x1x0.01 = 0.01 m3. 10 kg.
# 10000 * 10000 * 10 = 1000000000 kg
#

# %% trusted=true
10000 * 10000 * 10

# %% [markdown]
# ### Sensitivity analysis of threshold choice on area

# %% trusted=true
try_thresholds = [1, 5, 10, 20]

# %% trusted=true
test_thresh = 5
thresh_outputs = {}
for test_thresh in try_thresholds:
    results = pd.DataFrame.from_dict(compute_area_mar(test_thresh), orient='index').T
    thresh_outputs[test_thresh] = results.sum(axis=1)
mar_thresh_outputs = pd.DataFrame.from_dict(thresh_outputs, orient='columns')

# %% trusted=true
racmo_thresh_outputs = {}
for test_thresh in try_thresholds:
    results = pd.DataFrame.from_dict(compute_area_racmo(test_thresh), orient='index').T
    racmo_thresh_outputs[test_thresh] = results.sum(axis=1)
racmo_thresh_outputs = pd.DataFrame.from_dict(racmo_thresh_outputs, orient='columns')

# %% trusted=true
mar_thresh_outputs.index = mar_thresh_outputs.index.to_pydatetime()

# %% trusted=true
mar_thresh_outputs.index

# %% trusted=true
# %matplotlib inline 
# widget

line_cols = sns.color_palette("crest", n_colors=4, as_cmap=True)
lines_racmo = sns.cubehelix_palette(as_cmap=True, n_colors=4)
lines_mar = sns.cubehelix_palette(start=.5, rot=-.5, as_cmap=True, n_colors=4)

from matplotlib.dates import date2num
from matplotlib import dates

letgen = letters(start=0)

def _format_common(ax, xstart=dt.datetime(1984,1,1), xend=dt.datetime(2021,1,1), every=5, xy=(0.04,0.93)):
    years = dates.YearLocator(every) 
    ax.xaxis.set_major_locator(years)
    ax.xaxis.set_major_formatter(dates.DateFormatter('%Y'))
    #plt.xlim(date2num(xstart), date2num(xend))
    ax.annotate(next(letgen), fontweight='bold', xy=xy, xycoords='axes fraction',
           horizontalalignment='left', verticalalignment='top')
    
plt.figure(figsize=(5.5,4))

ax1 = plt.subplot(221)
mar_thresh_outputs.plot(ax=ax1, colormap=lines_mar, x_compat=True, legend=False)
ax1.legend(frameon=False, #title='MAR threshold\n (mmWE a$^{-1}$)', 
           labelspacing=0.2, ncol=2, title_fontsize=7,
          loc='lower right', handletextpad=0.5, handlelength=1)
plt.ylim(100000, 300000)
plt.ylabel('Runoff area (km$^2$)')
_format_common(ax1)

ax2 = plt.subplot(222)
racmo_thresh_outputs.plot(ax=ax2, colormap=lines_racmo, x_compat=True)
ax2.legend(frameon=False, #title='RACMO threshold\n(mmWE a$^{-1}$)',
           labelspacing=0.2, ncol=2, title_fontsize=7,
           loc='lower right', handletextpad=0.5, handlelength=1)
plt.ylim(100000, 300000)
_format_common(ax2)

ax3 = plt.subplot(223)
as_perc = mar_thresh_outputs.apply(lambda x: (100 / mar_thresh_outputs.loc[:,1]) * x)
as_perc = as_perc.drop(columns=[1])
as_perc.plot(ax=ax3, colormap=lines_mar, x_compat=True, legend=False)
plt.ylim(85, 102)
plt.ylabel('% of 1 mm WE a$^{-1}$ area')
_format_common(ax3)

ax4 = plt.subplot(224)
as_perc = racmo_thresh_outputs.apply(lambda x: (100 / racmo_thresh_outputs.loc[:,1]) * x)
as_perc = as_perc.drop(columns=[1])
as_perc.plot(ax=ax4, colormap=lines_racmo, x_compat=True, legend=False)
plt.ylim(85,102)
_format_common(ax4)

sns.despine()
sns.despine(ax=ax2, left=True)
ax2.set_yticks([])
sns.despine(ax=ax4, left=True)
ax4.set_yticks([])
plt.subplots_adjust(hspace=0.35, left=0.15)

plt.savefig(os.path.join(os.environ['FIGS_OUT'], 'sf_sub2_model_runoff_areas_thresholds.pdf'))

# %% trusted=true
# %matplotlib widget


# %% trusted=true
# %matplotlib widget


# %% [markdown]
# ## Runoff Volumes

# %% trusted=true
def calc_rcm_runoff(ru_base_all, 
                    rcm_region_msks, 
                    rcm_annual_msks, 
                    rcm_8592_msk, 
                    scale_cell_area):
    """ 
    ru_base_all must be supplied with RCM-specific topo/ice-sheet mask already applied !
    
    IMPORTANT: masking operations must be multiplier, NOT where, because the MAR
    masks are fractional !
    
    """
        
    # Calculate runoff quantities for each region
    regions_ru_vis = {}
    regions_ru_all = {}
    regions_ru_zone_all = {}
    regions_ru_zone_vis = {}
    for region in gis.regions_boxes:
        
        ru_base = ru_base_all.where(rcm_region_msks.sel(region=region))
       
        # MAR: cell_area is provided by Xavier in km3 units.
        #      Convert ru_base from mmWE a-1 to kmWE a-1
        #      Then multiply by cell_area to get runoff volume/flux.
        # RACMO : scaling factor is computed above. For the 1km res this can be applied directly.
        ru_base = (ru_base * 1e-6)  * scale_cell_area

        # All runoff which occurs in region up to that year's VISIBLE runoff limit.
        region_ru_vis_gt = ( ru_base \
            * rcm_annual_msks ) \
            .sum(dim=('x', 'y')) \
            .squeeze() \
            .to_pandas()
        regions_ru_vis[region] = region_ru_vis_gt

        # All runoff in region, ignoring visible runoff limit.
        region_ru_all_gt = ru_base \
            .sum(dim=('x', 'y')) \
            .squeeze() \
            .to_pandas()
        regions_ru_all[region] = region_ru_all_gt

        # Just runoff which occurs above the 1985-92 limit in any given year, but below VISIBLE
        region_ru_zone_all = ( ru_base \
            * (1-rcm_8592_msk) \
            * rcm_annual_msks ) \
            .sum(dim=('x', 'y')) \
            .squeeze() \
           .to_pandas()
        regions_ru_zone_vis[region] = region_ru_zone_all

        # All runoff in region above 1985-92 limit, ignoring visible runoff limit.
        region_ru_vis_zone = ( ru_base \
            * (1-rcm_8592_msk) ) \
            .sum(dim=('x', 'y')) \
            .squeeze() \
            .to_pandas()
        regions_ru_zone_all[region] = region_ru_vis_zone

    regions_ru_vis_pd = pd.DataFrame.from_dict(regions_ru_vis, orient='columns')
    regions_ru_all_pd = pd.DataFrame.from_dict(regions_ru_all, orient='columns')
    regions_ru_zone_vis_pd = pd.DataFrame.from_dict(regions_ru_zone_vis, orient='columns')
    regions_ru_zone_all_pd = pd.DataFrame.from_dict(regions_ru_zone_all, orient='columns')
    
    return (regions_ru_vis_pd, regions_ru_all_pd, regions_ru_zone_vis_pd, regions_ru_zone_all_pd)


# %% [markdown]
# ### MAR

# %% [markdown]
# At the moment I'm using both the +/- MAD 1985-92 line, AND the +/- MAD annual runoff limits. Is this actually what I want to do?

# %% trusted=true
mar_8592_msk = xr.open_dataarray(os.path.join(os.environ['ANALYSIS_OUT'], 'runoff_limit_polys', 
                                        'poly_1985_1992_median_edited_mar10km_weights.nc'))
annual_msks = xr.open_dataarray(os.path.join(os.environ['ANALYSIS_OUT'], 'runoff_limit_polys', 
                                        'polys_as_mar_rasters_annual_10k.nc'))
mar_ru = calc_rcm_runoff(mar_ru_base, 
                                   gis.regions_as_mar_masks, 
                                   annual_msks, 
                                   mar_8592_msk,
                                   mar_cell_area)

# %% trusted=true
# %matplotlib widget
mar_ru_base.where(gis.regions_as_mar_masks.sel(region='SW')).where(annual_msks.sel(time='2020')).plot()
#mar_ru_base.where(gis.regions_as_mar_masks.sel(region='SW')).where(annual_msks.sel(time='2020')).where(1-mar_8592_msk).plot(cmap='Reds', alpha=0.5)

# %% trusted=true
mar_8592_msk_minus_mad = xr.open_dataarray(os.path.join(os.environ['ANALYSIS_OUT'], 'runoff_limit_polys', 
                                        'poly_1985_1992_median_xytpd_minus_mad_edited_mar_weights.nc'))
annual_msks_minus_mad = xr.open_dataarray(os.path.join(os.environ['ANALYSIS_OUT'], 'runoff_limit_polys', 
                                        'polys_as_mar_rasters_annual_10k_xytpd_minus_mad.nc'))
mar_ru_minus_mad = calc_rcm_runoff(mar_ru_base, 
                                  gis.regions_as_mar_masks, 
                                  annual_msks_minus_mad, 
                                  mar_8592_msk, 
                                  mar_cell_area)

# %% trusted=true
mar_8592_msk_plus_mad = xr.open_dataarray(os.path.join(os.environ['ANALYSIS_OUT'], 'runoff_limit_polys', 
                                        'poly_1985_1992_median_xytpd_plus_mad_edited_mar_weights.nc'))
annual_msks_plus_mad = xr.open_dataarray(os.path.join(os.environ['ANALYSIS_OUT'], 'runoff_limit_polys', 
                                        'polys_as_mar_rasters_annual_10k_xytpd_plus_mad.nc'))
mar_ru_plus_mad = calc_rcm_runoff(mar_ru_base, 
                                  gis.regions_as_mar_masks, 
                                  annual_msks_plus_mad, 
                                  mar_8592_msk, 
                                  mar_cell_area)

# %% trusted=true
# %matplotlib widget
# vis, all, zone_vis, zone_all
mar_ru_minus_mad[0].plot()
mar_ru_plus_mad[0].plot()
(mar_ru_plus_mad[0] - mar_ru_minus_mad[0]).plot()

# %% trusted=true
# Example proof of correct mask functioning
# %matplotlib widget
mar_ru_base.where(1-mar_8592_msk_plus_mad) \
.where(1-annual_msks_plus_mad) \
.where(gis.regions_as_mar_masks.sel(region='SW')) \
.sel(time='2020').plot()


# %% trusted=true
# %matplotlib widget
annual_msks_plus_mad.sel(time=slice('2018','2020')).plot(col='time')

# %% [markdown]
# ### RACMO

# %% trusted=true
racmo_scale_factors

# %% trusted=true
racmo_8592_msk = xr.open_dataarray(os.path.join(os.environ['ANALYSIS_OUT'], 'runoff_limit_polys', 
                                        'poly_1985_1992_median_edited_racmo_weights.nc'))
racmo_annual_msks = xr.open_dataarray(os.path.join(os.environ['ANALYSIS_OUT'], 'runoff_limit_polys', 
                                        'polys_as_racmo_rasters_annual_1k.nc'))
racmo_ru = calc_rcm_runoff(racmo_ru_base, 
                                    gis.regions_as_racmo_masks, 
                                    ~racmo_annual_msks, 
                                    ~racmo_8592_msk,
                                    racmo_scale_factors)

# %% trusted=true
racmo_8592_msk_minus_mad = xr.open_dataarray(os.path.join(os.environ['ANALYSIS_OUT'], 'runoff_limit_polys', 
                                        'poly_1985_1992_median_xytpd_minus_mad_edited_racmo_weights.nc'))
racmo_annual_msks_minus_mad = xr.open_dataarray(os.path.join(os.environ['ANALYSIS_OUT'], 'runoff_limit_polys', 
                                        'polys_as_racmo_rasters_annual_1k_xytpd_minus_mad.nc'))
racmo_ru_minus_mad = calc_rcm_runoff(racmo_ru_base, 
                                    gis.regions_as_racmo_masks, 
                                    ~racmo_annual_msks_minus_mad, 
                                    ~racmo_8592_msk,
                                    racmo_scale_factors)

# %% trusted=true
racmo_8592_msk_plus_mad = xr.open_dataarray(os.path.join(os.environ['ANALYSIS_OUT'], 'runoff_limit_polys', 
                                        'poly_1985_1992_median_xytpd_plus_mad_edited_racmo_weights.nc'))
racmo_annual_msks_plus_mad = xr.open_dataarray(os.path.join(os.environ['ANALYSIS_OUT'], 'runoff_limit_polys', 
                                        'polys_as_racmo_rasters_annual_1k_xytpd_plus_mad.nc'))
racmo_ru_plus_mad = calc_rcm_runoff(racmo_ru_base, 
                                    gis.regions_as_racmo_masks, 
                                    ~racmo_annual_msks_plus_mad, 
                                    ~racmo_8592_msk,
                                   racmo_scale_factors)

# %% trusted=true
# %matplotlib widget
((racmo_ru_plus_mad[0] - racmo_ru_minus_mad[0]) / 1e6).plot()

# %% trusted=true
# Example proof of correct mask functioning
# %matplotlib widget
racmo_ru_base.where(~racmo_8592_msk_minus_mad) \
.where(~racmo_annual_msks_minus_mad) \
.where(gis.regions_as_racmo_masks.sel(region='SW')) \
.sel(time='2020').plot()

# %% trusted=true
racmo_ru

# %% [markdown]
# ## Runoff volume inter-comparison

# %% trusted=true
# Start with a region through time
# Need to scale RACMO data down to km3 of runoff earlier in the process.

# %matplotlib widget
# vis, all, zone_vis, zone_all
plt.vlines(mar_ru_minus_mad[0]['SW'].index, mar_ru_minus_mad[0]['SW'], mar_ru_plus_mad[0]['SW'])
plt.plot(mar_ru[0]['SW'].index, mar_ru[0]['SW'], linestyle=':')
plt.vlines(racmo_ru_minus_mad[0]['SW'].index, racmo_ru_minus_mad[0]['SW']/1e6, racmo_ru_plus_mad[0]['SW']/1e6, colors='red')
plt.plot(racmo_ru[0]['SW'].index, racmo_ru[0]['SW']/1e6, linestyle=':')

# %% trusted=true

## SUPPLEMENTARY FIGURE - region-by-region runoff area and volume
# ---------------------------------------------------------------

# %matplotlib inline
# Visible runoff limit, quantity above the 1985-1992 runoff limit only
from matplotlib import dates

letgen = letters()
col_mar = 'tab:cyan'
col_racmo = 'tab:pink'
lw = 0.5

# See https://stackoverflow.com/questions/45445180/how-to-do-nested-subplots-in-python
from mpl_toolkits.axes_grid1 import axes_grid
nrows = 4
ncols = 2
naxes = 2
f = plt.figure(figsize=(5.5, 7))
for i, region in enumerate(gis.regions_boxes):              # hoz, vert
    ag = axes_grid.Grid(f, (nrows, ncols, i+1), (naxes, 1), axes_pad=(0.1, 0.1))
    
    ag[0].axhline(y=100, linestyle=':', color='k', alpha=0.5)
    mar_area = (100 / res_km[region] * reg_ru_area[region]).where(comparison['r_%s'%region] != 'N')
    ag[0].plot(mar_area.index, mar_area, 
             linestyle='none', marker='s', label='MAR', color=col_mar, markersize=3, alpha=0.7)

    racmo_area = (100 / res_km[region] * racmo_reg_ru_area[region]).where(comparison['r_%s'%region] != 'N')
    ag[0].plot(racmo_area.index, racmo_area, 
             linestyle='none', marker='^', label='RACMO', color=col_racmo, markersize=3, alpha=0.7)

    #ag[0].set_ylabel('% model area of obs. area')
    if i == 6:
        ag[0].legend(frameon=False, ncol=2, loc='center right', bbox_to_anchor=(2.2, 0.4))

    years = dates.YearLocator(5) 
    ag[0].xaxis.set_major_locator(years)
    ag[0].xaxis.set_major_formatter(dates.DateFormatter('%Y'))
    ag[0].xaxis.set_minor_locator(dates.YearLocator())
    ag[0].set_ylim(60, 225)


    ax2 = ag[1]
    # vis, all, zone_vis, zone_all
    
    ax2.plot(mar_ru[3][region].index, mar_ru[3][region], linestyle='-', color='#ccc', 
            label='MAR (all)', linewidth=lw)
    ax2.plot(racmo_ru[3][region].index, racmo_ru[3][region], linestyle='--', color='#ccc',
            label='RACMO (all)', linewidth=lw)

    ax2.vlines(mar_ru_minus_mad[2][region].index, mar_ru_minus_mad[2][region], mar_ru_plus_mad[2][region], 
               linewidth=1, colors='k', zorder=1000)
    ax2.plot(mar_ru[2][region].index, mar_ru[2][region], linestyle='-', color='k', 
             label='MAR (vis.)', linewidth=lw)

    ax2.vlines(racmo_ru_minus_mad[2][region].index, racmo_ru_minus_mad[2][region], 
               racmo_ru_plus_mad[2][region], 
               colors='k', linewidth=1, zorder=1000)
    ax2.plot(racmo_ru[2][region].index, racmo_ru[2][region], 
             linestyle='--', color='k', label='RACMO (vis.)', linewidth=lw)

    #ax2.set_ylabel('R above 1985-92 limit (Gt)')
    
    if i in [0, 2, 4, 6]:
        ag[0].set_ylabel('%')
        ax2.set_ylabel('Gt')
    
    if i == 6:
        ax2.legend(frameon=False, ncol=2, loc='center right', bbox_to_anchor=(2.4, 0.5))

    ag[1].annotate('{r}'.format(r=region), verticalalignment='center', xy=(0.5, 0.8), horizontalalignment='center', xycoords='axes fraction', fontweight='bold', color=regions_colors[region])
    years = dates.YearLocator(10) 
    ax2.xaxis.set_major_locator(years)
    ax2.xaxis.set_major_formatter(dates.DateFormatter('%Y'))
    ax2.xaxis.set_minor_locator(dates.YearLocator(2))

    ax2.set_ylim(0, 20) 

sns.despine()
plt.subplots_adjust(hspace=0.35)

plt.savefig(os.path.join(os.environ['FIGS_OUT'], 'suppl', 'regions_runoff_vols.pdf'))

# %% trusted=true

## MAIN FIGURE - visible-area-wide runoff
# ---------------------------------------

col_mar = 'tab:cyan'
col_racmo = 'tab:pink'

# %matplotlib inline

from matplotlib.dates import date2num
letgen = letters(start=0)

def _format_common(ax, xstart=dt.datetime(1984,1,1), xend=dt.datetime(2021,1,1), every=10, xy=(0.04,0.93)):
    years = dates.YearLocator(every) 
    minor = dates.YearLocator(base=5)
    ax.xaxis.set_major_locator(years)
    ax.xaxis.set_major_formatter(dates.DateFormatter('%Y'))
    ax.xaxis.set_minor_locator(minor)
    plt.xlim(date2num(xstart), date2num(xend))
    ax.annotate(next(letgen), fontweight='bold', xy=xy, xycoords='axes fraction',
           horizontalalignment='left', verticalalignment='top')

from matplotlib import gridspec
gs = gridspec.GridSpec(3, 2, hspace=0.35)
plt.figure(figsize=(3.5,3.5))#(5,4)

ax1 = plt.subplot(gs[0,:])
plt.axhline(y=100, linestyle=':', color='k', alpha=0.5)
mar_area = (100 / res_km.sum(axis=1) * reg_ru_area.sum(axis=1)) #.where(comparison['r_%s'%region] != 'N')
plt.plot(mar_area.index, mar_area, 
             linestyle='none', marker='s', label='MAR', color=col_mar, markersize=3.5)

racmo_area = (100 / res_km.sum(axis=1) * racmo_reg_ru_area.sum(axis=1)) #.where(comparison['r_%s'%region] != 'N')
plt.plot(racmo_area.index, racmo_area, 
             linestyle='none', marker='^', label='RACMO', color=col_racmo, markersize=3.5)

plt.ylim(80,150)
plt.ylabel('% model area\n of obs. area')

#plt.legend(loc='center right', fontsize=7, frameon=False, bbox_to_anchor=(1.3,0.5), handletextpad=0, labelcolor='mfc')
_format_common(ax1)


ax2 = plt.subplot(gs[1,:])
plt.fill_between(mar_ru_minus_mad[2].index, mar_ru_minus_mad[2].sum(axis=1), mar_ru_plus_mad[2].sum(axis=1), 
               linewidth=1, color=col_mar, alpha=0.25)
plt.plot(mar_ru[2].index, mar_ru[2].sum(axis=1), color=col_mar)

plt.fill_between(racmo_ru_minus_mad[2].index, racmo_ru_minus_mad[2].sum(axis=1), 
               racmo_ru_plus_mad[2].sum(axis=1), color=col_racmo, alpha=0.25)
plt.plot(racmo_ru[2].index, racmo_ru[2].sum(axis=1), 
             color=col_racmo, label='RACMO (vis.)')
plt.ylabel('R above\n 1985-92 lim. \n(Gt)')
_format_common(ax2)


ax3 = plt.subplot(gs[2, 0])
ax3.stackplot(mar_ru[2].index, mar_ru[2].cumsum().T, labels=mar_ru[2].columns, colors=[regions_colors[v] for v in regions_colors], linewidth=0.5)
plt.xlim()
plt.ylim(0, 450)
plt.ylabel('Cum. R above \n1985-92 lim. (Gt)')
ax3.annotate('MAR', xy=(0.1,0.5), xycoords='axes fraction',
           horizontalalignment='left', verticalalignment='center', color=col_mar)
_format_common(ax3, xstart=dt.datetime(2000,1,1), every=10, xy=(0.08,0.93))


ax4 = plt.subplot(gs[2, 1])
ax4.stackplot(racmo_ru[2].index, racmo_ru[2].cumsum().T, labels=racmo_ru[2].columns, colors=[regions_colors[v] for v in regions_colors], linewidth=0.5)
#plt.xlim('2000-01-01', '2021-01-01')
#plt.legend(loc='center right', bbox_to_anchor=(1.7,0.35), fontsize=7, frameon=False)
plt.legend(loc='lower center', fontsize=7, frameon=False, ncol=4, columnspacing=1, bbox_to_anchor=(-0.15,-1.))
plt.ylim(0,450)
ax4.annotate('RACMO', xy=(0.1,0.5), xycoords='axes fraction',
           horizontalalignment='left', verticalalignment='center', color=col_racmo)

_format_common(ax4, xstart=dt.datetime(2000,1,1), every=10, xy=(0.08,0.93))


sns.despine()
sns.despine(ax=ax4, left=True)
ax4.set_yticks([])
plt.subplots_adjust(left=0.2, right=1)

plt.savefig(os.path.join(os.environ['FIGS_OUT'], 'sub2_fig2_runoffvolumes.pdf'), bbox_inches='tight', pad_inches=1)

# %% trusted=true
mar_area

# %% trusted=true
# output data

area = pd.concat((mar_area, racmo_area), axis=1)
area.columns = ['MAR', 'RACMO']
labels = ['area', 'runoff_MAR', 'runoff_MAR_minus_1mad', 'runoff_MAR_plus_1mad', 'runoff_RACMO', 'runoff_RACMO_minus_1mad', 'runoff_RACMO_plus_1mad']
with pd.ExcelWriter(os.path.join(os.environ['FIGS_OUT'],'main/final/fig3_modelled_runoff.xlsx')) as writer:  
    for label, df in zip(labels, [area, mar_ru[2], mar_ru_minus_mad[2], mar_ru_plus_mad[2], racmo_ru[2], racmo_ru_minus_mad[2], racmo_ru_plus_mad[2]]):
        print(label)
        df.to_excel(writer, sheet_name=label)

# %% trusted=true
mar_ru

# %% trusted=true
# Small plot for SPI/Horst's talk, October 2021, Lausanne
sns.set_context("poster", font_scale=0.7)
# %matplotlib widget
fig = plt.figure(figsize=(5,6))

letgen = letters()

ax3 = plt.subplot(211)
ax3.stackplot(mar_ru[2].index, mar_ru[2].cumsum().T, labels=mar_ru[2].columns, colors=[col_mar], linewidth=1, edgecolor=col_mar)
plt.xlim()
plt.ylim(0, 450)
fig.text(0.06, 0.5, 'Cumulative runoff above 1985-92 limit (Gt)', rotation=90, horizontalalignment='center', verticalalignment='center')
ax3.annotate('MAR', xy=(0.1,0.5), xycoords='axes fraction',
           horizontalalignment='left', verticalalignment='center', color=col_mar)
_format_common(ax3, xstart=dt.datetime(2000,1,1), every=5, xy=(0.08,0.93))


ax4 = plt.subplot(212)
ax4.stackplot(racmo_ru[2].index, racmo_ru[2].cumsum().T, labels=racmo_ru[2].columns, colors=col_racmo, linewidth=1, edgecolor=col_racmo)
#plt.xlim('2000-01-01', '2021-01-01')
#plt.legend(loc='center right', bbox_to_anchor=(1.7,0.35), fontsize=8, frameon=False)
plt.ylim(0,450)
ax4.annotate('RACMO', xy=(0.1,0.5), xycoords='axes fraction',
           horizontalalignment='left', verticalalignment='center', color=col_racmo)

_format_common(ax4, xstart=dt.datetime(2000,1,1), every=5, xy=(0.08,0.93))

sns.despine()
plt.tight_layout()
plt.subplots_adjust(left=0.23)

plt.savefig(os.path.join(os.environ['FIGS_OUT'], 'fig2_cumulative_only.png'), dpi=200)

# %% trusted=true
available = rlims.rlims_plot.notna().sum()
available.index = [dt.datetime(int(y), 1, 1) for y in available.index]
as_perc.where(available > 30).plot()

# %% trusted=true
# Overall ranges
print('MAR:')
mar_combo = pd.concat([
    mar_ru_minus_mad[2].sum(),
    mar_ru[2].sum(),
    mar_ru_plus_mad[2].sum(),
    mar_ru[1].sum(),
    mar_ru[3].sum()
], axis=1)
mar_combo.columns = ['-1MAD', 'MED', '+1MAD', 'Model_All', 'Model_Above8592']
print(mar_combo)
print('\n1985-2020 totals:')
print(mar_combo.sum())

print('\n\nRACMO:')
racmo_combo = pd.concat([
    racmo_ru_minus_mad[2].sum(),
    racmo_ru[2].sum(),
    racmo_ru_plus_mad[2].sum(),
    racmo_ru[1].sum(),
    racmo_ru[3].sum()
], axis=1)
racmo_combo.columns = ['-1MAD', 'MED', '+1MAD', 'Model_All', 'Model_Above8592']
print(racmo_combo)
print('\n1985-2020 totals:')
print(racmo_combo.sum())


# print('\n\nArea differences c/w observed:')
# print('MAR:')
# print(reg_ru_area.mean())
# print('RACMO:')
# print(racmo_reg_ru_area.mean())

# %% trusted=true
mar_area.index.year

# %% trusted=true
import statsmodels.api as sm
m_mar = sm.OLS(mar_area, sm.add_constant(mar_area.index.year))
print(m_mar.fit().summary())

m_racmo = sm.OLS(racmo_area, sm.add_constant(racmo_area.index.year))
print(m_racmo.fit().summary())

# %% trusted=true
# Extra panel : cumulative by region? But which model outputs to use?
# Basically, might need some kind of summary of these rather detailed plots. Perhaps they are best 
# located in supplementary, in which case the question is - what's the best fig. 2 option?


# %% trusted=true
(racmo_ru[2].sum()/1e6).sum()

# %% trusted=true
# need RACMO cell area weightings !!!

# %% trusted=true
# Percentage of total area runoff which occurs above the 1985-1992 limit.

# %% trusted=true
# How much retention is RACMO doing, given that its 'invisible' runoff limit closely matches the 
# observed visible runoff limit? 

# %% trusted=true
imbie_mar_minus = mar_ru_minus_mad[2].loc['2012':'2017', :].sum(axis=1).mean()
imbie_mar_plus = mar_ru_plus_mad[2].loc['2012':'2017', :].sum(axis=1).mean()
imbie_mar = mar_ru[2].loc['2012':'2017', :].sum(axis=1).mean()
imbie_racmo_minus = racmo_ru_minus_mad[2].loc['2012':'2017', :].sum(axis=1).mean()
imbie_racmo_plus = racmo_ru_plus_mad[2].loc['2012':'2017', :].sum(axis=1).mean()
imbie_racmo = racmo_ru[2].loc['2012':'2017', :].sum(axis=1).mean()
print('MAR: {m} ({l}-{u}) Gt a-1'.format(m=imbie_mar, l=imbie_mar_minus, u=imbie_mar_plus))
print('RACMO: {m} ({l}-{u}) Gt a-1'.format(m=imbie_racmo, l=imbie_racmo_minus, u=imbie_racmo_plus))

# %% trusted=true
imbie_total = 244
lower = 100/imbie_total * imbie_racmo_minus
upper = 100/imbie_total * imbie_mar_plus
perc_racmo = 100/imbie_total * imbie_racmo
perc_mar = 100/imbie_total * imbie_mar
print('Largest perc range: {l}-{u}'.format(l=lower, u=upper))
print('Model perc range: {racmo}-{mar}'.format(racmo=perc_racmo, mar=perc_mar))

# %% trusted=true
# 2012-2017 total runoff from expanded runoff area, for comparison with IMBIE
imbie_compare = regions_roff_zone_pd.loc['2012':'2017', :].sum()
print('IMBIE period total Gt', imbie_compare.sum())
print('IMBIE period annual avg Gt', imbie_compare.mean())
