# ---
# jupyter:
#   jupytext:
#     cell_metadata_filter: all
#     formats: ipynb,py:percent
#     notebook_metadata_filter: all,-language_info
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.7.1
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Runoff-retention factors near the runoff limit, according to elevation
#
# This script calculates the retention factor for the area that lies within 200 m elevation down-hill of the runoff limit. It uses GIMP DEM (currently at 2 km resolution) initially, then resamples to coarser MAR resolutions after. This is needed as, at MAR resolution, the region of interest can be too small in terms of pixels, making it difficult to find the downslope limit as a function of elevation.

# %% trusted=true
from load_env import *
gis = gisloader()

from geoutils.georaster import Raster

from copy import deepcopy
from scipy.spatial import cKDTree
from rasterio import fill
from rasterio.enums import Resampling

# %% trusted=true
from weight_mar_runoff import polygon_to_mar_mask, rasterize_runoff_limit, create_index_grids

# %% trusted=true
gimpdem = Raster('/home/geoscience/nobackup_cassandra/L0data/GIMPDEM/gimpdem_90m_v01.1_EPSG3413_grisonly_filled_2km.tif')
gimp_xds = gimpdem.to_xarray().squeeze()
ii, jj = create_index_grids(gimp_xds)

# %% trusted=true
cells_of_interest = np.ones((gimpdem.shape))


# %% trusted=true
def define_retention_limits(gimpdem, ii, jj, rlim_line_rast,
                            win=50, min_ediff=200, max_ediff=350):
    """ Find pixels which are down-ice of runoff limit.
    
    This function defines the lower retention limits pixel-wise.
    
    It operates in matrix coordinate space, not geographic space.
    
    gimpdem : Raster of GIMPDEM 
    ii, jj : integer matrix coordinate grids (*not* geographic)
    rlim_line_rast : matrix @ gimpdem dims of runoff limit location
    win : search window half-size (pixels) in which to look for lower-elevation points
    min_ediff : chosen points must be at least this much lower than runoff limit
    max_ediff : chosen points cannot be any lower away from runoff limit than this
    
    returns:
        new_grid : matrix. runoff limit still labelled as 1. Chosen lower
        limit pixels denoted by their relative elevation from the runoff limit,
        e.g. -200 m.
        
    """
    
    ir = np.where(rlim_line_rast > 0, ii, np.nan)
    ir = ir[~np.isnan(ir)]
    jr = np.where(rlim_line_rast > 0, jj, np.nan)
    jr = jr[~np.isnan(jr)]

    new_grid = deepcopy(rlim_line_rast).astype(np.float32)

    coords = []
    values = []

    for ip, jp in zip(ir, jr):
        ip = int(ip)
        jp = int(jp)

        jp_min = np.maximum(jp-win, 0)
        ip_min = np.maximum(ip-win, 0)
        jp_max = np.minimum(jp+win, gimpdem.data.shape[1])
        ip_max = np.minimum(ip+win, gimpdem.data.shape[2])

        data_win = gimpdem.data[:, jp_min:jp_max, ip_min:ip_max].squeeze()
        e_rlim = gimpdem.data[:,jp,ip].squeeze()
        data = np.where( (data_win < (e_rlim-min_ediff)) & (data_win > (e_rlim-max_ediff)), data_win-e_rlim, np.nan)

        x = ii[jp_min:jp_max, ip_min:ip_max]
        x = np.where(~np.isnan(data), x, np.nan)
        x = x[~np.isnan(x)]

        if len(x) == 0:
            continue

        y = jj[jp_min:jp_max, ip_min:ip_max]
        y = np.where(~np.isnan(data), y, np.nan)
        y = y[~np.isnan(y)]

        tree = cKDTree(np.array((x, y)).swapaxes(0, 1))

        distances, indices = tree.query((ip, jp), k=1)

        data_flat = data[~np.isnan(data)]
        xx = int(x[indices])
        yy = int(y[indices])
        dd = data_flat[indices]

        new_grid[yy,xx] = dd

        coords.append((xx,yy))
        values.append(dd)
        
    return new_grid

def interp_retention_zone(retention_limits, norm_min=-200, win=25):
    """ 
    
    Input raster has runoff limit still labelled at 1 m asl elevation
    Here, convert to -0.01, in order that all values >= 0 can be assigned
    to have zero-retention.
    
    """
    elev_interp = np.where(retention_limits == 1, -0.01, retention_limits)
    mask = np.where(retention_limits == 0, 0, 1)
    elev_interp = fill.fillnodata(elev_interp, mask=mask, 
                                  smoothing_iterations=0, max_search_distance=win)
    # Normalise elevations
    norm_elev = np.where(elev_interp > norm_min, elev_interp, norm_min)
    norm_elev = np.where(norm_elev < -0.01, norm_elev, -0.01)
    return norm_elev

def resample_to_mar(src_geo, hires, dst_geo):
    elev_R = Raster.from_array(hires, src_geo.transform, src_geo.crs)
    # left bottom right top ... but actually not. bottom and top are flipped :[
    b = dst_geo.rio.bounds()
    b = dict(left=b[0], bottom=b[3], right=b[2], top=b[1])

    elev_R_MAR = elev_R.reproject(dst_crs=dst_geo.rio.crs, 
                                  resampling=Resampling.average, 
                                  dst_res=dst_geo.rio.resolution()[0], 
                                  dst_bounds=b)
    return elev_R_MAR


# %% trusted=true
## RUN ANNUALLY

xa_stack = []
for year in range(1985, 2021):

    print(year)
    
    poly_f = os.path.join(os.environ['ANALYSIS_OUT'], 'runoff_limit_polys/poly_{y}.shp'.format(y=year))
    poly = gpd.read_file(poly_f)
    
    rlim_poly_rast, rlim_line_rast = rasterize_runoff_limit(poly, gimp_xds, cells_of_interest)

    retention_pts = define_retention_limits(gimpdem, ii, jj, rlim_line_rast)
    elev_interp = interp_retention_zone(retention_pts, norm_min=-200)
    
    #elev_interp = np.where(rlim_poly_rast > 0, 0, elev_interp)
    
    retain = 0.0015 * elev_interp + 0.28
    retain2 = np.where(rlim_poly_rast > 0, 0.28, retain)
    retain2 = np.where((elev_interp == -0.01) & (rlim_poly_rast < 1), 0, retain2)
    
    at_mar = resample_to_mar(gimpdem, retain2, gis.mar_ds)
    at_mar.save(os.path.join(os.environ['ANALYSIS_OUT'], 'retention_zones', 'retention_{y}.tif'.format(y=year)))
    
    xa_stack.append(np.flipud(at_mar.data))

# %% trusted=true
print('Saving .nc')
xda = xr.DataArray(np.array(xa_stack).squeeze(), 
                   dims=('time', 'y', 'x'), 
                   coords={'time':np.arange(1985,2021,1), 'y':gis.mar_ds.y[::-1], 'x':gis.mar_ds.x})
xda.to_netcdf(os.path.join(os.environ['ANALYSIS_OUT'], 'retention_zones', 'retention_10k.nc'))

# %% trusted=true
## RUN FOR THE TWO COMMON PERIODS

for period in ('1985_1992', '2013_2020'):

    print(period)
    
    poly_f = os.path.join(os.environ['ANALYSIS_OUT'], 'runoff_limit_polys/poly_{p}_median_edited.shp'.format(p=period))
    poly = gpd.read_file(poly_f)
    
    rlim_poly_rast, rlim_line_rast = rasterize_runoff_limit(poly, gimp_xds, cells_of_interest)

    retention_pts = define_retention_limits(gimpdem, ii, jj, rlim_line_rast)
    elev_interp = interp_retention_zone(retention_pts, norm_min=-200)
    
    #elev_interp = np.where(rlim_poly_rast > 0, 0, elev_interp)
    
    retain = 0.0015 * elev_interp + 0.28
    retain2 = np.where(rlim_poly_rast > 0, 0.28, retain)
    retain2 = np.where((elev_interp == -0.01) & (rlim_poly_rast < 1), 0, retain2)
    
    at_mar = resample_to_mar(gimpdem, retain2, gis.mar_ds)
    at_mar.save(os.path.join(os.environ['ANALYSIS_OUT'], 'retention_zones', 'retention_{p}.tif'.format(p=period)))
    at_mar.to_xarray().to_netcdf(os.path.join(os.environ['ANALYSIS_OUT'], 'retention_zones', 'retention_10k_{p}.nc'.format(p=period)))

# %% [markdown]
# ## Various checks/debugging

# %% trusted=true
plt.figure()
plt.imshow(rlim_poly_rast, interpolation='none')
plt.colorbar()

# %% trusted=true
xda.plot(col='time', col_wrap=5)

# %% trusted=true
plt.figure()
#plt.imshow(np.where(retention_pts < 0, 1, retention_pts), interpolation='none')
plt.imshow(retain2, interpolation='none')
plt.colorbar()


# %% trusted=true
# %matplotlib widget
plt.figure()
ax= plt.subplot(111)
gis.mar_ds.SH.plot(ax=ax)
at_mar.show(interpolation='none', alpha=0.5, ax=ax, zorder=100)

# %% trusted=true
plt.figure()
#plt.imshow(test_fill, interpolation='none')
plt.imshow(np.where(elev_interp > -200, elev_interp, -200), interpolation='none')
plt.colorbar()

# %% trusted=true

# %% trusted=true
retain = 0.0015 * elev_R_MAR.data + 0.28
# Make sure that we can't "negatively" retain water!
retain = np.where(retain < 0, 0, retain)
retain = np.where(np.isnan(retain), 0, retain)




# %% trusted=true
print(retain)
# %matplotlib widget
plt.figure()
retain.show(interpolation='none')

# %% trusted=true
plt.figure()
plt.imshow(1-retain.data.squeeze())
plt.colorbar()

# %%
# Geographic coordinates version - takes much longer (probably due to the raster clipping, memory faff etc)

"""
xx, yy = gimpdem.coords()

xxr = np.where(rlim_line_rast > 0, xx, np.nan)
xxr = xxr[~np.isnan(xxr)]
yyr = np.where(rlim_line_rast > 0, yy, np.nan)
yyr = yyr[~np.isnan(yyr)]

new_grid = deepcopy(rlim_line_rast).astype(np.float32)

min_ediff = 200
max_ediff = 350

coords = []
win = 50000

for xp, yp in zip(xxr, yyr):
  
    clipped = gimpdem.copy()
    clipped.crop([xp-win, yp-win, xp+win, yp+win])
    e_rlim = clipped.value_at_coords(xp, yp)
    data_win = clipped.data
    data = np.where( (data_win < (e_rlim-min_ediff)) & (data_win > (e_rlim-max_ediff)), data_win-e_rlim, np.nan)
    
    xxc, yyc = clipped.coords()
    
    xc = np.where(~np.isnan(data), xxc, np.nan)
    xc = xc[~np.isnan(xc)]
    
    if len(xc) == 0:
        continue
        
    yc = np.where(~np.isnan(data), yyc, np.nan)
    yc = yc[~np.isnan(yc)]
    
    tree = cKDTree(np.array((xc, yc)).swapaxes(0, 1))
    
    _, ix = tree.query((xp, yp), k=1)
    
    data_flat = data[~np.isnan(data)]
    x = int(xc[ix])
    y = int(yc[ix])
    dd = data_flat[ix]
    
    i, j = clipped.xy2ij(x, y)
    
    new_grid[j,i] = dd
    
    #coords.append((xx,yy))
"""
