# %% trusted=true
"""

Seasonality analysis to estimate annual maximum runoff limit elevation.
- Does not find coordinates. This is done with estimate_coords_from_elev.py.

Script has two main sections:
1. Compute the polynomial that describes the seasonality of each box.
2. Apply the polynomial to xytpd runoff limit elevations, to estimate the 
maximum runoff elevation each year in each slice.

AT August 2021


"""

# %% trusted=true
import pandas as pd
from sklearn.preprocessing import normalize, StandardScaler
import numpy as np

# %% trusted=true
from load_env import *
gis = gisloader()
rlims = rlimsloader()

# %% trusted=true
from sqlalchemy import create_engine
from find_annual_rlims import query_box

# %% [markdown]
# ------------------------------------------------------------------------------

# %% trusted=true
doy_range = np.arange(182, 245)
# Degree of polynomial to fit
p_degrees = 2

# %% trusted=true
plot = True

# %% [markdown]
# ------------------------------------------------------------------------------

# %% trusted=true
engine = create_engine(db_href)

# %% trusted=true
store = []
for region in gis.regions_boxes:
    
    print(region)
    
    bmin, bmax = gis.regions_boxes[region]
    df = []
    for n in range(bmin, bmax+1):
        df_box = query_box(n, engine)
        df.append(df_box)
    df = pd.concat(df, axis=0)

    # Prepare the data for conversion to seasonality - this requires DOYs.
    df.loc[:,'doy'] = pd.DatetimeIndex(df.date).dayofyear
    df.index = pd.RangeIndex(start=0,stop=len(df))

    # Identify products which are clearly very noisy.
    retain_ids = df.groupby('product_id').elev.mad()
    retain_ids.name = 'product_id__mad'
    retain_ids = retain_ids[retain_ids < 75] 

    # Delete the noisy products.
    df_culled = pd.merge(left=df, right=retain_ids, how='right', 
        left_on='product_id', right_index=True)

    # Do scaling
    outputs = []
    for year in df_culled.year.unique():
        df_year = df_culled[df_culled.year == year]
        
        # do some filtering to get rid of very small values
        #df_year = df_year[df_year.elev > ]
        ts = pd.DatetimeIndex(df_year.date)
        df_year.loc[:,'doy'] = ts.dayofyear

        dd = df_year.filter(items=['elev'], axis='columns').to_numpy()
        if len(dd) > 0:
            scaler = StandardScaler(with_std=False).fit(dd)

            norm = scaler.transform(dd)    

            outputs.append(pd.Series(norm.squeeze(), index=df_year.doy))
            
    all_data = pd.concat(outputs)
    
    if plot:
        plt.figure()
        plt.plot(all_data.index, all_data, 'o', alpha=0.3, label=str(year))   

    # Fit a polynomial through the averaged seasonality series
    doys_ts = all_data.groupby('doy').median()
    doys_ts_filt = doys_ts[doys_ts > all_data.median()-(all_data.mad()*2)]
    
    p = np.polynomial.Polynomial.fit(doys_ts_filt.index.values, doys_ts_filt.values, p_degrees)
    if plot:
        xx, yy = p.linspace()
        plt.plot(xx, yy, color='orange')
        plt.savefig(os.path.join(os.environ['ANALYSIS_OUT'], 'seasonality', 'season_poly2_region_{n}'.format(n=region)))
        plt.close()
        
    nph, _, _ = np.histogram2d(all_data.values, all_data.index, bins=[np.arange(-500, 500, 20), np.arange(180,250,1)])
    np.savetxt(os.path.join(os.environ['ANALYSIS_OUT'], 'seasonality', 'region_{n}_binned_freqs.csv'.format(n=region)),
              nph, delimiter=',')
    
    c1, c2, c3 = p.convert()
    if region == 'NW':
        region_save = 'NW_unfilt'
    else:
        region_save = region
    store.append(dict(region=region_save, c1=c1, c2=c2, c3=c3))
    
    if region == 'NW':
        print('NW-specific filtering')
        doys_ts_filt = doys_ts_filt[doys_ts_filt.index <= 230]
        p = np.polynomial.Polynomial.fit(doys_ts_filt.index.values, doys_ts_filt.values, p_degrees)
        c1, c2, c3 = p.convert()
        store.append(dict(region='NW', c1=c1, c2=c2, c3=c3))



# %% trusted=true
all_polys = pd.DataFrame(store)
all_polys.index = all_polys.region
all_polys.to_csv(os.path.join(os.environ['ANALYSIS_OUT'], 'seasonality', 'region_polynomial_d2_coefs.csv'))

# %% trusted=true
# %matplotlib widget

output = sns.histplot(data=df, x='doy', y='elev', pthresh=0.1, binwidth=(1,20))  
doy_range = np.arange(182,244)
ediff = (c3 * doy_range**2) +     \
        (c2 * doy_range) +        \
        c1
plt.plot(doy_range, ediff)

# %% [markdown]
# ------------------------------------------------------------------------------

# %% trusted=true
e_max_store = []
for region in gis.regions_boxes:

    ## Estimate maximum runoff limit elevation for every slice in the box.
    # Adjust polynomial curve using the identified maximum runoff limit as the centred 0-value for the DOY of observation.
    # This allows to then find the maximum value according to the adjusted curve.
    bmin, bmax = gis.regions_boxes[region]
    region_rlims = rlims.xytpd[(rlims.xytpd['box_id'] >= bmin) & (rlims.xytpd['box_id'] <= bmax)]
    region_rlims.loc[:,'doy'] = pd.DatetimeIndex(region_rlims.date).dayofyear

    coefs = all_polys.loc[region]

    #ediff = (coefs.c4 * doy_range**3) + \
    ediff =    (coefs.c3 * doy_range**2) +     \
        (coefs.c2 * doy_range) +        \
        coefs.c1
    ediffs = pd.Series(ediff, index=doy_range)

    outputs = []
    for ix, row in region_rlims.iterrows():
        yi = ediffs.loc[row.doy]
        adj_yy = ediffs - yi + row.elev
        est_max = np.max(adj_yy)
        outputs.append(dict(year=row.year, region=region, box_id=row.box_id, slice_id=row.slice_id, 
            est_max=est_max, loc_id=row.loc_id))

    e_max_store.extend(outputs)

# %% trusted=true
e_maxes = pd.DataFrame(e_max_store)
e_maxes.to_csv(os.path.join(os.environ['ANALYSIS_OUT'], 'seasonality', 'annual_emax_region_poly2.csv'))

# %%
