"""

Extract centreline elevation profiles as 2nd degree polynomials, from
rectangular polygons (in this case, slices for the Landsat runoff limits work).

July 2021 AJT

"""

import pandas as pd
from shapely.geometry import LineString
import rasterio
import numpy as np
import geoutils as gu
import geopandas as gpd

dem_in = "/flash/tedstona/L0data/GIMPDEM/gimpdem_30m_3413.vrt"
msk_in = "/flash/tedstona/L0data/GIMPMASK/gimpmask_30m_3413.vrt"


def _find_midpt(asdf, origin_iloc):
    origin = asdf.iloc[origin_iloc]
    dest = asdf.loc[origin.name - 1]
    mx = (origin.x + dest.x) / 2
    my = (origin.y + dest.y) / 2
    return (mx, my)


def calculate_centreline(geom):
    """
    Given a rectangular Polygon (5 nodes), return a LineString representing
    the centre line of the long axis of the rectangle.
    """

    # Find the start and end positions of the slice's centre line.
    x, y = geom.exterior.coords.xy
    assert len(x) == 5
    x = pd.Series(x)
    y = pd.Series(y)

    dists = np.sqrt((x.shift(1) - x)**2 + (y.shift(1) - y)**2)

    corners = pd.concat([x, y, dists], axis=1)
    corners.columns = ['x', 'y', 'd']

    # Sort by distance so that we get the corner positions associated with the
    # short bounds
    corners = corners.sort_values('d')

    # Find the mid points at each end of the slice.
    pt1 = _find_midpt(corners, 0)
    pt2 = _find_midpt(corners, 1)
    # Connect these points with a line.
    ls = LineString([pt1, pt2])

    return ls


def dem_values_on_centreline(ls, flip_y=True):
    """ Given a LineString, return the unique values of from a DEM that fall
    on the line.

    Returns pandas.Dataframe(x,y,z)
    """

    # Load the raster data coincident with the line.
    # DEM
    with rasterio.open(dem_in) as src:
        out_image, out_transform = rasterio.mask.mask(src, [ls], crop=True)
        out_meta = src.meta
        gr = gu.Raster.from_array(
            out_image, transform=out_transform, crs=out_meta['crs'])
    # Mask
    with rasterio.open(msk_in) as src:
        msk, msk_trans = rasterio.mask.mask(src, [ls], crop=True)
        msk_meta = src.meta
        gr_msk = gu.Raster.from_array(
            msk, transform=msk_trans, crs=msk_meta['crs'])

    # Mask the DEM to the on-ice portions
    gr.data = np.where(gr_msk.data > 0, gr.data, -1)

    # Make a data frame of the x,y,z coordiantes.
    xx, yy = gr.coords()
    zz = np.ravel(gr.data)
    z = zz[zz > 0]
    x = np.ravel(xx)[zz > 0]
    if flip_y:
        yy = np.flipud(yy)
    y = np.ravel(yy)[zz > 0]
    xyz = pd.DataFrame({'x': x, 'y': y, 'z': z})
    xyz = xyz.groupby('z').mean()
    xyz = xyz.sort_values('z')

    return xyz


def calculate_polys(xyz):
    # Calculate polynomials
    # Elev --> X coordinate
    poly_x = np.polynomial.Polynomial.fit(xyz.index, xyz.x, 2)

    # Elev --> Y coordinate
    poly_y = np.polynomial.Polynomial.fit(xyz.index, xyz.y, 2)

    return(poly_x, poly_y)


def control_mp(box_id):

    slices = gpd.read_file(
        '/home/geoscience/nobackup_cassandra/L0data/100k_boxes/boxes_fid{:d}.shp'.format(box_id))

    polynomials = []

    for s_ix, s in slices.iterrows():
        centre = calculate_centreline(s.geometry)
        xyz = dem_values_on_centreline(centre)

        if calc_polys:
            pX, pY = calculate_polys(xyz)

            xc1, xc2, xc3 = pX.convert()
            yc1, yc2, yc3 = pY.convert()

            polynomials.append(dict(slice_id=s_ix,
                                    xc1=xc1, xc2=xc2, xc3=xc3,
                                    yc1=yc1, yc2=yc2, yc3=yc3))
        else:
            xyz.to_csv(os.path.join(os.environ['ANALYSIS_OUT'],
                    'slice_centre_elevations',
                    'slice_elevations_box{n}_s{s}.csv'.format(n=box_id, s=s_ix)))   
            
    if calc_polys:
        aspd = pd.DataFrame.from_dict(polynomials)
        aspd.index = aspd.slice_id
        aspd.to_csv(os.path.join(os.environ['ANALYSIS_OUT'],
            'slice_centre_elevations',
            'slice_polynomials_box{n}.csv'.format(n=box_id)))


if __name__ == '__main__':

    print('Executing ...')

    calc_polys = False

    from multiprocessing import Pool
    from tqdm import tqdm

    from load_env import *
    gis = gisloader()

    # Do extraction box-by-box
    try:
        nworkers = int(os.environ['EOLEARN_WORKERS'])
    except KeyError:
        print('WARNING: EOLEARN_WORKERS env-var not set, defaulting to one worker.')
        nworkers = 1
    pool = Pool(processes=nworkers,
              maxtasksperchild=1)
    for _ in tqdm(pool.imap_unordered(control_mp, gis.boxes.FID), total=len(gis.boxes)):
        pass
    pool.close()
