# ---
# jupyter:
#   jupytext:
#     cell_metadata_filter: all
#     formats: ipynb,py:percent
#     notebook_metadata_filter: all,-language_info
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.7.1
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Plotting the difference between the observed and Emax-estimated runoff limits

# %% trusted=true
from load_env import *
gis = gisloader()
rlims = rlimsloader()

# %% trusted=true
xytpd_emax = pd.read_csv(os.path.join(os.environ['ANALYSIS_OUT'], 'rlim_annual_maxm', 'xytpd_region_emax.csv'))

# %% trusted=true
xytpd_emax

# %% trusted=true
med_filt = rlims.xytpd.filter(items=['elev', 'year', 'box_id', 'slice_id'], axis='columns')
med_filt.index = [med_filt.year, med_filt.box_id, med_filt.slice_id]
emax_filt = xytpd_emax.filter(items=['x', 'y', 'z', 'year', 'box_id', 'slice_id'], axis='columns')
emax_filt.index = [emax_filt.year, emax_filt.box_id, emax_filt.slice_id]

combo = pd.merge(med_filt, emax_filt, left_index=True, right_index=True)

# %% trusted=true
diff = (combo.z - combo.elev)
diff.name = 'diff'
diff = diff.reset_index()

# %% trusted=true
diff

# %% trusted=true
## Load polynomial curves describing each region
region_polys = pd.read_csv(os.path.join(os.environ['ANALYSIS_OUT'], 
                                        'seasonality', 
                                        'region_polynomial_d2_coefs.csv'), 
                           index_col='region')

## Set DOY range over which to plot them
doy_range = np.arange(182,244)


# %matplotlib inline
# width, height
plt.figure(figsize=(5,6.5))

from matplotlib import gridspec
# rows, cols
gs = gridspec.GridSpec(7, 4, wspace=0.7, hspace=0.6)

labels = list(string.ascii_lowercase)[:7]
n = 6

for region in gis.regions_boxes:
    
    ## Part 1: Histogram and seasonal poly
    ax = plt.subplot(gs[n,-1]) #plt.subplot(7,1,n)
    
    # Frequency matrices are generated by seasonality_for_max_rlim.py
    region_freqs = np.genfromtxt(os.path.join(os.environ['ANALYSIS_OUT'], 
                            'seasonality', 
                            'region_{n}_binned_freqs.csv'.format(n=region)),
                                delimiter=',')

    plt.imshow(np.where(region_freqs > 2, region_freqs, np.nan), origin='lower', 
               extent=(180, 250, -500, 500),
               aspect=0.1, 
               cmap=sns.color_palette('Greys',as_cmap=True), 
               interpolation='none')
    
    # Special case of NW - show unfiltered and filtered polynomial fits
    linestyle = '-'
    alpha = 1
    if region == 'NW':
        ediff = (region_polys.loc['NW_unfilt'].c3 * doy_range**2) + \
            (region_polys.loc['NW_unfilt'].c2 * doy_range) + \
            region_polys.loc['NW_unfilt'].c1
        plt.plot(doy_range, ediff, linewidth=1, color=regions_colors[region], linestyle='--', alpha=0.7)
    
    # Plot chosen polynomial fit
    ediff = (region_polys.loc[region].c3 * doy_range**2) +     \
            (region_polys.loc[region].c2 * doy_range) +        \
            region_polys.loc[region].c1
    plt.plot(doy_range, ediff, linewidth=1, color=regions_colors[region], alpha=1)
    
    # Set grid properties
    plt.ylim(-200, 200)
    plt.xlim( 182, 250)
    plt.xticks([190, 220, 250])
    
    
    ## Part 2: Box plots
    ax2 = plt.subplot(gs[n,0:-1])
    bmin, bmax = gis.regions_boxes[region]
    
    status = rlims.rlims_plot[(rlims.rlims_plot.index >= bmin) & 
                                  (rlims.rlims_plot.index <= bmax)]
    status = status.mean()
    
    for year in range(1985, 2021):
        
        if np.isnan(status.loc[year]):
            continue
        
        data = diff[(diff.year == year) & 
                    (diff.box_id >= bmin) & 
                    (diff.box_id <= bmax)]
        if len(data) < 1:
            continue
        plt.boxplot(data['diff'], positions=[year], widths=0.65, 
                    patch_artist=True,
                    flierprops={'markersize':3, 'marker':'.', 'mfc':'grey', 'mec':'none'}, 
                    boxprops={'facecolor':regions_colors[region], 'edgecolor':regions_colors[region]}, 
                    medianprops={'color':'black'}, 
                    whiskerprops={'color':regions_colors[region]},
                    capprops={'color':regions_colors[region]},
                    showfliers=False)
        
    plt.ylim(-20, 200)

    if n == 3:
        plt.ylabel('Difference (m)')
        
    if n != 6:
        ax2.tick_params(axis='x', labelbottom=False)
        ax.tick_params(axis='x', labelbottom=False)
    if n == 6:
        ax.set_xlabel('Day of Year')
    

    xlabels = np.arange(1985, 2021, 5)
    plt.xticks(xlabels, xlabels)
    plt.xlim(1984, 2021)
    
    ax2.annotate(labels[n], fontweight='bold', xy=(0.03,0.9), xycoords='axes fraction',
           horizontalalignment='left', verticalalignment='bottom', bbox=dict(fc='white', ec="none", alpha=0.6))
    
    ax2.annotate(region, xy=(0.09,0.9), xycoords='axes fraction', color=regions_colors[region],
           horizontalalignment='left', verticalalignment='bottom', bbox=dict(fc='white', ec="none", alpha=0.6))
    
    n -= 1

sns.despine()
    
plt.tight_layout(h_pad=1)
plt.savefig(os.path.join(os.environ['FIGS_OUT'],'suppl/sf_full_temporal_regions_withpolys.pdf'))

# %%
