# ---
# jupyter:
#   jupytext:
#     cell_metadata_filter: all
#     formats: ipynb,py:percent
#     notebook_metadata_filter: all,-language_info
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.7.1
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# %% trusted=true
""" 
2021-02-17 : Note that some functions in here are not currently being used,
but they are retained for future reference. For an implementation where they are
in use, check the repo history on/prior to this date.

N.b. this is very expensive to run for finer grid resolutions! (e.g. RACMO 1km)
"""

# %% [markdown]
# # Calculate weighted RCM runoff masks
# This script calculates the percentage of each RCM cell which lies within a given observed runoff limit. It produces a mask of weights, which should be applied to calculations of runoff volumes made using RCM data.

# %% trusted=true

from shapely.geometry import Polygon
from shapely.errors import TopologicalError
import osmnx as ox
from skimage.morphology import binary_erosion
import geopandas as gpd
import numpy as np
import pandas as pd
import xarray as xr
import argparse

from load_env import polygon_to_mar_mask


# %% trusted=true
def compute_grid_polys(da, dims=['x', 'y'], grid_crs=None, aoi=None):

    if len(dims) == 2:
        xdim, ydim = dims
   
    Xs = da[xdim].values
    Ys = da[ydim].values
    Xr = (Xs[1] - Xs[0]) / 2
    Yr = (Ys[1] - Ys[0]) / 2
    grid_polys = []
    for i in range(da[xdim].size):
        for j in range(da[ydim].size):
            
            if aoi is not None:
                if aoi.isel(x=i, y=j) < 1:
                    continue
                
            grid = Polygon(
                [(Xs[i]-Xr, Ys[j]-Yr),
                 (Xs[i]+Xr, Ys[j]-Yr),
                 (Xs[i]+Xr, Ys[j]+Yr),
                 (Xs[i]-Xr, Ys[j]+Yr),
                 (Xs[i]-Xr, Ys[j]-Yr)]
            )
            grid_polys.append(dict(
                i=i, 
                j=j, 
                geometry=grid))        

    gdf = gpd.GeoDataFrame(pd.DataFrame(grid_polys), crs=da.rio.crs)
    return gdf


# %% trusted=true
def _matches(geometry, spatial_index, gdf, quadrats=None):
    """ Find polygons in gdf which intersect with geometry.
    
    Uses gdf spatial index to speed up the search.
    
    geometry : polygon to check for intersections
    spatial_index : .sindex of gdf
    gdf : GeoDataFrame of features to check for intersectiona against polygon
    quadrats : None or distance in metres
    
    returns : rows of gdf which intersect with geometry.
    
    Based on:
    https://github.com/gboeing/urban-data-science/blob/master/19-Spatial-Analysis-and-Cartography/rtree-spatial-indexing.ipynb
    """
    
    def _check(g):
        possible_matches_index = list(spatial_index.intersection(g.bounds))
        possible_matches = gdf.iloc[possible_matches_index]
        if len(possible_matches) < 1:
            return None
        precise_matches = possible_matches[possible_matches.intersects(g)]
        return precise_matches
    
    if quadrats is not None:
            geometry_cut = ox.utils_geo._quadrat_cut_geometry(geometry.iloc[0].geometry, quadrats)
            matches = []
            for poly in geometry_cut:
                matches.append(_check(poly))
            m = pd.concat(matches, axis=0)
    else:
        m = _check(geometry.iloc[0].geometry)
    
    return m


def find_cells_at_limit(poly_rlim, mar_grid, grid_sindex):
    """ Find cells which are crossed by the runoff limit.
    
    poly_rlim : polygon of the runoff limit
    mar_grid : gdf of polygons which form the grid to search
    grid_sindex : the gdf's .sindex
    
    returns : gdf of cells which are crossed by the poly_rlim.
    
    """
    # technically this is a 'crosses' operation but the native GeoPandas functionality doesn't seem to work.
    intersects = _matches(poly_rlim, grid_sindex, mar_grid, quadrats=500000)
    within = intersects.within(poly_rlim.iloc[0].geometry)
    intersects['crosses'] = np.where((within == False), True, False)

    return intersects[intersects['crosses'] == True]


def calculate_weights(geometry, cells, cartesian=True, interior=True):
    """
    
    interior : bool. If True, the weight is for the area of the cell within the geometry.
        If False, the weight is for the area outside the geometry.
        
    """
    if cartesian:
        cells_cart = cells.to_crs(8857)
        geometry = geometry.to_crs(8857)
    a = cells_cart.geometry.intersection(geometry.iloc[0].geometry).area
    cells['w'] = 1. / cells_cart.geometry.area * a
    if interior is False:
        cells['w'] = 1 - cells['w']    
    
    return cells


def create_weights_mask(weights_gdf, ds, ds_xdim='x', ds_ydim='y', wcol='w', into=None):
    """
    
    weights_gdf : gdf of cells which contains `weights` column
    ds : DataArray of interest, must match geospatial attributes of weights_gdf
    ds_xdim, ds_ydim : names of the x and y coordinates in the DataArray
    wcol : name of `weights` column in weights_gdf
    into : None, or numpy.array whose shape corresponds to ds. Usually a binary mask (1/0),
        this function will substitute values of 1 for the cell weight where relevant. Weights
        for cells labelled 0 in `into` will not be substituted in.
        
    returns : msk, into as DataArrays.
    
    """
    msk = np.zeros(( len(ds[ds_ydim]), len(ds[ds_xdim]) ))
    
    if into is not None:
        if into.shape != msk.shape:
            raise ValueError('Shape of `into` does not match `ds`.')
        into = into.astype(np.float32)
            
    for ix, g in weights_gdf.iterrows():
        msk[g.j,g.i] = g[wcol] 
        
        if into is not None:
            if into[g.j, g.i] > 0:
                into[g.j, g.i] = g[wcol]
                
    msk_nc = xr.DataArray(msk, dims=(ds_ydim, ds_xdim), 
                          coords={ds_xdim:ds[ds_xdim], ds_ydim:ds[ds_ydim]})
    into_nc = xr.DataArray(into, dims=(ds_ydim, ds_xdim), 
                           coords={ds_xdim:ds[ds_xdim], ds_ydim:ds[ds_ydim]})
                
    if into is not None:
        return (msk_nc, into_nc)
    else:
        return msk_nc

    
def weights_for_rlim(poly, mar_ds, mar_grid, sindex, msk_into):
    """ Find weights for a given runoff limit polygon 
    
    Convenience function.
    
    """
    c = find_cells_at_limit(poly, mar_grid, sindex)
    w = calculate_weights(poly, c, interior=False)
    m, i = create_weights_mask(w, mar_ds, into=msk_into)
    return m, i


# -----------------------------------------------------------------------------


def create_index_grids(xds, xdim='x', ydim='y'):
    ni = np.arange(0, len(xds['x']))
    nj = np.arange(0, len(xds['y']))
    ii, jj = np.meshgrid(ni, nj)
    return (ii, jj)


def rasterize_runoff_limit(poly, xds, cells_of_interest):
    """ Create rasters showing runoff limit which corresponds to limits of xds.

    At the moment relies on `polygon_to_mar_mask`

    
    returns: tuple (area, outline)
        area: all of the polygon is set to 1. 
        outline: just the outline of the runoff limit with 1s, i.e. the 
        interior of the runoff limit area is set to zero.

    """
    # Convert to straightforward raster mask
    year_mask = polygon_to_mar_mask(xds, polygons=poly, invert=False, all_touched=True)
    
    rast_line_raw = year_mask - binary_erosion(year_mask).astype(np.int8)
    rast_line_raw = np.where(cells_of_interest == 1, rast_line_raw, 0)
    
    return (year_mask, rast_line_raw)


def calculate_cell_cover(cell, p, area='inside'):
    """ Given a Polygon of a cell and a Polygon p, calculate the area covered.

    cell : Polygon
    p : Polygon to intersect
    interior : if 'inside', area returned is that of the cell which
        falls inside the Polygon p. If 'outside', area corresponds to that outside 
        the Polygon P.

    Returns: (a, w) where a=area in square metres, w=weight (proportion of cell)

    """

    # Use the polygonised runoff limit to cut through the cell
    # Work out proportional cover
    a = cell.geometry.intersection(p.iloc[0].geometry).area
    w = (1./cell.geometry.area) * a

    if area == 'outside':
        w = 1. - w

    return (a, w)


def calculate_rlim_cover(rlim_line_rast, rlim_poly, cells_gdf, ii, jj, cartesian=True):
    """ For a given runoff limit polygon, calculate the % area of each cell
    that runs off.

    rlim_line_rast : np.array. The outline of the runoff limit (interior not filled!)
    rlim_poly : Polygon. Runoff limit as polygon, must be same as contents of rlim_line_rast
    cells_gdf : GeoDataFrame. The MAR cells to be checked against.
    ii : np.array. All x-direction (column) indices.
    jj : np.array. All y-direction (row) indices.
    cartesian: bool. Default True. If True, do area calculations in cartesian system.

    returns : np.array. Complete weighted mask.  0<val<=1 denotes runoff (as fraction of cell).

    """

    CRS_EQUAL_AREA = 8857

    if cartesian:
        cells_gdf = cells_gdf.to_crs(CRS_EQUAL_AREA)
        rlim_poly = rlim_poly.to_crs(CRS_EQUAL_AREA)
        rlim_poly['geometry'] = rlim_poly.buffer(0)

    # Proportions raster
    rlim_rast_prop = np.zeros(rlim_line_rast.shape)

    # Find the column and row indices of the runoff limit
    p_iii = np.where(rlim_line_rast > 0, ii, np.nan)
    p_ii = p_iii[~np.isnan(p_iii)]
    p_jjj = np.where(rlim_line_rast > 0, jj, np.nan)
    p_jj = p_jjj[~np.isnan(p_jjj)]

    # Use these indices to find corresponding MAR cell
    for p_i, p_j in zip(p_ii, p_jj):
        # Find polygon for the cell matching each runoff limit location
        cell = cells_gdf[(cells_gdf.i == p_i) & (cells_gdf.j == p_j)]
        cell = cell.iloc[0]            

        a, w = calculate_cell_cover(cell, rlim_poly, area='outside')
    
        rlim_rast_prop[int(p_j), int(p_i)] = w

    return rlim_rast_prop



################################################################################

# %% trusted=true
if __name__ == '__main__':

    parser = argparse.ArgumentParser(description='Generate weighted masks corresponding to runoff limit for specified RCM.')
    parser.add_argument('rcm', type=str, 
        help='str, mar or racmo, or racmo5 (for 5.5km RACMO masks')
    parser.add_argument('method', type=str,
        help='str, binary or fraction')
    parser.add_argument('-plot', dest='plot', action='store_true')
    parser.add_argument('-poly_type', dest='poly_type', type=str, default=None,
        help='str, default is to use the median polygons (i.e. unlabelled) but can be set to the relevant label, e.g. xytpd_plus_mad')
    parser.add_argument('-noannual', dest='noannual', action='store_true')
    args = parser.parse_args()

    from load_env import *
    gis = gisloader()

    if args.rcm == 'mar':
        rcm_ds = gis.mar_ds
        fn_out = gis.fn_roff_as_mar_masks
    elif args.rcm == 'racmo':
        rcm_ds = gis.racmo_ds
        fn_out = gis.fn_roff_as_racmo_masks
    elif args.rcm == 'racmo5':
        rcm_ds = xr.open_dataset(os.path.join(gis.RACMO_DS_ROOT, '5.5km', 'FGRN055_Masks_5.5km.nc'))
        rcm_ds.rio.write_crs('+init=pstere +lat_0=90' inplace=True)
        fn_out = os.path.join(os.environ['ANALYSIS_OUT'], 'runoff_limit_polys/polys_as_racmo_rasters_annual_5k.nc')
    else:
        raise ValueError('Unknown RCM specified.')


    # Generate an 'RCM mask' of the areas of the runoff limit that we are interested in - this 
    # will limit the production of RCM grid polygons to just these areas.

    # Create a polygon of just the region of interest '& (gis.boxes.FID != 21)'
    boxes_region = gis.boxes[(gis.boxes.FID >= 2) & (gis.boxes.FID <= 32)].to_crs(rcm_ds.rio.crs)
    boxes_region['merge'] = 1
    boxes_region['geometry'] = boxes_region.buffer(2000)
    just_region = boxes_region.dissolve(by='merge').to_crs(rcm_ds.rio.crs)

    # Raster @ RCM dimensions of cells/pixels that we're interested in
    cells_of_interest = polygon_to_mar_mask(rcm_ds, polygons=just_region, nc=True)

    if args.method == 'fraction':
        # Polygons of RCM grid
        rcm_grid = compute_grid_polys(rcm_ds, aoi=cells_of_interest)

        # Row and column indexes of RCM grid
        ii, jj = create_index_grids(rcm_ds)

    if not args.noannual:
        fn_template = 'poly_{y}'
        if args.poly_type is not None:
            fn_template += '_' + args.poly_type
            fn_out = fn_out[:-3] + '_' + args.poly_type + '.nc'
        fn_template += '.shp'

        print(fn_out)

        # %% trusted=true
        # Run through the runoff limits
        yearly_masks_store = {}
        for year in range(1985, 2021):    
            
            poly_f = os.path.join(os.environ['ANALYSIS_OUT'], 'runoff_limit_polys', fn_template.format(y=year))
            print(poly_f)
            
            if os.path.exists(poly_f):
                print(year)
                
                # Load year's polygon
                p = gpd.read_file(poly_f).to_crs(rcm_ds.rio.crs)
                if len(p) != 1:
                    raise ValueError('File with runoff limit does not contain single polygon required.')
                p['geometry'] = p.buffer(0)

                rlim_mask, rlim_line = rasterize_runoff_limit(p, rcm_ds, cells_of_interest)
                if args.method == 'fraction':
                    weights = calculate_rlim_cover(rlim_line, p, rcm_grid, ii, jj)
                    # Now generate the complete weighted mask, which consists of the runoff area
                    # and the weights for the line.
                    # Note that following invert (~), runoff area (margins) = 1, interior = 0.
                    weighted_mask = weights + ~rlim_mask
                elif args.method == 'binary':
                    weighted_mask = rlim_mask
                else:
                    raise ValueError('unknown method')                

                yearly_masks_store[year] = weighted_mask
                
        def concat_to_xr(yearly_masks_store, ds):
            data = np.array([yearly_masks_store[y] for y in yearly_masks_store])
            yearly_masks = xr.DataArray(data, dims=('time', 'y', 'x'), 
                                    coords={'time':[dt.datetime(y,1,1) for y in yearly_masks_store], 
                                    'y':ds.y, 'x':ds.x})
            return yearly_masks

        yearly_masks_xr = concat_to_xr(yearly_masks_store, rcm_ds)
        yearly_masks_xr.to_netcdf(fn_out)


    # %% trusted=true
    # Common periods
    fn_template = 'poly_{ys}_{ye}_median'
    if args.poly_type is not None:
        fn_template += '_' + args.poly_type
    fn_template += '_edited.shp'

    periods = [fn_template.format(ys=1985, ye=1992), fn_template.format(ys=2013, ye=2020)]

    for period in periods:
        print(period)
        poly_f = os.path.join(os.environ['ANALYSIS_OUT'], 'runoff_limit_polys', period)
        # Load year's polygon
        p = gpd.read_file(poly_f).to_crs(rcm_ds.rio.crs)
        if len(p) != 1:
            raise ValueError('File with runoff limit does not contain single polygon required.')
        p['geometry'] = p.buffer(0)

        rlim_mask, rlim_line = rasterize_runoff_limit(p, rcm_ds, cells_of_interest)
        if args.method == 'fraction':
            weights = calculate_rlim_cover(rlim_line, p, rcm_grid, ii, jj)
            # Now generate the complete weighted mask, which consists of the runoff area
            # and the weights for the line.
            # Note that following invert (~), runoff area (margins) = 1, interior = 0.
            weighted_mask = weights + ~rlim_mask
        elif args.method == 'binary':
            weighted_mask = rlim_mask
            
        as_xr = xr.DataArray(weighted_mask, dims=('y', 'x'), 
                                coords={'y':rcm_ds.y, 'x':rcm_ds.x})
        as_xr.to_netcdf(poly_f[:-4] + '_{m}_weights.nc'.format(m=args.rcm))

    # %% trusted=true
    if args.plot:
        
        # %matplotlib widget
        plt.figure()
        plt.imshow(weighted_mask, origin='lower', interpolation='none')
        plt.colorbar()

        plt.figure()
        ax = plt.subplot(111)
        yearly_masks_xr.sel(time='1985').plot(ax=ax)
        p.plot(ax=ax, facecolor='none')
