# README

## USAGE SEQUENCE

The whole processing chain is based around configuration parameters specified in a `template` file. See `model.template` for an example.

The workflow operates under the following assumptions/restrictions:

* Only run with one satellite/sensor platform at a time (e.g. Landsat 5 TM sensor)

There are several steps in running the full river detection processing chain.

References to `beo05` in this document refer to the computing cluster at University of Fribourg, CH for which this workflow was designed.


### Acquiring imagery

Landsat functionality is currently implemented with Google Cloud as the imagery supplier but with a copy of the Google Cloud Landsat database held and queried locally.

This requires the tools found in `github.com/atedstone/gs_landsat.git`.


#### Local database

Download `index.csv.gz` from Google Cloud then run `create_landsat_spatialite_database.py` on a system with SQLite and Spatialite available to build an SQLite database with geometries.


##### Updating the database

For example once a year following each melt season.

Can use Google Cloud Platform to make this quicker. Open the Google Cloud Console. Select 'My First Project'. Run a query that looks something like this:

```sql
SELECT * FROM `bigquery-public-data.cloud_storage_geo_index.landsat_index` WHERE date_acquired > '2019-10-19' AND south_lat > 50 AND cloud_cover < 50
```

(Of course, this does some basic northern hemisphere subsetting, so the database is no longer actually complete)

Then save the result as a csv - it will do this to Google Drive. Download the csv and plug it into create_landsat_spatialite_db.py, with create_db=False.

Finally the local database will then need to be 'initially spatially subset' again (see the next topic.)

There's probably a more elegant way of doing this involving creating a database out of the new csv, then joining this onto the subset databases directly, but I haven't worked out the practicalities.


#### Initial spatial sub-setting

In principle there are lots of ways of doing this. Here, use `ogr2ogr` to split the main database into four quadrants according to polygons covering the expected GrIS runoff limit. It isn't very elegant or quick, but it works. The ogr2ogr command is as follows:

```bash
for region in SW NW NE SE; do
echo ${region}
ogr2ogr $L0lib/landsat_gs_GrIS_${region}.sqlite filter_database.vrt -f SQLite -dsco SPATIALITE=YES -dialect sqlite -sql "SELECT landsat.SCENE_ID, landsat.PRODUCT_ID, landsat.SPACECRAFT_ID, landsat.SENSOR_ID, landsat.DATE_ACQUIRED, landsat.COLLECTION_NUMBER, landsat.COLLECTION_CATEGORY, landsat.SENSING_TIME, landsat.DATA_TYPE, landsat.WRS_PATH, landsat.WRS_ROW, landsat.CLOUD_COVER, landsat.BASE_URL,  landsat.geom, gris_quadrants_runoff.name, ST_Area(ST_Intersection(ST_Transform(landsat.geom, 3413), ST_Transform(gris_quadrants_runoff.geometry, 3413))) as area FROM landsat INNER JOIN gris_quadrants_runoff ON ST_Intersects(landsat.geom, gris_quadrants_runoff.geometry) WHERE gris_quadrants_runoff.name='${region}' AND landsat.EAST_LON > landsat.WEST_LON";
done
```

This ogr2ogr call relies on three files:

1. `filter_database.vrt`. This GDAL VRT file contains links to the main Google Cloud Landsat Database and to the shapefile dividing the GrIS runoff limit into quadrants. Make sure the absolute paths contained in this file are correct.
2. The Google Cloud Landsat Database.
3. The quadrants shapefile (created manually).

Contents of `filter_database.vrt`:

```xml
<OGRVRTDataSource>
    <OGRVRTLayer name="landsat">
        <SrcDataSource>/home/geoscience/nobackup_cassandra/L0data/landsat_gs.sqlite</SrcDataSource>
    </OGRVRTLayer>
    <OGRVRTLayer name="gris_quadrants_runoff">
        <SrcDataSource>/home/tedstona/TEMPLATES/river_detection/gris_quadrants_runoff.shp</SrcDataSource>
    </OGRVRTLayer>
</OGRVRTDataSource>
```

Then use the quadrant-specific database to further refine the query and submit a job for processing.

Approach to consider in future: add WRS-2 polygons to database instead of building unique polygons for each Landsat product. Then do a join on Path/Row.



#### Build SQL query to define job

Do this in a template file. 

Specify the database as one of the quadrant-specific databases created above.

For the Greenland runoff limits work, the ice sheet was split into four quadrants, and then each quadrant was run via two jobs: one at <5% cloud cover, the other at 5>=<50% cloud cover. Given three satellites (LT5, LE7, LC8) this yields 6 jobs per quadrant, 24 jobs total.


#### Download imagery

Use `run_landsat_query.py` in `-download` mode. Remember that the workflow is designed to handle only one specific satellite sensor at a time, so the template file that you use with `run_landsat_query.py` has to uniquely specify a single sensor in the SQL query.

Use `run_landsat_query.py` to examine Landsat archive availability according to the SQL query specified in the `template` file, to download Landsat imagery and to check what imagery is already available on local storage. Run it before executing the rest of the workflow to verify that the query is returning the expected selection of products. 

If `$L0lib` is not the same as `$L0data`, for example if the job is to be executed using the beo05 flash storage, then execute `run_landsat_query.py <job> -stage` to copy the data returned by the query onto flash storage. You must have downloaded the data first! 



### Execute main workflow

`run_workflow.py` controls the main image processing (pre-processing, gabor filtering, path opening, thresholding). It is implemented as an EO-Learn pipeline. It creates a job ID associated with this execution of the workflow, corresponding to a new directory in `$PROCESS_DIR`  where all outputs are saved, together with a copy of the `.template` file renamed as `.job` file.

`run_workflow.py` can be run in testing mode by supplying the parameter `-test`. In this configuration, only the first product returned by the query is processed, and it is returned to the interactive namespace as `first_result`. Optionally, a restricted bounding box may also be specified in the `.template` file. This is useful for executing the workflow much more quickly than running an entire Landsat product. See Parameters section below for information on setting bounding box.

On `beo05.unifr.ch` this part of the workflow is run as a batch job, submitted to the job scheduler. Instead of executing `run_workflow.py` directly, run rivers_job_beo05.py:

```bash
    $ qsub_prep_workflow.py $TEMPLATES/my.template
```

This script generates a shell script suitable for submission with `qsub.` Note that you will need to modify `qsub_prep_workflow.py` to your own installation path information.

Testing suggests that Landsat-5 products need a minimum of 10 Gb per image process to execute successfully.


#### Re-running some steps

**IMPORTANT** If using the template file option to export automatically to the desired EPSG then this functionality does not work. It only works if the previously saved steps were output in the native projection of each Landsat product.

`run_workflow.py` offers basic partial re-running functionality. This is useful if you wish to re-run some later steps of the workflow and have saved outputs from some useful previous stage. There are a few manual steps needed to set this up.

1. This functionality is based on a modified .job file. So, go to `$PROCESS_DIR/<job_id>` and first create a backup copy of the .job file.
2. Edit the .job file as follows: 
  - Delete tasks to repeated from chain->tasks. NOTE: retain `load_single_band` if you have kept tasks that rely on metainfo loaded into the patch (e.g. MultiCloudMaskTask). Ensure that main->bands is set to an appropriate value.
  - Add task `restart_from`.
  - Add key: chain->restart_from_step. Specify this as the name of the output step to begin from, e.g. 'opening'.
  - Add key: chain->restart_from_number. This is the sequence number `S<n>` in the output filenames.
  - Delete tasks from the chain->save and chain->save_masks lists appropriately.
3. Run workflow using `-j` option as usual. If running on a qsub system like beo, you can use the `qsub.*.sh` file found in `$PROCESS_DIR/<job_id>` without any modifications.
4. Afterwards, remember to restore the backup copy of the .job file.


#### Job restart functionality

If the job halted unexpectedly during processing, then after debugging use `-restart` to run the remaining products which did not execute previously.


#### Retrying failed products

This option is useful if a full EOExecution completed but individual products/patches failed (e.g. as indicated in the EOExecution report). In these circumstances, the failed products are listed in `<job_id>.failures`. To retry the failed products, add `-retry_failures` to the `run_workflow.py` call.




### Post-processing to retrieve runoff limit of each product

#### Background on outlier removal
We cull erroneous runoff limit detection at several points:

1. Initial std. dev. based thresholding of the path-opened Gabor-filtered image. This is an early step in `ppj_hydro_features.py`.
2. During `ppj_hydro_features.py`, statistics for each label are generated, which are then used as criteria for culling in `ppj_runoff_limit.py`.
3. During `ppj_runoff_limit.py`, once undesired hydrological features are culled, we generate information about the proximity of each chosen runoff limit to (a) the path row and (b) clouds. See product specification for more details.
4. `ppj_filter_rlims.py` uses the master map approach to produce merged runoff limits for the whole job.

#### Runoff limit retrieval
From this point, processing chain scripts take the Job ID generated by `run_workflow.py`.

1. run `ppj_warp.py <job_id> -bands` to generate the 'feature'/NIR band in the target projection. They only are needed if `ppj_hydro_features.py` is set up to compute intensity image statistics.
* `ppj_hydro_features.py`. Finds and labels each 'hydrological' region, then produces a range of statistics which characterise each region. For Landsat-5 needs 4 Gb memory per product. If intensity image statistics are set up, then reprojected NIR band is needed. The MTL file is also required as `ppj_hydro_features.py` needs the radiance->reflectance information contained in this file.
* `ppj_runoff_limit.py`. Retrieves runoff limit position at the nominal full spatial resolution of the Landsat product. For a full, cloud-free scene, an estimate will be made every approx. 30 m. These estimates are saved to an SQLite database for further examination. For Landsat-5 needs 5 Gb per product. This script requires a `FilterFeaturesCritera.py` file to be present in the root of the job directory - usually the model version of this in this repository will be sufficient, but it does need to copied into the job directory manually.

On the `beo05` cluster, all of the scripts beginning `ppj` can be submitted to the queue scheduler by using  `qsub_prep_ppj.py` to create the necessary shell-script:

```bash
  $ qsub_prep_ppj.py <script_filename, no path> <job_id>
```

Note that the `-bands` option of `ppj_warp.py` must be added manually to the `qsub` file.

Note:  `ppj_runoff_limit.py` requires an ice mask that covers the entire loaded EOpatch, not just the central 'diamond' covered by a warped landsat image. This is made automatically if the variable `icemsk_in` in the job configuration file is set to a Greenland-wide product such as GIMP.


#### Further filtering

This requires master maps to be created.

Build using `create_master_maps.py` with a `merge.template` file.

After building master map for a quadrant, make a raster mask for it from the manually-defined GrIS runoff mask (`master_map_mask.shp`). For example:

```bash
gdal_rasterize $L0lib/master_map_mask/master_map_mask.shp master_map_mask_nw.tif -ts 20000 45000 -te -609000.000 -2292100.000 -9000.000 -942100.000  -burn 1 -ot Byte

gdal_rasterize greenland_icesheet_fix.shp $OUTPUT_DIR/merge_nw/gris_outline_nw.tif -ts 20000 45000 -te -609000.000 -2292100.000 -9000.000 -942100.000  -burn 1 -ot Byte

gdal_calc.py -A gris_outline_nw.tif -B master_map_mask_nw.tif --calc=A-B --outfile=mask_combined_nw.tif
```

Replace target extent and dimensions with the values gained from the relevant VRT master map file for the quad.

Or, this could be done automatically within create_master_maps.py and the actual maps masked already, which would save needing to run ppj_filter_rlims with the -mask option.

`create_master_maps.py` uses the final outputs of `run_workflow.py`, so can be run anytime after `run_workflow.py` has finished. It does however need to be run using all the jobs which compose the quadrant of the ice sheet under examination. Options for this are specified in a `merge.template` file. See section below for more information.

Once `create_master_maps.py` has run then run `ppj_filter_rlims.py`. Remember to run this with the `-mask` option!

Once `create_master_maps.py` has been run on all quadrants, run `build_gris_vrts_annual_counts.py` to make study-wide VRTs of annual counts.


#### DEM and Masks used in post-processing

Using GIMP DEM and IceMask at 30 m resolution.
This requires building of 30 m tiles into gdal virtual rasters as follows...

https://gis.stackexchange.com/questions/200407/merging-gdal-tiles-and-crop-via-bounding-box

```bash
gdalbuildvrt gimpdem_30m.vrt 30m/*tif
gdalwarp -t_srs EPSG:3413 gimpdem_30m.vrt gimpdem_30m_3413.vrt
```

And the same for the GIMP masks.

Set the config file keys `postprocess->dem_in` and `postprocess->mask_in` to these VRT files.

Note that these VRT files can only be employed once the Landsat product in question has been warped into the common projection.


### Assessment of potential spatio-temporal coverage

The first step here is to produce pixel-based counts of the number of image acquisitions per year (only for the months of interest). Use `create_master_maps.py` in `-yearly`, `-count_only` mode. See the script header for more information.


### Optional: per-image checks of runoff limit

Use `examine_rl.py` to plot extract runoff limits on top of their parent products. This script is *not* part of the workflow needed to arrive at the final results!


### Collating all runoff limits: PostgreSQL database

Requires various pre-requisites, they can all be installed within conda environment (PostgreSQL itself included).

Boilerplate code for DB creation can be found in `db_create.sql`.

Use `db_add_rlims.py` to add the filtered points CSV file of each job into the database.

The DB also has the 100 km analysis boxes stored in a table, which allows for easy spatial querying. 


### Note on -chunk option

This is an option that was trialled for a couple of weeks during I/O problems. It has not been removed systematically from where it was implemented, but best not to try to use it.



## WORKFLOW PARAMETERS

The section titles below correspond to their section names in the template/job configuration files.

### `main` - General options

- `sensor`. One of MSS, TM, ETM, OLI.
- `satellite`. One of LT5, LE7, LC8.
- `bands`. A list of bands to process, e.g. `[4,5]`.
- `wrs2`. Only needed if running Landsat imagery. Absolute path of shapefile containing WRS-2 polygons.


### `testing` - Options used during chain testing

This is an optional section.

- `bbox`. A decimal latitiude/longitude (EPSG:4326) defining a bounding box within which to constrain processing of the single product that `run_workflow.py` processes when run in `-test` mode. Provide in the order `[min_x, min_y, max_x, max_y]`.


### `query` - Controlling which products to process

- `db`. Full path to SQLite database file.
- `table_name`. Optional-ish. If using database generated by `create_landsat_spatialite_db.py` then the table name is `landsat`. If using the database after running it through `ogr2ogr` for spatial sub-setting, the table name is `select`.
- `q1..n`. Multiple SQL query components which are automatically joined together by `AND`.
- `spatial`. An optional spatial query string making use of SpatialLite functionalities.


### `download` - What to download

- `bands`. Specify both the bands and anything else you want to download. Generally, this needs to look like: `[4, "MTL", "BQA"]` if we are working only with Band 4. Providing MTL and BQA options means that the MTL.txt and BQA.tif files also get downloaded, which are essential for use later on in the chain.


### `chain` - Tasks to run in main processing chain

- `tasks`. A list of tasks to run, in the order that they need to be executed. Place the name of each task in quote marks.
- `save`. A list of tasks to save DATA outputs from.
- `save_masks`. A list of tasks to save MASK outputs from.
- `remove_previous`. True or False, if True then each processing step will remove the results of the previous processing step from memory.
- `warp`. If True, immediately warp outputs of run_workflow to target projection as defined in the `warp` section (see below). By default, when True, the files in the old (Landsat) projection are deleted immediately and only newly-warped files are retained.


### `filt_med` - Median image filtering

Use this step to remove 'speckle' noise, for example seen in some scan lines of Landsat TM data, from raw imagery prior to other steps such as FFT. A small neighbourhood, 3-5 px, should suffice for this purpose. 

- `size`: size in pixels of square "selem" neigbourhood element for skimage.filter.median.


### `hdfft` - High-pass FFT denoising

Contains task-specific options for the "hipass" task.

- `data`. The data label to denoise. **Is this still used?**
- `f1`. The frequency at which ramp-up begins.
- `f2`. The frequency by which the ramp-up is unity.


### `obnlm` - Denoising with 'Optimised Bayesian Non-Local Means'

Contains task-specific options for the "OBNLM" task.

- `m1`. Spatial extent in which to search for potentially similar image patches.
- `m2`. Patch size for conducting similarity calculations.
- `smooth`. Determines smoothing intensity.


### `gabor` - Gabor filtering

- `W`. Minimum width in pixel space of rivers to identify (1 or greater). Note that if this step is run on the outputs of FFT denoising then the size of this parameter corresponds to river width in the FFT result, *not* in the original unfiltered imagery! (Hi-pass FFT tends to "increase" apparent channel width).
- `ndirs`. Number of directions over which to run filter.
- `mask_borders`. Number of pixels along each border/margin to mask out after the Gabor filtering has completed.


### `pathopen` - Path Opening

Contains task-specific options for the path opening task.

- `length`.  The minimum length of path that will be detected by the algorithm and assessed as a candidate for path opening.
https://diplib.github.io/diplib-docs/group__morphology.html#gae750966565f85da6dc32dd413e179826


### `thresholding_stddev` - Thresholding using Std. Dev. approach

Contains task-specific options for thresholding with standard deviation of image.

- `number_std_dev`. The number of standard deviations to add to the image mean to use as the threshold.


### `clouds` - Cloud mask creation and application

Contains task-specific options for MultiCloudMaskTask.

- `mask_types`. A list of the cloud masks to combine into one cloud mask. Options are `cloud`, `cloud_shadow`, `cloud_swir`.
- `swir_band`. If using `cloud_swir` mask, specify the Band number which corresponds to the SWIR band here.
- `swir_threshold`. If using `cloud_swir` mask, specify the reflectance threshold to apply to the SWIR band. Values greater than this value will be treated as cloud.
- `dilate_by`. If using `cloud_swir` mask, dilate the mask by this many pixels.


### `warp` - Options for warping outputs from Landsat projection to custom projection

- `to_epsg`. Give the numeric EPSG code here, e.g. 3413, to warp to.
- `to_xyres`. Give the spatial resolution to force images to conform to, in metres. E.g. 30.


### `postprocess` - Options used by post-processing tools

The files specified by the `*_in` parameters need to be located where you want to read them from - on beo05 this is probably on /flash rather than /home.

- `dem_in`. Full path to source DEM.
- `aspect_in`. Full filepath to a geotiff of aspect generated from a DEM, can be a much larger area than the tile - will be subset automatically during ingestion.
- `mask_in`. Full path to ice mask.

- `cleanup`. If True, clean up the interim files used in processing.
- `nde`. The number of dilate-erode iterations used to close holes by `ppj_hydro_features.py`.
- `label_properties`. A list of properties about each label to return, see https://scikit-image.org/docs/dev/api/skimage.measure.html#skimage.measure.regionprops for full list. Note that the `label_id` is returned automatically. 
  - If `region_intensity` (below) is specified, then `max_<name>`, `min_<name>` and `mean_<name>` will also be returned in `regions` dataframe.
  - `max_dem`, `min_dem`, `mean_dem` are labels always returned in `skel_regions` dataframe.
- `region_intensity`. Optional. If provided, this sets the image source that will be used to generate 'intensity' information of each labelled region during `hydro_features`. Available options are `B<n>`, e.g. `B4`. Note that the image source must be in the same EPSG as the rest of post-processed imagery, so in the case of 'raw' image bands you will need to run `ppj_warp.py` in `-bands` mode to generate the requisite files.

#### Filtering in `runoff_limit`
`runoff_limit` expects to find a Python file named `FilterFeaturesCriteria.py` located in the `$OUTPUT_DIR/<job_id>` directory. This file is *not* copied here automatically, you have to explicitly create it yourself. If the file is not present then `runoff_limit` will fail.

The file is run directly through the Python `exec()` command. See the main repository for an example script, but the essence is that:

1. Two dataframes are available: `regions` and `regions_skel`.
2. These dataframes have all the keys specified by `postprocess->label_properties`. 
3. In `region_intensity` is specified, `regions` will have the relevant labels.
4. You are generating a list of image labels/regions to remove, by querying the dataframes to return the regions you want to get rid of. 
5. You add to the list of labels to remove by calling `self._queue_labels` with the filtered dataframe.
5. Thus, an example line of code is as follows:

```python  
self._queue_labels(skel_regions[(skel_regions['area'] < 34)])
```



## MERGE PARAMETERS

See `merge.template` for an example. These parameters mainly define the processing done in `create_master_maps.py`.

### `default`
- `tile_size_px`. The desired dimensions of each tile of the merged product. (They are subsequently all linked together by a VRT file).
- `native_res`. The native resolution in metres of the processed Landsat products.
- `epsg`. The EPSG code of the projection to merge onto.
- `file_suffix`. The file suffix of the processed Landsat products to merge.
- `qa_file_suffix`. The quality information (QA) file suffix of the processed Landsat products to merge.
- `mask_in`. Path to mask - usually a VRT of the whole ice sheet.

### `master_map`
- `data_threshold`. Minimum number of occurences needed for a pixel to be retained.
- `fill_holes_threshold`. Holes smaller than this number of pixels are filled prior to skeletonisation.

### `query`
- `db`. The path to the database of Landsat products (correct for the quadrant).
- `sql`. The SQL query to run to find the products which compose the merge area. This should be left as-is - various parameters are filled in dynamically by `create_master_maps.py`.
- `main_cc_thresh`. In principle set this to the most cloudy threshold used.
- `jobs_cc_thresh`. The threshold which defines which images were processed in the "clear sky" job and which were processed in the "cloudy sky" job.

### `quad`
- `shp`. The shapefile which defines the extents of each quad.
- `process`. The ID of the quadrant to process.

### `jobs`
Contains an entry per satellite, each entry consists of a Python list of constituent jobs, e.g.:
    
    LANDSAT_5=["clear.job", "cloudy.job"]


### `filtering`
- `master_map_type`. Always `skel`.
- `npx_from_edge`. Number of pixels away from edge at which to clip retrievals.
- `max_elev`. Maximum elevation at which to retain retrievals.
- `clds_row_dist_thresh`. Minimum pixels distance required between a retrieval and a flagged cloud pixel.
- `search_radius_m`. The radius in metres over which to search for/include runoff limit retrievals relative to the master hydrological network.
- `search_radius_epsilon`. The degree of "roughness" allowed by the k-d search operation.
- `elev_std`. Number of standard deviations away from elevation mean allowed to be retained.

### `rcm`
This section is no longer actively used by any part of the analysis.


## INSTALLATION

Clone this repository to production machine.

The system was created within a miniconda-3-based Python environment. Main dependencies beyond the basic Python ecosystem:

- GDAL
- Pandas
- GeoPandas
- EO-Learn
- SentinelHub

```bash
conda install -c conda-forge python=3.7 ipython matplotlib pandas geopandas eo-learn-core eo-learn-features eo-learn-geometry eo-learn-io eo-learn-mask eo-learn-visualization rasterio scikit-image sentinelhub georaster requests 

pip install spatialite
```

Note also that PyDIP is compiled against the Python version in the active conda environment! If you change/upgrade the Python version later then PyDIP will (probably) stop working.

For `ppj_runoff_limit.py`, the `RichDEM` package is also needed.


### User environment setup

The following environment variables are needed:

* `L0lib`. Path to the directory holding the 'Level-0' data. `run_landsat_query.py` will automatically make Landsat-sensor sub-directories here. Each product will be saved to its own directory, within the relevant sensor directory. 
* `L0data`. In simple cases this can be set the same as `L0lib`, but for more complex cases such as using beo05 flash storage, set this to where the Level-0 data need to be staged to. You can achieve the staging using `run_landsat_query.py <template/job file> -stage`.    
* `OUTPUT_DIR`. Every job will automatically create a directory within this location.
* `WRKDIR`. Location where execution is to be carried out. For simple cases this can be set to the same as `OUTPUT_DIR`. To make use of 'chunked' execution where data are copied to the compute node's local hard disk prior to execution, set this appropriately - e.g. `/scratch/tedstona`. 
* `FINAL_DIR`. For convenience, set this if `OUTPUT_DIR` is not the same as the final destination for your results. For example on beo05, `OUTPUT_DIR` needs to be set to /flash storage, but the final location of the data is back in /home.
* For using beo flash storage: you must set OUTPUT_DIR to the flash storage! Otherwise, final outputs will be saved to cluster storage, which won't help...
* `EOLEARN_WORKERS`. Used to control number of workers used by EO-Learn. On beo05, set this to correspond to the number of slots being requested. If you are using the helper script `qsub_prep_workflow.py` then this variable is set for you automatically.
* `TEMPLATES`. The folder holding the `.template` files used to initialise the main workflow and which become `.job` control files.
* `JOBLIB_TEMP_FOLDER`. This controls where `joblib` dumps temporary data. Usually we wouldn't need to touch this, but on `beo05` this defaults to `/dev/shm`, which potentially fills quickly. On `beo05` it is adequate to set this to `/scratch/<username>`, taking advantage of local node HDD scratch space. Further info: `joblib: _memmapping_reducer.py`. 
* Also add the location of this repository of scripts to your `PATH` and `PYTHONPATH`.

It's recommended to set all these up in a shell script which can be executed (`source`d) whenever needed.

If using a system with a batch job scheduler, make sure to update the paths coded into the `qsub_*.py` files to reflect your installation environment.


### DIPLib

The main workflow requires DIPLib to be compiled on host machine. See DIPLib readme for minimum gcc/g++ versions. 

On `beo05`:
```bash

cd <diplib_source>

module load gcc73/gcc73

cmake . -DCMAKE_INSTALL_PREFIX=/home/geoscience/cassandra/lib/dip -DPYDIP_INSTALL_PATH=/home/geoscience/cassandra/lib/python3.7/site-packages/ -DCMAKE_C_COMPILER=gcc -DCMAKE_CXX_COMPILER=g++ -DDIP_BUILD_JAVAIO=Off

```

Replace `python3.7` in path with the version of Python you are installing bindings against. 

Note that the `-DCMAKE_C_COMPILER` and `DCMAKE_CXX_COMPILER` flags are important on beo05, otherwise the build fails as DIPLib will not find the correct compilers.

Then `make`, `make install` as usual.


#### FFTW3

I tried to install with FFTW3 dependency but it is unclear how to do this because of environment variables.


#### DIPLib Python bindings

Add `[INSTALLATION_PATH]/lib/python3.6/site-packages/` to `PYTHONPATH` env. variable.




## LOG FILES

Logging occurs only under the following conditions:

* If an EO-Execution pipeline is executed (i.e. `run_workflow.py`, `ppj_hydro_features.py`, when not in `-test` mode). In these cases, EO-Learn exports an `eoexecution` report folder within the job folder, containing an HTML report file and individual execution logs for each patch.
* For all jobs sent to the job scheduler (e.g. on beo05): stdout and stderr are redirected to text files located in the job folder, named `<job_id>`.



## SPECIFIC INFO to UniFR beo05

## Database connection

* Located in `/flash/tedstona/rlims`
* Running on Psql 12.3, installed within `geospa` environment
* The database server must be running on the compute node for analysis to work
* Execute: `pg_ctl -D /flash/tedstona/rlims start`
* Remember to `stop` the database server afterwards.


### Backing up/transferring from /flash to /home

Example:

```bash
rsync --dry-run -r -u -v --exclude=*eoexecution* /flash/tedstona/<job>/ /home/geoscience/nobackup_cassandra/river_detection/<job>  
```

### Re-running runoff_limits with new poly mask

```bash
export TMP_JOB="LC8_NW.20201001-123212.job"

mkdir $OUTPUT_DIR/${TMP_JOB}/
cd $OUTPUT_DIR/$TMP_JOB
rsync $FINAL_OUTPUTS/$TMP_JOB/ . -v -r  --include='*/' --exclude='*LABELS_FILT.tif' --exclude='*S1_hipass*' --exclude='*S2_obnlm*' --exclude='*S3_mask_cloud*' --exclude='*RLIM.csv' --exclude='*B4_EPSG3413*' --exclude='*filtered_MM.csv' -m --progress 

echo 'master_map_mask=/home/geoscience/nobackup_cassandra/L0data/master_map_mask/master_map_mask.shp' >> $TMP_JOB

qsub qsub_${TMP_JOB}.runoff_limit.sh
```

### Commands for partial backup

To copy just the new RLIM files over to /home, into a new directory:
```bash
export TMP_JOB='LC8_NW.20201001-123212.job'
mkdir $FINAL_OUTPUTS/${TMP_JOB}.premask
rsync $OUTPUT_DIR/${TMP_JOB} $FINAL_OUTPUTS/${TMP_JOB}.premask/ --include="*/" --include='*RLIM*' --exclude='*' -v -r --progress
rm -rf $OUTPUT_DIR/${TMP_JOB}
```
This was used following re-running ppj_runoff_limit with pre-masking using the manually-drawn, master-map-derived shapefile mask to remove aquifer type areas from the raster images, prior to them then being scanned for the runoff limits.

This was done in October 2020, only on NW jobs. SW jobs have not been re-processed, i.e. the mask was only applied in -posthoc mode with ppj_filter_rlims.py.



### Manual transfer of selected files from /home to /flash

E.g. to copy over stage 3 outputs only, for instance to regenerate master maps.

```bash
cd $OUTPUT_DIR/<job>/
rsync $FINAL_OUTPUTS/LE7_SW.20200604-105019.job/ . -v -r  --include='*/' --include='*S3_mask_cloud*.tif' --exclude='*' -m --progress --dry-run
```

Remember that you'll probably need to QA fill tiffs too!


### CPU and memory resource management

The salient issues are:

* Number of cores and amount of memory to request using job scheduler. On beo05, generally each node has 2x CPUs, each of which has 8 cores. Each core is dual-threaded, so 16 threads available in total.
* EOExecutor n.workers and multiprocess flag. When multiprocess=True, n.workers controls the number of processors over which the workflow is allocated. When multiprocess=False, workflow is distributed only among threads.
* Potential conflict with this: sub-EOExecutor worker/thread management. E.g. the Gabor implementation uses a multiprocessing.Parallel pool. In April 2020 I had substantial problems associated with this: with Parallel threads=2 but also EOExecutor n.workers=2 *and* multiprocess=False, joblib was trying to dump out mem-map pkl files to scratch but it kept resulting in a FileNotFoundError for most patches after the first 10 or so. Eventually, resulting RAM leakages would cause complete qsub job failure. This suggests there was some thread-level conflict, though I'm not sure of the details. Changing Parallel threads=1 solved the memory leak.
* I/O (Input-Output). If using beo05 then it is critical to address this properly. The main process must be 'ioniced' as follows: `ionice -c 3 <script>`. This MUST run at the 'highest' level, e.g. `ionice -c 3 python run_workflow.py -j <job_id>`. It has no effect if implemented within one of the Python control scripts. The command is added to qsub shell scripts generated by the `qsub_prep*` scripts automatically.


### Bugs in DIPLib v3.0-beta.1

According to `make check` on 06.11.2019.

```
===============================================================================
/home/geoscience/cassandra/lib/diplib-3.0-beta.1.src/src/analysis/findshift.cpp:419:
TEST CASE:  [DIPlib] testing the FindShift function

/home/geoscience/cassandra/lib/diplib-3.0-beta.1.src/src/analysis/findshift.cpp:450: ERROR: CHECK( std::abs( result[ 0 ] - shift[ 0 ] ) < 0.15 ) is NOT correct!
  values: CHECK( 60.3405074584 <  0.15 )

/home/geoscience/cassandra/lib/diplib-3.0-beta.1.src/src/analysis/findshift.cpp:451: ERROR: CHECK( std::abs( result[ 1 ] - shift[ 1 ] ) < 0.15 ) is NOT correct!
  values: CHECK( 44.0518955165 <  0.15 )

===============================================================================
/home/geoscience/cassandra/lib/diplib-3.0-beta.1.src/src/distance/separable_dt.cpp:224:
TEST CASE:  [DIPlib] testing the distance transform

/home/geoscience/cassandra/lib/diplib-3.0-beta.1.src/src/distance/separable_dt.cpp:308: ERROR: CHECK( dip::MaximumAbsoluteError( gt, out ) == doctest::Approx( 0.0 ) ) is NOT correct!
  values: CHECK( 0.0267949224 == Approx( 0.0 ) )

/home/geoscience/cassandra/lib/diplib-3.0-beta.1.src/src/distance/separable_dt.cpp:311: ERROR: CHECK( dip::MaximumAbsoluteError( gt, out ) == doctest::Approx( 0.0 ) ) is NOT correct!
  values: CHECK( 0.0267949224 == Approx( 0.0 ) )

[dip::testing::CompareImages] Maximum absolute error = 133.914 > 0.015
===============================================================================
/home/geoscience/cassandra/lib/diplib-3.0-beta.1.src/src/geometry/interpolation.cpp:1112:
TEST CASE:  [DIPlib] testing dip::ShiftFT() against dip::Shift()

/home/geoscience/cassandra/lib/diplib-3.0-beta.1.src/src/geometry/interpolation.cpp:1120: ERROR: CHECK( dip::testing::CompareImages( shifted, shifted2, 0.015 ) ) is NOT correct!
  values: CHECK( false )

100 s
1 m
===============================================================================
[doctest] test cases:     70 |     67 passed |      3 failed |      0 skipped
[doctest] assertions:   2348 |   2343 passed |      5 failed |
[doctest] Status: FAILURE!
make[3]: *** [src/CMakeFiles/check] Error 1
make[2]: *** [src/CMakeFiles/check.dir/all] Error 2
make[1]: *** [src/CMakeFiles/check.dir/rule] Error 2
make: *** [check] Error 2
```

