"""
Generate a 'master map' of river locations for given quadrant.

However, this script also has a second mode: production of annual 'master maps'.
The principal use of this mode is not actually to to identify river locations
but instead to simply generate counts of cloud-free observations made annually
across the quadrant.

In both usage modes, this script uses the 'mask_clouds' output step of 
run_workflow.py.

For main 'master map' mode, only use these options:
- template
- retry_failures
- test

For 'annual count' mode, set:
- template
- -yearly <year>
- -only_count
- -cct (cloud cover threshold) - only needed if this is different to 'main_cc_thresh'
  set in the merge template. This is likely to be the case, as the main_cc_thresh
  defines the maximum cloud threshold value for images to be candidates for 
  inclusion in river master map. On the other hand, for the observation counts, 
  we want to include all images potentially available for runoff limit detection,
  so -cct should usually be set to the maximum cloud threshold used to choose
  the images processed by the main workflow.

Both modes work in image tiles. At the end of processing, VRT file(s) are 
generated to combine these tiles.

This script outputs its files to $OUTPUT_DIR/merge_{QUADRANT}/.

AT, September 2020

"""

import numpy as np
import os
import argparse
import configparser
import subprocess
import json

import geopandas as gpd
from sentinelhub import BBox, CRS

from eolearn.core import LinearWorkflow, EOExecutor, EOTask
from eolearn.io.local_io import ExportToTiff, ImportFromTiff
from eolearn.mask import masking
from eolearn.core.constants import *
from eolearn.core.eodata import EOPatch
import rasterio
from eolearn.mask import MaskFeature

from skimage.morphology import skeletonize, remove_small_holes

import run_landsat_query

################################################################################
parser = argparse.ArgumentParser(description='Create a master map of specified quadrant')
parser.add_argument('template', type=str, 
	help='str, path to template file which defines the merging process')
parser.add_argument('-retry_failures', dest='retry_failures', default=False,
	action='store_true',
	help='Only try to process tiles of master map which failed previously.')
parser.add_argument('-test', dest='test', default=False, 
	action='store_true', 
	help='Run for first tile of merge only.')
parser.add_argument('-yearly', dest='year', type=int, default=None)
parser.add_argument('-only_count', dest='only_count', default=False, action='store_true', help='only use with -yearly!')
parser.add_argument('-file_suffix', dest='file_suffix', type=str, default=None)
parser.add_argument('-cct', dest='cct', type=int, default=None)
args = parser.parse_args()


config = configparser.ConfigParser()
config.read_file(open(args.template, 'r'))


################################################################################
## Main settings 

# Define the tile size, in pixels, of the square tiles which will compose the master map.
TILE_SIZE_PX = int(config.get('default', 'tile_size_px'))

# The native resolution, in metres, of the input files.
NATIVE_RES = float(config.get('default', 'native_res'))

# EPSG code of the input data and for all processing.
EPSG = config.get('default', 'epsg')

# Corresponding QA product.
QA_FILE_SUFFIX = config.get('default', 'qa_file_suffix')

# Number of EOLearn workers over which to run EOExecutor.
try:
	N_WORKERS = int(os.environ['EOLEARN_WORKERS'])
except KeyError:
	print('EOLEARN_WORKERS not set, defaulting to 1 worker')
	N_WORKERS = 1


################################################################################
## Options

# Load the quadrants in which processing is being undertaken
quads = gpd.read_file(config.get('quad', 'shp'))

# Quadrant to process.
QUAD = config.get('quad', 'process')

jobs_ld = dict(config.items('jobs'))
jobs = {}
for key in jobs_ld.keys():
	jobs[key.upper()] = json.loads(jobs_ld[key])

# Load config file contents for each job composing this merge.
# At the moment this is only to determine the file ending - this depends on the n
# umber of outputs written by run_workflow.py
job_files = {}
av_out = {}
for job in jobs:
	for j in jobs[job]:
		tmp_job_conf = configparser.ConfigParser()
		fp = os.path.join(os.environ['OUTPUT_DIR'], j, j)
		tmp_job_conf.read_file(open(fp, 'r'))

		save_tasks = json.loads(tmp_job_conf.get('chain', 'save'))
		available_outputs = {}
		n = 1
		for t in save_tasks:
			available_outputs[t] = n
			n += 1

		av_out[j] = available_outputs
		job_files[j] = tmp_job_conf


base_sql = config.get('query', 'sql')
if args.year is not None:
	base_sql = base_sql + " AND strftime('%Y',DATE_ACQUIRED) = '" + str(args.year) + "'"

if args.cct is not None:
	cc_thresh = args.cct
else:
	cc_thresh = config.get('query', 'main_cc_thresh')

template_name = args.template.split('/')[-1].split('.template')[0]
save_path = os.path.join(os.environ['OUTPUT_DIR'], template_name)
if not os.path.exists(save_path):
	os.makedirs(save_path)
print('Saving to {s}'.format(s=save_path))


################################################################################
## Tasks specific to this processing approach.

class CreateTileTask(EOTask):

	def __init__(self, pixel_size_m=None, tile_dim_px=None,
		blank_features=[(FeatureType.DATA, 'sum'), (FeatureType.DATA, 'count')]):

		self.blank_features = blank_features
		self.pixel_size_m = pixel_size_m
		self.tile_dim_px = tile_dim_px

		# Tile side dimension in metres
		self.tile_size_m = self.pixel_size_m * self.tile_dim_px

		return


	def execute(self, min_x=None, min_y=None):
		
		max_x = min_x + self.tile_size_m
		max_y = min_y + self.tile_size_m
		roi = BBox(bbox=[min_x, min_y, max_x, max_y], crs=CRS(EPSG))	
		eopatch = EOPatch(bbox=roi)

		for feature in self.blank_features:

			feat = self._parse_features(feature)
			feat_type, feat_name = next(feat())
			eopatch[feat_type][feat_name] = np.zeros((1,TILE_SIZE_PX, TILE_SIZE_PX, 1))
														
		return eopatch



class AdditiveProductsTask(EOTask):

	def __init__(self, 
		sum_feature=(FeatureType.DATA, 'sum'),
		count_feature=(FeatureType.DATA, 'count')):
		
		self.sum_feature = self._parse_features(sum_feature)
		self.count_feature = self._parse_features(count_feature)


	def execute(self, eopatch, product_files):
		"""
		product_files : dict of dicts {<product_id>:{'data':<path_to_data>, 'qa':<path_to_qa>}

		"""

		for product_id in product_files:

			# Import the data

			if not(os.path.exists(product_files[product_id]['data'])):
				print('{p} not available, skipping'.format(p=product_id))
				continue

			try:
				eopatch = ImportFromTiff((FeatureType.DATA, product_id)).execute(
					filename=product_files[product_id]['data'], eopatch=eopatch)
			except rasterio.errors.RasterioIOError as err:
				print('Error loading {f} in AdditiveProductsTask ({e})'.format(f=product_id, e=err))
				continue

			# Import the QA info
			try:
				eopatch = ImportFromTiff((FeatureType.MASK, product_id+'_QA')).execute(
					filename=product_files[product_id]['qa'], eopatch=eopatch)
			except rasterio.errors.RasterioIOError as err:
				print('Error loading {f} QA in AdditiveProductsTask ({e})'.format(f=product_id, e=err))
				continue

			# Mask based on QA to remove edge features
			eopatch = masking.MaskFeature(
				(FeatureType.DATA, product_id, product_id), 
				(FeatureType.MASK, product_id+'_QA'), [1,], 
				no_data_value=-1).execute(eopatch)

			# Do addition
			sum_feature_type, sum_feature_name = next(self.sum_feature())
			eopatch[sum_feature_type][sum_feature_name] = eopatch[sum_feature_type][sum_feature_name] + \
				eopatch[FeatureType.DATA][product_id]

			# Do counting
			count_feature_type, count_feature_name = next(self.count_feature())
			eopatch[count_feature_type][count_feature_name] = eopatch[count_feature_type][count_feature_name] + \
				np.where(eopatch[FeatureType.DATA][product_id] > 0, 1, 0)

			# Remove the just-processed product
			eopatch.remove_feature(FeatureType.DATA, product_id)
			eopatch.remove_feature(FeatureType.MASK, product_id+'_QA')

		return eopatch



class CalculateMeanMasterMapTask(EOTask):

	def __init__(self, sum_feature=(FeatureType.DATA, 'sum'),
		count_feature=(FeatureType.DATA, 'count'),
		output_feature=(FeatureType.DATA, 'mean')):
		
		self.sum_feature = self._parse_features(sum_feature)
		self.count_feature = self._parse_features(count_feature)
		self.output_feature = self._parse_features(output_feature)
		

	def execute(self, eopatch):

		sum_feature_type, sum_feature_name = next(self.sum_feature())
		count_feature_type, count_feature_name = next(self.count_feature())
		out_feature_type, out_feature_name = next(self.output_feature())

		eopatch[out_feature_type][out_feature_name] =                          \
			eopatch[sum_feature_type][sum_feature_name] /                      \
			eopatch[count_feature_type][count_feature_name]

		return eopatch



class SkelMasterMapTask(EOTask):
	
	def __init__(self, data_threshold, fill_area_threshold,
		feature_in=(FeatureType.DATA, 'mean'),
		feature_out=(FeatureType.DATA, 'skel')):
		"""
		data_threshold : the value above which data pixels are treated as runoff
			features.

		fill_area_threshold : the maximum size of holes to fill, expressed as
			number of pixels.
		"""

		self.data_threshold = data_threshold
		self.fill_area_threshold = fill_area_threshold
		self.feature_in = self._parse_features(feature_in)
		self.feature_out = self._parse_features(feature_out)


	def execute(self, eopatch):

		feat_type, feat_name = next(self.feature_in())
		thresh = np.where(eopatch[feat_type][feat_name][0,:,:,0] >= self.data_threshold, 
							1, 0)

		filled = remove_small_holes(thresh, 
			area_threshold=self.fill_area_threshold)

		skel = skeletonize(filled)

		feat_type, feat_name = next(self.feature_out())
		eopatch[feat_type][feat_name] = np.reshape(skel, (1, skel.shape[0], skel.shape[1], 1))

		return eopatch




################################################################################
## Define tasks and workflow.


eo_tasks = {}
eo_tasks['task_create_tile'] = CreateTileTask(NATIVE_RES, TILE_SIZE_PX)
eo_tasks['task_add_products'] = AdditiveProductsTask()

eo_tasks['task_load_icemsk'] = ImportFromTiff((FeatureType.MASK, 'icemsk'), 
	config.get('default', 'mask_in'))
eo_tasks['task_mask_ice_count'] = MaskFeature(
	(FeatureType.DATA, 'count', 'count'),
	(FeatureType.MASK, 'icemsk'), [0,],
	no_data_value=0)

if not args.only_count:
	eo_tasks['task_mask_ice_sum'] = MaskFeature(
		(FeatureType.DATA, 'sum', 'sum'),
		(FeatureType.MASK, 'icemsk'), [0,],
		no_data_value=0)

	eo_tasks['task_calc_mean'] = CalculateMeanMasterMapTask()
	eo_tasks['task_skel'] = SkelMasterMapTask(
		float(config.get('master_map', 'data_threshold')),
		int(config.get('master_map', 'fill_holes_threshold')))

	eo_tasks['task_export_sum'] = ExportToTiff((FeatureType.DATA, 'sum'), 
		image_dtype=np.float32, compress='lzw', no_data_value=-1,
		folder=save_path)
	eo_tasks['task_export_mean'] = ExportToTiff((FeatureType.DATA, 'mean'), 
		image_dtype=np.float32, compress='lzw', no_data_value=-1,
		folder=save_path)
	eo_tasks['task_export_skel'] = ExportToTiff((FeatureType.DATA, 'skel'), 
		image_dtype=np.int8, compress='lzw', no_data_value=0,
		folder=save_path)

eo_tasks['task_export_count'] = ExportToTiff((FeatureType.DATA, 'count'), 
	image_dtype=np.float32, compress='lzw', no_data_value=-1,
	folder=save_path)

workflow = LinearWorkflow(*[eo_tasks[t] for t in eo_tasks])



################################################################################
# Query data and run pipeline.

# Load config files for each job ...


q = quads[quads.name == QUAD]
q_ps = q.to_crs('EPSG:{e}'.format(e=EPSG))

bounds = round(q_ps.bounds.iloc[0] / 100) * 100

tile_size_m = TILE_SIZE_PX * NATIVE_RES
x_mins = np.arange(bounds.minx, bounds.maxx, tile_size_m)
y_mins = np.arange(bounds.miny, bounds.maxy, tile_size_m)

if args.retry_failures:
	print('Retry Failures mode')
	print('-'*80)

workflow_args = []
for min_x in x_mins:
	for min_y in y_mins:

		patch_args = {}

		# Find all imagery for tile.
		# This SQL query is hard coded here based on criteria used during the job design phase.
		max_x = min_x + tile_size_m
		max_y = min_y + tile_size_m
		
		if args.year is not None:
			file_out_basic = 'master_{q}_{year}_x{x}_y{y}'.format(q=QUAD,x=min_x, y=min_y, year=args.year)
		else:
			file_out_basic = 'master_{q}_x{x}_y{y}'.format(q=QUAD,x=min_x, y=min_y)
			

		if args.retry_failures:
			out_path = os.path.join(save_path, '{b}_mean.tif'.format(b=file_out_basic))
			if os.path.exists(out_path):
				continue
			else:
				print(min_x, min_y)

		sql_q = base_sql.format(cc=cc_thresh, x1=min_x, y1=min_y, x2=max_x, y2=max_y, epsg=EPSG)
		products = run_landsat_query.execute_query(args.template, sql_q)
		print(min_x, len(products))

		product_paths = {}
		for ix, p in products.iterrows():

			if p.SPACECRAFT_ID not in jobs.keys():
				continue

			if p.CLOUD_COVER < int(config.get('query', 'jobs_cc_thresh')):
				job = jobs[p.SPACECRAFT_ID][0]
			else:
				job = jobs[p.SPACECRAFT_ID][1]
			
			product_path = os.path.join(os.environ['OUTPUT_DIR'], 
				job, 
				p.PRODUCT_ID)

			if args.file_suffix is not None:
				file_fn = p.PRODUCT_ID + args.file_suffix
			else:
				file_fn = p.PRODUCT_ID + '_S%s_mask_cloud_EPSG%s.tif' %(av_out[job]['mask_cloud'], job_files[job].get('warp','to_epsg'))
			
			filt_file = os.path.join(product_path, file_fn)

			qa_file = os.path.join(product_path, p.PRODUCT_ID + QA_FILE_SUFFIX)

			product_paths[p.PRODUCT_ID] = dict(data=filt_file, qa=qa_file)


		# Inputs	
		patch_args[eo_tasks['task_create_tile']] = {'min_x': min_x, 'min_y': min_y}
		patch_args[eo_tasks['task_add_products']] = {'product_files' :product_paths}

		# Outputs
		patch_args[eo_tasks['task_export_count']] = {'filename': '{f}_count.tif'.format(f=file_out_basic)}
		
		if not args.only_count:
			patch_args[eo_tasks['task_export_sum']] = {'filename': '{f}_sum.tif'.format(f=file_out_basic)}
			patch_args[eo_tasks['task_export_mean']] = {'filename': '{f}_mean.tif'.format(f=file_out_basic)}
			patch_args[eo_tasks['task_export_skel']] = {'filename': '{f}_skel.tif'.format(f=file_out_basic)}

		workflow_args.append(patch_args)



################################################################################
# Run workflow

if args.test:
	results = workflow.execute(workflow_args[0])
	first_result = list(results.values())[0]
else:
	executor = EOExecutor(workflow, workflow_args, save_logs=True, 
		logs_folder=save_path)
	

	results = executor.run(workers=N_WORKERS, multiprocess=False)
	executor.make_report()
	report_fn = executor.get_report_filename()
	print('Report saved to:\n' + report_fn)

	# Log any failures so that they can be reprocessed.
	failures = executor.get_failed_executions()



################################################################################
	# Create vrts

	if args.year is not None:
		fn_vrt = 'master_{q}_{y}_count.vrt'.format(q=QUAD, y=args.year)
		fn_tiffs = 'master_{q}_{y}_x*_count.tif'.format(q=QUAD, y=args.year)
	else:
		fn_vrt = 'master_{q}_count.vrt'.format(q=QUAD)
		fn_tiffs = 'master_{q}_x*_count.tif'.format(q=QUAD)

	cmd = 'gdalbuildvrt {s}/{fnv} {s}/{fnt}'.format(s=save_path, fnv=fn_vrt, fnt=fn_tiffs)
	subprocess.check_output(cmd, shell=True)

	if not args.only_count:
		cmd = 'gdalbuildvrt {s}/master_{q}_mean.vrt {s}/master_{q}_*_mean.tif'.format(s=save_path, q=QUAD)
		subprocess.check_output(cmd, shell=True)

		cmd = 'gdalbuildvrt {s}/master_{q}_skel.vrt {s}/master_{q}_*_skel.tif'.format(s=save_path, q=QUAD)
		subprocess.check_output(cmd, shell=True)



