""" 

Implement Discrete Fourier Transform for 2-d matrices [images] as EO-Learn task.

Based on Yang et al. 2017 approach, which uses 2DSpecTools-v1.1.
Combines only the relevant parts of these functions:

> SpecFilt2D
	> fft2D (to generate 'freq_mat')
	> Make2DFilt
		> Make2DFilt
		> Gaussian

Andrew Tedstone, 15 Nov 2019

"""


from eolearn.core.constants import *
from eolearn.core import EOTask
import numpy as np
import gc

class TwoDFiltTask(EOTask):
	"""
	Probably still more cleaning up/data type conversion to do here

	"""

	def __init__(self, band, feature_name, filter_type, filter_freq,
		remove_previous=False, invert=False):
		self.band = band
		self.feature_name = feature_name
		self.filter_type = filter_type
		self.filter_freq = filter_freq
		self.remove_previous = remove_previous
		self.invert = invert

	def execute(self, eopatch):
		
		# Pixel size
		dx = eopatch.meta_info['xres']
		dy = eopatch.meta_info['yres']

		# Image size, no padding
		Lx = eopatch.data[self.band].shape[2]
		Ly = eopatch.data[self.band].shape[1]

		# Calculate the frequency increments
		dfx = 1/(dx*Lx)
		dfy = 1/(dy*Ly)

		# Create a matrix of radial frequencies
		xc = Lx/2+1
		yc = Ly/2+1
		# matrix indices of zero frequency --> matrices of column and row indices
		cols, rows = np.meshgrid(np.arange(0,Lx),np.arange(0,Ly)) 
		# frequency matrix
		freq_mat = np.sqrt((dfy*(rows-yc))**2 + (dfx*(cols-xc))**2) 
		cols = None
		rows = None

		trans = eopatch.data[self.band][0,:,:,0]
		if self.invert:
			trans = np.invert(trans)
			
		trans = np.fft.fft2(trans)
		M = np.fft.fftshift(trans)

		F = Make2DFilt(freq_mat, self.filter_freq, self.filter_type)

		result = np.real(np.fft.ifft2(np.fft.ifftshift(M*F)))

		if self.remove_previous:
			eopatch.remove_feature(FeatureType.DATA, self.band)

		# Add to patch
		result = np.reshape(result, (1, result.shape[0], result.shape[1], 1))
		eopatch.add_feature(FeatureType.DATA, self.feature_name, result)
		gc.collect()

		return eopatch




def Make2DFilt(fmat, f, filttype):
	if filttype == 'lowpass':
	        flo = f[0] 
	        fhi = f[1]
	        mu = flo
	        sigma = np.abs(fhi-flo)/3
	        F = Gaussian(fmat, mu, sigma)
	        F[fmat < flo] = 1       

	elif filttype == 'highpass':
	        flo = f[0]
	        fhi = f[1]
	        mu = fhi
	        sigma = np.abs(fhi-flo)/3
	        F = Gaussian(fmat, mu, sigma)
	        F[fmat >= fhi] = 1

	elif filttype == 'bandpass':
	        flo1 = f[0] 
	        flo2 = f[1]
	        fhi1 = f[2] 
	        fhi2 = f[3]        
	        sigmalo = np.abs(flo2-flo1)/3
	        sigmahi = np.abs(fhi2-fhi1)/3
	        mulo = flo2
	        muhi = fhi1
	        Flo = Gaussian(fmat, mulo, sigmalo)
	        Fhi = Gaussian(fmat, muhi, sigmahi)
	        F = (Flo * np.where(fmat<=mulo, 1, 0)) + (Fhi * np.where(fmat>=muhi, 1, 0)) + (1 * np.where((fmat>mulo) & (fmat<muhi), 1, 0))

	else:
		ValueError('Filter type not known.')

	return F



def Gaussian(freqmat,mu,sigma):

	G = np.exp(-(freqmat - mu)**2 / (2 * sigma**2))
	G = G / np.max(G)

	return G