"""
Plot retrieved runoff limit on top of NIR image.
There is also some additional cloud masking logic in here, which does not tie
up with the wider cloud masking approach undertaken by the workflow.

At the moment, needs to be run from within job directory

"""

import glob
import pandas as pd
import datetime as dt
import rasterio
from rasterio.enums import Resampling
import geopandas as gpd
import os
import matplotlib.pyplot as plt
import numpy as np
import configparser
import argparse

import xarray as xr

parser = argparse.ArgumentParser(description='Runoff limit examination')
parser.add_argument('job_id', type=str, 
	help='str, job ID, as stored in $OUTPUT_DIR')
parser.add_argument('-plot', dest='plot', default=False, 
	action='store_true',
	help='Generate plots')
args = parser.parse_args()

config = configparser.ConfigParser()

# Identify job path
job_path = os.path.join(os.environ['OUTPUT_DIR'], args.job_id)

# Parse job file
job_file = os.path.join(job_path, args.job_id)
config.read_file(open(job_file, 'r'))

# This pipeline is run exclusively on warped imagery, so find out the epsg
epsg = config.get('warp', 'to_epsg')

sensor_id = config.get('main', 'sensor')

#########################################

# exclude = [
# 'LT05_L1TP_010011_19950908_20180220_01_T1',
# 'LT05_L1GS_007011_19910807_20180220_01_T2'
# ]

#########################################

max_elev = 2100
clds_row_dist_thresh = 500 #200px=6 km

fancy_clouds = False
cld_win_m = 2000
clds_win_thresh_perc = 0.2
clds_win_thresh_px = int(cld_win_m**2 / 30**2 * clds_win_thresh_perc)

#elev_below_med_tol = 100.


"""
NEED TO GET RID OF EASTERN POINTS !!!
"""


#########################################


csvs = glob.glob(os.path.join(job_path, '*/*RLIM.csv'))
#csvs = glob.glob('LE07_L1TP_010011_20120712_20161201_01_T1/LE07_L1TP_010011_20120712_20161201_01_T1_RLIM.csv')

store = []
for f in csvs:
	df = pd.read_csv(f)
	fn = f.split('/')[-1]
	df['date'] = dt.datetime.strptime(fn[17:25], '%Y%m%d')
	df['PRODUCT_ID'] = fn[:-9]
	store.append(df)

df = pd.concat(store, axis=0)
df.index = pd.MultiIndex.from_arrays((df['PRODUCT_ID'], df.x, df.y), names=('PRODUCT_ID','x','y'))


df_cull = df



df_cull = gpd.GeoDataFrame(df_cull, geometry=gpd.points_from_xy(df_cull.x, df_cull.y), crs=3413)


# Do not change cld_half
cld_half = cld_win_m / 2

retain = []
plt.ioff()
n = 0
ntot = len(df_cull.groupby(level=0))
for ix, p in df_cull.groupby(level=0):

	n += 1
	print(('{n}/{t} {ix}').format(n=n, t=ntot, ix=ix))

	# Delete points above max elev
	p_cull = p[p.elev < max_elev]
	p_cull = p_cull[p_cull.rlim]
	# Delete edge-proximal points
	p_cull = p_cull[p_cull.px_to_pr > 50]
	p_cull['px_to_cld_up'] = p_cull.px_to_cld_up.fillna(0)
	p_cull['px_to_cld_down'] = p_cull.px_to_cld_down.fillna(0)

	# Only keep points where clouds up-glacier are at least XX px away.
	# For some reason, a lot of _up pixels == 1, so threshold needs to be 2.
	p_cull = p_cull[((p_cull.px_to_cld_up > clds_row_dist_thresh) | (p_cull.px_to_cld_up < 2)) &
	 ((p_cull.px_to_cld_down > clds_row_dist_thresh) | (p_cull.px_to_cld_down < 1))]

	if fancy_clouds:
		# For points which remain, check within a larger search box for clouds 
		# We do this as 2nd step as it is a bit more expensive, so want to minimise number of calls
		cld_f = os.path.join(os.environ['OUTPUT_DIR'], args.job_id, ix, ix + '_mask_cloud_EPSG3413.tif')
		with xr.open_rasterio(cld_f) as clds:
			store = []
			for ix_here, row in p_cull.iterrows():
				win = clds.sel(band=1).sel(x=slice(row.x-cld_half, row.x+cld_half), y=slice(row.y+cld_half,row.y-cld_half))
				win_tot = win.sum()
				store.append(int(win_tot))
		# Add extra cloud info to dataframe
		p_cull['cld_neigh'] = store
		
		p_cull = p_cull[p_cull['cld_neigh'] < clds_win_thresh_px]
	
	# We assume that the runoff elevation of a scene should be pretty even
	# as the size of a scene is climatically quite constant.
	# Thus, compute the median elevation of points which remain then cull any points
	# which are less than the median minus a tolerance.
	# 2020-06-19: this doesn't always work as expected, so disabled. Should probably be done even later in processing, using a stipulation of
	# minimum n.points for each product, as it seems quite sensitive to this. Also, -100 m results in data loss in south Greenland, which is 
	# much less homogeneous per tile than along the western margin. 
	# Consider doing it by neighbourhood somehow. (rolling, or kd-tree...)
	#p_cull = p_cull[p_cull.elev >= (p_cull.elev.median()-elev_below_med_tol)]

	p_cull.to_csv(os.path.join(job_path, ix, ix+'_RLIM_retain.csv'))

#	retain.append(p_cull)

	if args.plot:
		## Now plot
		# Import background image for plotting - in original projection
		if sensor_id == 'OLI_TIRS':
			band = 'B5'
		else:
			band = 'B4'
		im_f = os.path.join(os.environ['L0data'],sensor_id,ix, ix+'_' + band + '.TIF')
		with rasterio.open(im_f) as im_h:
			im = im_h.read(out_shape=(int(im_h.height/4),int(im_h.width/4)), resampling=Resampling.nearest)
			b = im_h.bounds
			extent = (b.left, b.right, b.bottom, b.top)
			epsg = im_h.crs.to_epsg()

		plt.figure()
		plt.imshow(np.squeeze(im), extent=extent, cmap='Greys_r')

		# Project points to Landsat projection then plot
		p_reproj = p.to_crs(epsg)
		plt.plot(p_reproj.geometry.x, p_reproj.geometry.y, 'x', alpha=0.1, markersize=1, label='All')
		
		cull_reproj = p_cull.to_crs(epsg)
		plt.plot(cull_reproj.geometry.x, cull_reproj.geometry.y, 'x', alpha=0.2, markersize=2, label='Retained')

		plt.title(ix)
		plt.savefig(os.path.join(os.environ['OUTPUT_DIR'],args.job_id, ix, ix+'_plotlimit_ecull_median.png'), dpi=200)
		plt.close()

plt.ion()

#retain_df = pd.concat(retain, axis=0)
#retain_df.to_csv(os.path.join(job_path, 'retain_df.csv'))


