"""
EO-Learn tasks for handling historic (non-SentinelHub) Landsat imagery.

Andrew Tedstone, November 2019.

"""

from eolearn import io
from eolearn.core.constants import *
from eolearn.core import EOPatch, EOTask
from eolearn.core.eotask import CompositeTask
from eolearn.geometry import VectorToRaster
from eolearn.mask import MaskFeature


import subprocess
import os
import numpy as np
import glob

import geopandas 
import rasterio
from rasterio import fill
from skimage import morphology

import morphological

import mtlutils


class ImportLandsatTask(EOTask):
	""" Import all specified Landsat bands into eopatch instance.
	
	See also https://github.com/sentinel-hub/eo-learn/issues/106.

	"""

	def __init__(self, bands, convert_to_refl=False):
		self.bands = bands
		self.convert_to_refl = convert_to_refl

	def execute(self, image_path, eopatch=None, load_mtl=True):

		check_slc = False
		slc = False

		for b in self.bands:
			
			if (type(b) is str):
				if len(b) > 1:
					b = b.split('B')[1]


			fn = image_path + '_B%s.TIF' %b
			input_task = io.ImportFromTiff((FeatureType.DATA, 'B%s' %b), fn)
	
			eopatch = input_task.execute(eopatch=eopatch)	

			if not check_slc:
				if os.path.exists(image_path + '_B%s_GM.TIF' %b):
					slc = True
				else:
					slc = False
				check_slc = True


		with rasterio.open(fn) as dataset:
			crs = dataset.crs

		eopatch.meta_info['scene_path'] = image_path
		eopatch.meta_info['image_fn'] = image_path.split('/')[-1]
		eopatch.meta_info['scene_dir'] = '/'.join(image_path.split('/')[:-1])
		eopatch.meta_info['crs'] = crs
		
		# Image resolution
		llx, lly = eopatch.bbox.lower_left
		urx, ury = eopatch.bbox.upper_right
		eopatch.meta_info['xres'] = np.abs(llx - urx) / eopatch.data['B%s' %b].shape[2]
		eopatch.meta_info['yres'] = np.abs(lly - ury) / eopatch.data['B%s' %b].shape[1]

		# Metadata from MTL file
		if load_mtl or self.convert_to_refl:
			search_mtl = glob.glob(image_path + '*MTL*')
			if len(search_mtl) > 0:
				t = LoadMTLTask()
				eopatch = t.execute(eopatch, search_mtl[0])

		# Convert to reflectance if requested
		if self.convert_to_refl:
			t = Rad2ReflTask(self.bands)
			eopatch = t.execute(eopatch)

		eopatch.meta_info['SLC_OFF'] = slc

		return eopatch



class LoadMTLTask(EOTask):

	def __init__(self, folder=None):
		self.folder = folder
		return

	def execute(self, eopatch, filename):
		if self.folder is not None:
			filename = os.path.join(self.folder, filename)

		mtl = mtlutils.parsemeta(filename)
		eopatch.meta_info['mtl'] = mtl['L1_METADATA_FILE']
		return eopatch




class Rad2ReflTask(EOTask):

	def __init__(self, bands):
		self.bands = bands
		return

	def execute(self, eopatch):
		"""
		bands : list of features in FeatureType.DATA to rescale, valid
		values are 'B1'...'Bn' depending on satellite.
		"""

		for b in self.bands:
			# Convert to TOA reflectance 
			# https://www.usgs.gov/land-resources/nli/landsat/using-usgs-landsat-level-1-data-product
			if type(b) is str:
				band_n = int(b.split('B')[1])
			elif type(b) is int:
				band_n = b
			else:
				raise('landsat.Rad2Ref :: lUnknown band identifier')
			M_p = eopatch.meta_info['mtl']['RADIOMETRIC_RESCALING']['REFLECTANCE_MULT_BAND_%s' %band_n]
			A_p = eopatch.meta_info['mtl']['RADIOMETRIC_RESCALING']['REFLECTANCE_ADD_BAND_%s' %band_n]
			eopatch['data']['B%s' %band_n] = M_p * eopatch['data']['B%s' %band_n] + A_p

		return eopatch




class ImportLandsatQATask(EOTask):
	""" Task for importing quality info from Landsat Level-1 BQA files into EOPatch.

	Requires Landsat_QA_Tools tool `unpack_collection_qa' to be present on PATH.

	More information:
	https://www.usgs.gov/land-resources/nli/landsat/landsat-level-1-quality-assessment-tools?qt-science_support_page_related_con=2#qt-science_support_page_related_con
	https://www.usgs.gov/media/images/landsat-1-8-collection-1-level-1-quality-bit-designations

	"""


	def __init__(self, qa_attribute, clean=True):
		"""
		:param qa_attribute: one of combine, fill, drop_pixel, terrain_occl,
		radiometric_sat, cloud, cloud_confidence, cloud_shadow, snow_ice, cirrus.
		Do not supply 'all'.
		"""

		self.qa_attr = qa_attribute.lower()
		self.clean = clean


	def execute(self, eopatch, errors='raise'):
		""" Execute method with adds QA layer to EOPatch.

		:param errors: 'ignore' or 'raise'

		:returns: eopatch

		"""


		# Check that BQA file is available.
		if not os.path.isfile(eopatch.meta_info['scene_path'] + '_BQA.TIF'):
			if errors == 'raise':
				raise FileNotFoundError
			elif errors == 'ignore':
				return eopatch
			else:
				raise NotImplementedError

		# Landsat QA tools have to be executed in folder containing image
		cwd = os.getcwd()
		os.chdir(eopatch.meta_info['scene_dir'])
		
		# For unpack_collection_qa to work, filename must be PRODUCT_ID
		# Set up a symlink if the filename is SCENE_ID
		symlink = False
		if eopatch.meta_info['image_fn'] == eopatch.meta_info['mtl']['METADATA_FILE_INFO']['LANDSAT_SCENE_ID']:
			symlink = True
			bqa_fn = eopatch.meta_info['mtl']['METADATA_FILE_INFO']['LANDSAT_PRODUCT_ID'] + '_BQA.TIF'
			cmd = 'ln -s ' + eopatch.meta_info['scene_path'] + '_BQA.TIF ' + \
				bqa_fn
			subprocess.check_output(cmd, shell=True)
		else:
			bqa_fn = eopatch.meta_info['image_fn'] + '_BQA.TIF'

		# Generate individual QA GeoTIFF file
		cmd = 'unpack_collection_qa --ifile=%s --%s --ofile=%s' \
					%(bqa_fn, 
					  self.qa_attr,
					  eopatch.meta_info['image_fn'])
		subprocess.check_output(cmd, shell=True)

		# Load the file
		qa_fn = os.path.join(eopatch.meta_info['scene_dir'], eopatch.meta_info['image_fn'] + '_%s.tif' %self.qa_attr)
		input_task = io.ImportFromTiff((FeatureType.MASK, self.qa_attr.upper()), 
			folder=qa_fn)
		eopatch = input_task.execute(eopatch=eopatch)

		if self.clean:
			os.remove(qa_fn)
		if symlink:
			os.remove(bqa_fn)

		# Return to main working directory
		os.chdir(cwd)

		return eopatch




class SLC_FillFillTask(EOTask):
	""" Generate a uniform 'designated fill' mask that removes striping, to set
	whole scene area post-fillnodata to 0-values.
	"""

	def __init__(self, feature=(FeatureType.MASK,'FILL')):
		self.feature = self._parse_features(feature)
		return


	def execute(self, eopatch):

		if not eopatch.meta_info['SLC_OFF']:
			return eopatch

		feature_type, feature_name = next(self.feature())
		shape = eopatch[feature_type][feature_name].shape
		im = np.squeeze(eopatch[feature_type][feature_name])
		# Invert so that we can work outwards from the 'bad' data
		im = np.where(im == 1, 0, 1) 
		for n in range(0, 25):
			im = morphology.binary_dilation(im)
		for n in range(0, 25):
			im = morphology.binary_erosion(im)
		# Flip back
		im = np.where(im == 1, 0, 1) 	
		
		# Return in-place
		im = np.reshape(im, shape)
		eopatch[feature_type][feature_name] = im

		return eopatch



class SLC_FillDataTask(EOTask):
	"""
	Use gdal_fillnodata to fill gaps in all bands requested. This task assumes
	that a gap mask file (*GM.TIF) is available for each band.

	"""

	def __init__(self, bands):
		self.bands = bands
		return

	def execute(self, eopatch):

		if not eopatch.meta_info['SLC_OFF']:
			return eopatch

		for b in self.bands:

			feature_type = FeatureType.DATA
			feature_name = 'B%s' %b
			shape = eopatch[feature_type][feature_name].shape

			gm_mask_f = eopatch.meta_info['scene_path'] + '_' + feature_name + '_GM.TIF'
			gm_mask_h = rasterio.open(gm_mask_f)
			gm_mask = gm_mask_h.read()

			filled = fill.fillnodata(np.squeeze(eopatch[feature_type][feature_name]), 
				mask=np.squeeze(gm_mask))

			eopatch[feature_type][feature_name] = np.reshape(filled, shape)
		
		return eopatch




class MaskByWRSPathRow(EOTask):

	def __init__(self, band, feature_name, shapefile,
		remove_previous=False):
		"""

		Supply either path to shapefile, or already loaded into a geodataframe

		Requires MTL file to have been loaded.

		"""
		self.band = band
		self.feature_name = feature_name
		self.remove_previous = remove_previous

		if type(shapefile) is str:
			self.shp = geopandas.read_file(shapefile)
		elif type(shapefile) is geopandas.geodataframe.GeoDataFrame:
			self.shp = shapefile
		else:
			raise ValueError
		

	def execute(self, eopatch):

		# Get polygon of Path/Row
		path = eopatch['meta_info']['mtl']['PRODUCT_METADATA']['WRS_PATH']
		row = eopatch['meta_info']['mtl']['PRODUCT_METADATA']['WRS_ROW']
		pr = self.shp[(self.shp.PATH == path) & (self.shp.ROW == row)]
		if len(pr) == 0:
			raise ValueError('No polygon for this path/row returned')
		
		# Re-project to system of current eopatch
		proj = eopatch['meta_info']['mtl']['PROJECTION_PARAMETERS']
		#crs = "+proj=UTM +zone=%s +datum=%s +units=m +no_defs" %(proj['UTM_ZONE'], proj['DATUM'])
		pr = pr.to_crs(eopatch['meta_info']['crs'])

		# Burn polygon to raster
		height, width = eopatch['data'][self.band].shape[1:3]
		burnTask = VectorToRaster(pr, (FeatureType.MASK_TIMELESS, self.feature_name), 
			values=1, 
			raster_shape=(height, width))
		eopatch = burnTask.execute(eopatch)

		# Erode to make sure cruddy edges are chopped off
		erodeTask = morphological.MultiDilateErodeTask(
			(FeatureType.MASK_TIMELESS, self.feature_name),
			(FeatureType.MASK_TIMELESS, self.feature_name),
			0, 100, remove_previous=False)
		eopatch = erodeTask.execute(eopatch)


		# Apply as mask, !!BODGE BODGE BODGE dtype conversion
		eopatch[FeatureType.DATA][self.band] = eopatch[FeatureType.DATA][self.band].astype(np.int16)
		maskTask = MaskFeature(
			(FeatureType.DATA, self.band, self.feature_name), 
			(FeatureType.MASK_TIMELESS, self.feature_name), [0,], 
			no_data_value=-1)

		eopatch = maskTask.execute(eopatch)

		if self.remove_previous:
			eopatch.remove_feature(FeatureType.DATA, self.band)


		return eopatch




class SWIRCloudMaskTask(EOTask):

	def __init__(self, band, mask_name, threshold, dilate_by=0,
		previous_feature_name=None, feature_name=None,  
		keep_swir_band=False,
		remove_previous=False,
		mask_with=None):
		"""
		band : Landsat band number to load - specify band number of SWIR band.
		mask_name : name to give this mask.

		previous_feature_name : None or name of FeatureType.DATA feature to apply this mask to.
		feature_name : None or name of resulting FeatureType.DATA feature.

		threshold : reflectance threshold above which pixels treated as clouds.
		dilate_by : numver of pixels to dilate thresholded cloud mask by.
		keep_swir_band : bool. Retain the loaded band in eopatch memory or not?
		
		"""

		self.band = band
		self.mask_name = mask_name
		self.threshold = threshold
		self.dilate_by = dilate_by
		
		self.previous_feature_name = previous_feature_name
		self.feature_name = feature_name
		self.keep_swir_band = keep_swir_band
		self.remove_previous = remove_previous

		if mask_with is not None:
			if type(mask_with) is tuple:
				self.mask_with = self._parse_features(mask_with)
			else:
				self.mask_with = mask_with


	def execute(self, eopatch):

		# Import the SWIR band
		load_swir = ImportLandsatTask([self.band])
		eopatch = load_swir(eopatch.meta_info['scene_path'], eopatch=eopatch, 
			load_mtl=False)

		# Apply mask with here (to remove possible artefacts before it gets blown by dilation)
		if self.mask_with is not None:
			mask_with_type, mask_with_name = next(self.mask_with())
			t = MaskFeature((FeatureType.DATA, 'B%s' %self.band, 'B%s' %self.band),
						(mask_with_type, mask_with_name), [1,], no_data_value=0)
			eopatch = t.execute(eopatch)

		# Convert to TOA reflectance 
		# https://www.usgs.gov/land-resources/nli/landsat/using-usgs-landsat-level-1-data-product
		band_n = int(self.band.split('B')[0])
		M_p = eopatch.meta_info['mtl']['RADIOMETRIC_RESCALING']['REFLECTANCE_MULT_BAND_%s' %band_n]
		A_p = eopatch.meta_info['mtl']['RADIOMETRIC_RESCALING']['REFLECTANCE_ADD_BAND_%s' %band_n]
		eopatch['data']['B%s' %self.band] = M_p * eopatch['data']['B%s' %self.band] + A_p

		# Generate the mask ...
		# Apply threshold
		tmp = np.where(eopatch['data']['B%s' %self.band] > float(self.threshold), 1, 0)

		# Dilate
		if int(self.dilate_by) > 0:
			for n in range(0, int(self.dilate_by)):
				tmp = morphology.binary_dilation(tmp)
		# Save the mask to patch
		eopatch.add_feature(FeatureType.MASK, self.mask_name, tmp)
		tmp = None

		# Apply the mask to the requested feature (band)
		if (self.previous_feature_name is not None) & \
		   (self.feature_name is not None):
			maskTask = MaskFeature(
				(FeatureType.DATA, self.previous_feature_name, self.feature_name),
				(FeatureType.MASK, self.mask_name), [1,],
				no_data_value = -1)
			eopatch = maskTask.execute(eopatch)

		# Unless otherwise requested, delete the SWIR band from memory
		if not self.keep_swir_band:
			eopatch.remove_feature(FeatureType.DATA, 'B%s' %self.band)

		if self.remove_previous:
			eopatch.remove_feature(FeatureType.DATA, self.previous_feature_name)


		return eopatch




class MultiCloudMaskTask(EOTask):

	def __init__(self, mask_name, mask_types,
		previous_feature_name=None, feature_name=None,
		task_options=None,
		remove_previous=False,
		mask_with=None):
		"""
		Process multiple types of cloud mask.

		mask_types: list of masks to use. Currently available options: 
			- cloud. Uses cloud mask from Landsat QA band at default confidence.
			- cloud_shadow. Uses cloud shadow mask from Landsat QA band at default confidence.
			- cloud_swir. Uses shortwave infrared band. Pass additional arguments via task_options.

		task_options: None, or dict of format:
			{
				'mask_type' : {'arg1':'val1'} 
			}
			for 'cloud_swir' task, provide keys 'band', 'threshold', 'dilate', 'keep_band'.

		"""

		self.mask_name = mask_name
		self.mask_types = mask_types
		
		self.previous_feature_name = previous_feature_name
		self.feature_name = feature_name
		self.task_options = task_options
		self.remove_previous = remove_previous
		if mask_with is not None:
			self.mask_with = self._parse_features(mask_with)


	def execute(self, eopatch):

		combo_mask = None

		if self.mask_with is not None:
			mask_with_type, mask_with_name = next(self.mask_with())
		
		for m in self.mask_types:
			
			if (m.lower() == 'cloud') | (m.lower() == 'cloud_shadow'):
				t = ImportLandsatQATask(m.upper(), clean=True)

				if self.mask_with is not None:
					t2 = MaskFeature(
						(FeatureType.MASK, m.upper(), m.upper()),
						(mask_with_type, mask_with_name), [1,], no_data_value=0)
					t = CompositeTask(t, t2)

				if eopatch.meta_info['SLC_OFF']:
					t2 = morphological.MultiDilateErodeTask(
							(FeatureType.MASK, m.upper()), 
							(FeatureType.MASK, m.upper()),
							25, 25,
							remove_previous=False, no_data_value=None)
					t = CompositeTask(t, t2)


			elif m.lower() == 'cloud_swir':
				t = SWIRCloudMaskTask(self.task_options['cloud_swir']['band'],
					m.upper(), 
					self.task_options['cloud_swir']['threshold'],
					self.task_options['cloud_swir']['dilate'], 
					keep_swir_band=self.task_options['cloud_swir']['keep_band'],
					mask_with=self.mask_with)

			eopatch = t.execute(eopatch)

			if combo_mask is None:
				combo_mask = np.zeros(eopatch['mask'][m.upper()].shape)

			combo_mask = combo_mask + eopatch['mask'][m.upper()]

		# Now combine them
		combo_mask = np.where(combo_mask >= 1, 1, 0)

		# Add to patch
		eopatch.add_feature(FeatureType.MASK, self.mask_name, combo_mask)

		# Apply the mask if requested
		if (self.previous_feature_name is not None) & \
		   (self.feature_name is not None):
			maskTask = MaskFeature(
				(FeatureType.DATA, self.previous_feature_name, self.feature_name),
				(FeatureType.MASK, self.mask_name), [1,],
				no_data_value=0)
			eopatch = maskTask.execute(eopatch)

		if self.remove_previous:
			eopatch.remove_feature(FeatureType.DATA, self.previous_feature_name)

		return eopatch



