from eolearn import io
from eolearn.core.constants import *
from eolearn.core import EOPatch, EOTask, LinearWorkflow
from joblib import Parallel, delayed

import numpy as np
import pandas as pd
import os

import PyDIP as dip

import gabor_filt

from skimage.measure import label, regionprops, regionprops_table
from skimage.morphology import binary_dilation, binary_erosion, skeletonize

## REMOVE FOR OPERATIONAL USE!!!
from importlib import reload
reload(gabor_filt)

class GaborTask(EOTask):
	
	def __init__(self, band, feature_name, ndirs, freq, remove_previous=False,
				 mask_borders=0, jobs=-1, prefer='threads', **kwargs):
		"""
		mask_borders: None or number of pixels along each border/margin to set to
		zero after the Gabor filter has executed. Use to remove edge effects.

		"""
		self.band = band
		self.feature_name = feature_name
		self.remove_previous = remove_previous
		self.kernels = gabor_filt.design_kernels_evenspaced(ndirs, freq, **kwargs)
		self.mask_borders = mask_borders
		# See https://joblib.readthedocs.io/en/latest/parallel.html
		self.jobs = jobs
		self.prefer = prefer
		

	def execute(self, eopatch):
		im = np.squeeze(eopatch.data[self.band])
		norm = gabor_filt.normalise(im)

		# Run gabor filter
		results = Parallel(n_jobs=int(self.jobs), prefer=self.prefer) \
					(delayed(gabor_filt.run)(norm, k) for k in self.kernels)

		# Collate results
		processed = np.zeros_like(norm)
		for i in results:
			processed += i

		if self.mask_borders > 0:
			msk = np.zeros(processed.shape)
			msk[self.mask_borders:-self.mask_borders, self.mask_borders:-self.mask_borders] = 1
			processed = processed * msk

		result = np.reshape(processed, (1, processed.shape[0], processed.shape[1], 1))
		eopatch.add_feature(FeatureType.DATA, self.feature_name, result)

		if self.remove_previous:
			eopatch.remove_feature(FeatureType.DATA, self.band)
		
		return eopatch
		




class PathOpeningTask(EOTask):
	
	def __init__(self, band, feature_name, length, remove_previous=False):
		self.band = band
		self.feature_name = feature_name
		self.length = length
		self.remove_previous = remove_previous

	def execute(self, eopatch):
		# needs to be a dip Image
		im = np.squeeze(eopatch.data[self.band])

		mask = np.ones(im.shape).astype(np.bool)
		result = np.array(dip.PathOpening(im, mask, length=self.length))

		# Pixels smaller than mean cannot be rivers.
		meanPO = np.nanmean(result)
		result = np.where(result > meanPO, result, meanPO)

		result = np.reshape(result, (1, result.shape[0], result.shape[1], 1))
		eopatch.add_feature(FeatureType.DATA, self.feature_name, result)

		if self.remove_previous:
			eopatch.remove_feature(FeatureType.DATA, self.band)
		
		return eopatch




class ThresholdStdDevTask(EOTask):
	""" Note that if masking has been applied to input feature with specific
	values given to masked pixels then this could cause problems here.
	1. Values less than zero are removed immediately and so do not affect
	the thresholding process
	2. Other values could affecting the process
	3. Values less than zero are not restored to the output image.

	min_abs_value : the minimum absolute value that the threshold must be. If
	the threshold calculated by the mean+(std*n_std) is < min_abs_value, then
	the threshold applied will equal min_abs_value.

	returns : eopatch, with metakey 'ThresholdStdDev_threshold' containing 
	the threshold value.
	"""

	def __init__(self, band, feature_name, n_std_dev,
		remove_previous=False,
		min_abs_value=0):
		self.band = band
		self.feature_name = feature_name
		self.n_std_dev = n_std_dev
		self.remove_previous = remove_previous
		self.min_abs_value = min_abs_value

	def execute(self, eopatch):

		im = np.squeeze(eopatch.data[self.band])
		
		# Make sure that zeros, -1s etc do not contribute to thresholding
		im = np.where(im < 1, np.nan, im)

		im_mean = np.nanmean(im)
		im_std = np.nanstd(im)
		threshold = im_mean + (im_std * self.n_std_dev)

		threshold = np.maximum(threshold, self.min_abs_value)

		result = np.uint8(np.where(im > threshold, 1, 0))

		result = np.reshape(result, (1, result.shape[0], result.shape[1], 1))
		eopatch.add_feature(FeatureType.DATA, self.feature_name, result)

		if self.remove_previous:
			eopatch.remove_feature(FeatureType.DATA, self.band)

		eopatch.meta_info['ThresholdStdDev_threshold'] = threshold
		return eopatch



class MultiDilateErodeTask(EOTask):

	def __init__(self, feature_in, feature_out, n_dilate, n_erode,
		remove_previous=False, no_data_value=None):
		self.feature_in = self._parse_features(feature_in)
		self.feature_out = self._parse_features(feature_out)
		self.n_dilate = n_dilate
		self.n_erode = n_erode
		self.remove_previous = remove_previous
		self.no_data_value = no_data_value

	def execute(self, eopatch):

		feature_type, feature_name = next(self.feature_in())

		shape = eopatch[feature_type][feature_name].shape
		im = np.squeeze(eopatch[feature_type][feature_name])

		if self.no_data_value is not None:
			# Convert back to binary image
			im[im == self.no_data_value] = 0

		for n in range(0,self.n_dilate):
			im = binary_dilation(im)

		for n in range(0,self.n_erode):
			im = binary_erosion(im)

		im = np.reshape(im, shape)

		if self.no_data_value is not None:
			im = np.where(eopatch[feature_type][feature_name] == self.no_data_value,
				self.no_data_value,
				im)

		new_feature_type, new_feature_name = next(self.feature_out())
		eopatch.add_feature(new_feature_type, new_feature_name, im)

		if self.remove_previous:
			eopatch.remove_feature(feature_type, feature_name)

		return eopatch




class SkeletonizeTask(EOTask):

	def __init__(self, feature_in, feature_out,
		remove_previous=False, no_data_value=None):
		self.feature_in = self._parse_features(feature_in)
		self.feature_out = self._parse_features(feature_out)
		self.remove_previous = remove_previous
		self.no_data_value = no_data_value

	def execute(self, eopatch):

		feature_type, feature_name = next(self.feature_in())
		im = np.squeeze(eopatch[feature_type][feature_name])

		if self.no_data_value is not None:
			# Convert back to binary image
			im[im == self.no_data_value] = 0

		im = skeletonize(im)
		im = np.reshape(im, (1, im.shape[0], im.shape[1], 1))

		if self.no_data_value is not None:
			# Then add the no data values back in
			im = np.where(eopatch[feature_type][feature_name] == self.no_data_value,
				self.no_data_value,
				im)
		
		new_feature_type, new_feature_name = next(self.feature_out())
		eopatch.add_feature(new_feature_type, new_feature_name, im)
		
		if self.remove_previous:
			eopatch.remove_feature(feature_type, feature_name)

		return eopatch



class LabelImageTask(EOTask):

	def __init__(self, feature_in, feature_out, no_data_value=None):
		"""
		feature_in : feature to label.
		feature_out : name of output feature to create containing labels.
		properties: None, or list of properties to return from skimage.measure.regionprops
		feature_intensity : None, or Feature to use as intensity image input to regionprops,
			n.b. this could be the same as feature_in, or a different feature!

		"""
		self.feature_in = self._parse_features(feature_in)
		self.feature_out = self._parse_features(feature_out)
		self.no_data_value = no_data_value

	def execute(self, eopatch, save_properties=None):
		"""
		save_properties: None or file path and name to save a csv of regionprops to
		
		returns:
			eopatch.
			if properties requested, these are returned in eopatch.meta_info,
			with key name <feature_out>_regions.

		"""

		feature_type, feature_name = next(self.feature_in())
		im = np.squeeze(eopatch[feature_type][feature_name])

		if self.no_data_value is not None:
			# Convert back to binary image
			im[im == self.no_data_value] = 0

		labels = label(im)
		labels_sv = np.reshape(labels, (1, labels.shape[0], labels.shape[1], 1))

		if self.no_data_value is not None:
			# Then add the no data values back in
			labels_sv = np.where(eopatch[feature_type][feature_name] == self.no_data_value,
				self.no_data_value,
				labels_sv)
		
		feature_type, feature_name = next(self.feature_out())
		eopatch.add_feature(feature_type, feature_name, labels_sv)

		return eopatch



class ApplyLabelsTask(EOTask):
	"""
	Replace 1-values of binary input feature with the equivalent values
	from feature_values.
	"""

	def __init__(self, feature_in, feature_out, feature_values):
		self.feature_in = self._parse_features(feature_in)
		self.feature_out = self._parse_features(feature_out)
		self.feature_values = self._parse_features(feature_values)

	def execute(self, eopatch):
		feature_type, feature_name = next(self.feature_in())
		im = np.squeeze(eopatch[feature_type][feature_name])

		feature_type, feature_name = next(self.feature_values())
		values = np.squeeze(eopatch[feature_type][feature_name])

		im = np.where(im == 1, values, 0)

		im = np.reshape(im, (1, im.shape[0], im.shape[1], 1))
		
		new_feature_type, new_feature_name = next(self.feature_out())
		eopatch.add_feature(new_feature_type, new_feature_name, im)

		return eopatch



class RegionPropertiesTask(EOTask):
	"""
	Generate region properties of a labelled image using scikit measure region_props.

	"""
	def __init__(self, feature_in, meta_key_out, properties, 
		feature_intensity=None, no_data_value=None):

		self.feature_in = self._parse_features(feature_in)
		self.meta_key_out = meta_key_out
		if feature_intensity is not None:
			self.feature_intensity = self._parse_features(feature_intensity)
		else:
			self.feature_intensity = None
		self.properties = properties
		self.no_data_value = no_data_value


	def execute(self, eopatch):
		
		# Load labelled image from which to extract properties
		feature_type, feature_name = next(self.feature_in())
		im = np.squeeze(eopatch[feature_type][feature_name])
		if self.no_data_value is not None:
			# Convert back to binary image
			im[im == self.no_data_value] = 0

		# Optionally, load the intensity image to extract further information from
		if self.feature_intensity is not None:
			intensity_type, intensity_name = next(self.feature_intensity())
			intensity_image = np.squeeze(eopatch[intensity_type][intensity_name])
			intensity_image = np.where(intensity_image <= 0, np.nan, intensity_image)
			self.properties.append('min_intensity')
			self.properties.append('mean_intensity')
			self.properties.append('max_intensity')
		else:
			intensity_image = None

		# Get properties of each labelled region	
		if 'label' not in self.properties:
			self.properties.append('label')

		# Run the region properties extraction
		regions = pd.DataFrame(
			regionprops_table(
				im,
				intensity_image=intensity_image,
				properties=self.properties
				))

		if self.feature_intensity is not None:
			# Rename the intensity properties to something more useful
			# according to name of supplied intensity feature
			regions = regions.rename(columns={
				'min_intensity' : ('min_{a}').format(a=intensity_name),
				'max_intensity' : ('max_{a}').format(a=intensity_name),
				'mean_intensity': ('mean_{a}').format(a=intensity_name)
				})

		eopatch.meta_info[self.meta_key_out] = regions

		return eopatch
