"""

Currently no multiprocessing, so kinda slow 
but not appropriate to use eolearn so would need to roll my own strategy
(as not raster based really at this point)


"""
import argparse
import configparser
import os
import glob
import datetime as dt
import pandas as pd
import geopandas as gpd
from shapely.geometry import Point
import rasterio
from rasterio.windows import Window
import numpy as np
from scipy.spatial import cKDTree
from scipy import ndimage
from skimage.morphology import skeletonize, remove_small_holes

from multiprocessing import Pool

import pyproj


parser = argparse.ArgumentParser(description='Runoff limit examination')
parser.add_argument('job_id', type=str,
                    help='str, job ID, as stored in $OUTPUT_DIR')
parser.add_argument('merge_template', type=str,
                    help='str, absolute path of culling/merging template for the quadrant')
parser.add_argument('-mar', dest='mar', default=False, action='store_true',
                    help='Generate MAR-based culling statistics according to merge template settings.')
parser.add_argument('-mask', dest='apply_mask_posthoc', default=None, type=str,
                    help='Apply the raster mask corresponding to the master map area (see README for more info!)')
parser.add_argument('-test', dest='test', default=False, action='store_true')

args = parser.parse_args()

if args.mar:
    import mar_raster

job_path = os.path.join(os.environ['OUTPUT_DIR'], args.job_id)
template_name = args.merge_template.split('/')[-1].split('.template')[0]

config = configparser.ConfigParser()
config.read_file(open(args.merge_template, 'r'))


data_proj = pyproj.Proj('epsg:{s}'.format(s=config.get('default', 'epsg')))
mar_proj = pyproj.Proj('+proj=stere +lat_0=70.5 +lon_0=-40 +k=1 +x_0=0 +y_0=0 +datum=WGS84 +units=m +no_defs')

def _get_mar_stats(bounds, image_dt, window=int(config.get('rcm', 'window_days'))):

    year = image_dt.year
    mar_f = os.path.join('/flash/tedstona/MARv3.11.2-6km-ERA5', 'MARv3.11.2-6km-daily-ERA5-{y}.nc'.format(y=year))
    mar = mar_raster.open_xr(mar_f)

    xmin, ymin = mar_proj(*data_proj(bounds[0], bounds[1], inverse=True))
    xmax, ymax = mar_proj(*data_proj(bounds[2], bounds[3], inverse=True))

    curr = mar.sel(X=slice(xmin, xmax), Y=slice(ymin, ymax))
    curr = curr.sel(TIME=slice((image_dt - dt.timedelta(days=window)
                                ).strftime('%Y-%m-%d'), image_dt.strftime('%Y-%m-%d')))

    me = curr.ME.sum(dim='TIME').values.flatten()
    sf = curr.SF.sum(dim='TIME').values.flatten()
    elev = curr.SH.values.flatten()

    combo = pd.DataFrame({'me': me, 'elev': elev, 'sf': sf})
    try:
        bins = pd.interval_range(combo.elev.min(), combo.elev.max(), freq=100)  # metres
        me_sf_per_100m = combo.groupby(pd.cut(combo.elev, bins)).median()
    except ValueError:
        me_sf_per_100m = None

    mar = None

    return me_sf_per_100m


def _filt_initial(rlims):
    # Initial filtering
    rlims = rlims[rlims.rlim == True]
    rlims = rlims[rlims.px_to_pr > int(config.get('filtering', 'npx_from_edge'))]
    rlims = rlims[rlims.elev < int(config.get('filtering', 'max_elev'))]

    # Proximity to clouds on row
    rlims['px_to_cld_up'] = rlims.px_to_cld_up.fillna(0)
    rlims['px_to_cld_down'] = rlims.px_to_cld_down.fillna(0)
    # Only keep points where clouds up-glacier are at least XX px away.
    # For some reason, a lot of _up pixels == 1, so threshold needs to be 2.
    c_thresh = float(config.get('filtering', 'clds_row_dist_thresh'))
    rlims = rlims[((rlims.px_to_cld_up > c_thresh) | (rlims.px_to_cld_up < 2)) &
                  ((rlims.px_to_cld_down > c_thresh) | (rlims.px_to_cld_down < 1))]

    return rlims


def _filt_master_map(rlims):

    # Get total bounds of rlim points which are being so far kept for this product
    bounds = rlims.total_bounds

    # Load the master map for these bounds
    with rasterio.open(map_file, 'r') as master_map:
        row_min, col_min = master_map.index(bounds[0], bounds[1])
        row_max, col_max = master_map.index(bounds[2], bounds[3])
        row_offset = row_min - row_max
        col_offset = col_max - col_min
        data = np.zeros((row_offset, col_offset))
        master_map.read(1, window=Window(col_min, row_max, col_offset, row_offset),
                        out=data, boundless=True, fill_value=0)

        # Look up exact coordinates of pixels chosen
        dx_min, dy_min = master_map.xy(row_min, col_min, offset='ul')  # consider adding offset parameter
        dx_max = dx_min + (30 * col_offset)
        dy_max = dy_min + (30 * row_offset)

    # Optionally, apply a mask (which must have exactly the same dims and georef as the master map file)
    if args.apply_mask_posthoc is not None:
        with rasterio.open(args.apply_mask_posthoc, 'r') as master_mask:
            data_mask = np.zeros((row_offset, col_offset))
            master_mask.read(1, window=Window(col_min, row_max, col_offset, row_offset),
                             out=data_mask, boundless=True, fill_value=0)
        data = np.where(data_mask > 0, data, 0)

    im_extent = (dx_min, dy_min, dx_min + (30 * col_offset), dy_min + (30 * row_offset))

    # Generate coordinates for each pixel of map
    x_coords = dx_min + (30 * np.arange(col_offset))
    y_coords = dy_min + (30 * np.arange(row_offset))
    X, Y = np.meshgrid(x_coords, y_coords)
    X = X.ravel()
    Y = Y.ravel()[::-1]

    data_flat = data.ravel()
    if np.sum(data_flat) < 1:
        print('Failed: no data')
        return None

    # Create look up tree
    X_rivers = X[data_flat > 0]
    Y_rivers = Y[data_flat > 0]
    XY_rivers = np.array((X_rivers, Y_rivers)).swapaxes(0, 1)
    tree = cKDTree(XY_rivers)

    # Check each point against tree
    # Might be possible to remove the loop here
    distances, indices = tree.query(np.array([rlims.x, rlims.y]).swapaxes(0, 1),
                                    k=1,
                                    distance_upper_bound=float(config.get('filtering', 'search_radius_m')),
                                    eps=float(config.get('filtering', 'search_radius_epsilon')))

    rlims['dist'] = distances
    rlims['i'] = indices

    rlims_keep = rlims[np.isfinite(rlims.dist)]

    # Finally, use statistics to preferentially get rid of any too-low estimates
    # - probably along diagonal upper edge of images if on SW of ice sheet.
    elev_std = float(config.get('filtering', 'elev_std'))
    if elev_std >= 0:
        rlims_keep = rlims_keep[rlims_keep.elev > (rlims_keep.elev.mean() - (rlims_keep.elev.std() * elev_std))]

    return rlims_keep


def _filt_mar(rlims_keep, mar_stats):

    # Highest elevation at which 50 mm we melting experienced in last two weeks, according to MAR
    me_elev = mar_stats[mar_stats.me > float(config.get('rcm', 'thresh_melt'))].elev.max()
    sf_elev = mar_stats[mar_stats.sf < float(config.get('rcm', 'thresh_snow'))].elev.max()

    me_tol = float(config.get('rcm', 'tol_elev_melt'))
    sf_tol = float(config.get('rcm', 'tol_elev_snow'))
    rlims_keep = rlims_keep[(rlims_keep.elev > me_elev - me_tol) & (rlims_keep.elev < me_elev + me_tol)]
    # only get rid of points if they are above the SF limit!
    rlims_keep = rlims_keep[(rlims_keep.elev < sf_elev + sf_tol)]

    return (rlims_keep, me_elev, sf_elev)


def control_filt(f):

    # Generate index information ...
    fn = f.split('/')[-1]

    #print('{c}/{t} {f}'.format(c=c, t=tot, f=fn))

    rlims = pd.read_csv(f)

    # Date of image acquisition
    im_date = dt.datetime.strptime(fn[17:25], '%Y%m%d')
    rlims['date'] = im_date
    # Product ID
    rlims['PRODUCT_ID'] = fn[:-9]

    # Geometry
    geometry = [Point(xy) for xy in zip(rlims.x, rlims.y)]
    rlims = gpd.GeoDataFrame(rlims, geometry=geometry, crs=int(config.get('default', 'epsg')))

    bounds = rlims.total_bounds

    # Various filtering stages
    rlims = _filt_initial(rlims)
    if len(rlims) < 1:
        print('No rlims left after initial filtering')
        return (fn, None, None, None)

    rlims_mm = _filt_master_map(rlims)
    if rlims_mm is not None:
        if len(rlims_mm) < 1:
            return (fn, None, None, None)

    if args.mar:
        mar_stats = _get_mar_stats(bounds, im_date)
        if mar_stats is None:
            return (fn, rlims_mm, None, None)

        rlims_mm_mar, me, sf = _filt_mar(rlims_mm, mar_stats)
        mar_elev = dict(me=me, sf=sf)

        # Consider saving to a new product csv here as well
        return (fn, rlims_mm, rlims_mm_mar, mar_elev)
    else:
        return (fn, rlims_mm)


# Establish filename to master map used for spatial filtering
map_type = config.get('filtering', 'master_map_type')
quad = config.get('quad', 'process')
map_file = os.path.join(os.environ['OUTPUT_DIR'], 'master_maps', template_name,
                        'master_{q}_{t}.vrt'.format(q=quad, t=map_type))


# Find all CSVs. By this point we don't need to worry about interrogating Landsat db,
# some products may have 'failed' as they don't have runoff features, so this isolates
# to the notionally successful products.
csvs = glob.glob(os.path.join(job_path, '*/*RLIM.csv'))

if args.test:
    csvs = csvs[0:10]

# csvs = csvs[0:10] ###!!! Debug!!


tot = len(csvs)

try:
    nworkers = int(os.environ['EOLEARN_WORKERS'])
except KeyError:
    print('WARNING: EOLEARN_WORKERS env-var not set, defaulting to one worker.')
    nworkers = 1
with Pool(processes=nworkers) as pool:
    results = pool.map(control_filt, csvs)


store_rlim_mm = []
store_rlim_mar = []
mar_store = {}
for result in results:
    if result[1] is not None:
        store_rlim_mm.append(result[1])
        if (args.mar) and (result[2] is not None):
            store_rlim_mar.append(result[2])
            mar_store[result[0]] = result[3]

all_rlim = pd.concat(store_rlim_mm, axis=0)

if not args.test:
    all_rlim.to_csv(os.path.join(job_path, args.job_id + '_RLIM_filtered_MM.csv'))

if args.mar:
    all_rlim_mar = pd.concat(store_rlim_mar, axis=0)
    all_rlim_mar.to_csv(os.path.join(job_path, args.job_id + '_RLIM_filtered_MM_MAR.csv'))

    if not args.test:
        mar_stats = pd.DataFrame.from_dict(mar_store, orient='index')
        mar_stats.to_csv(os.path.join(job_path, args.job_id + '_RLIM_filtered_MM_MAR_stats.csv'))


# Consider adding option to export results directly to database

"""

# still need to think about how rcm stats translate over to the 50 km bins

# need to remove the odd residual stripe from the master map ...

# run one product per thread/worker. Make sure that multiprocess=False to reduce access problems to master map vrt.

"""


# store = []
# nitems = 0
# for ix,g in all_rlim_pre.groupby(all_rlim_pre.PRODUCT_ID):
# 	sf_lim = mar_stats.loc[ix+'_RLIM.csv'].sf
# 	g = g[g.elev < sf_lim + 150]
# 	nitems += len(g)
# 	store.append(g)
