"""
Hydrological feature extraction from Landsat images processed through the RiverDetection chain.

Outputs:
- Image with labelled regions.
- Associated CSV file of propertties of each labelled region (label column 
	corresponds to raster value).

Andrew Tedstone, March 2020

"""
import numpy as np
import pandas as pd
from copy import deepcopy
import argparse
import configparser
import os

from eolearn.core import LinearWorkflow, EOExecutor, EOTask
from eolearn.core.constants import FeatureType
from eolearn.core.eodata import EOPatch
from eolearn.io.local_io import ImportFromTiff, ExportToTiff
from eolearn.mask import MaskFeature

import json

import run_landsat_query

import io_dataframes
import morphological
import landsat



################################################################################
# Get basic configuration information

parser = argparse.ArgumentParser(description='Hydrological feature post-processing')
parser.add_argument('job_id', type=str, 
	help='str, job ID, as stored in $OUTPUT_DIR')
parser.add_argument('-test', dest='test_first_product', default=False, 
	action='store_true', 
	help='Run for first product in query only.')
parser.add_argument('-restart', dest='restart', default=False,
	action='store_true',
	help=('Restart the processing chain to process remaining products which were '
		' not processed previously. ').format())
args = parser.parse_args()

config = configparser.ConfigParser()

# Identify job path
job_path = os.path.join(os.environ['OUTPUT_DIR'], args.job_id)

# Parse job file
job_file = os.path.join(job_path, args.job_id)
config.read_file(open(job_file, 'r'))

# Identify saved tasks with associated order number, so that we can load these 
# files in dynamically.
save_tasks = json.loads(config.get('chain', 'save'))
available_outputs = {}
n = 1
for t in save_tasks:
	available_outputs[t] = n
	n += 1

# This pipeline is run exclusively on warped imagery, so find out the epsg
epsg = config.get('warp', 'to_epsg')

# If true, the output of the previous step is removed by the next step, saving memory
remove_previous = config.getboolean('chain', 'remove_previous')



################################################################################
## Set up tasks and workflow
# Important: tasks are STRICTLY listed in order of processing.

eotasks = {}

# Load data
eotasks['task_load_opening'] = ImportFromTiff((FeatureType.DATA, 'opening'),
	job_path)

# Relies on a gdalvirtualraster then eolearn/rasterio cropping extent automatically
eotasks['task_load_dem'] = ImportFromTiff((FeatureType.DATA, 'dem'),
	config.get('postprocess', 'dem_in'))

eotasks['task_load_mask'] = ImportFromTiff((FeatureType.MASK, 'ICE_MASK'),
	config.get('postprocess', 'mask_in'))

eotasks['task_load_mask_desfill'] = ImportFromTiff((FeatureType.MASK, 'mask_desfill'),
	job_path)


# Decide on source of intensity values for labelled regions (this is optional)
intensity_choice = config.get('postprocess', 'region_intensity', fallback=None)
if intensity_choice is not None:
	
	feature_intensity = (FeatureType.DATA, intensity_choice)
	eotasks['task_load_intensity'] = ImportFromTiff(feature_intensity,
		job_path)

	if intensity_choice[0] == 'B':

		# Option requires warped band images to already be available inside
		# $OUTPUT_DIR/<product_ID>/.
		eotasks['task_load_mtl'] = landsat.LoadMTLTask(
			folder=os.path.join(os.environ['L0data'], 
			config.get('main', 'sensor')))

		eotasks['task_refl_band'] = landsat.Rad2ReflTask([intensity_choice,])

		intensity_fn_label = intensity_choice

else:
	feature_intensity = None


# Apply masks
eotasks['task_mask_ice'] = MaskFeature(
			(FeatureType.DATA, 'opening', 'opening_icemask'), 
			(FeatureType.MASK, 'ICE_MASK'), [0,], 
			no_data_value=0)

eotasks['task_mask_desfill'] = MaskFeature(
			(FeatureType.DATA, 'opening_icemask', 'opening_icemask_desmask'), 
			(FeatureType.MASK, 'mask_desfill'), [1,], 
			no_data_value=-1)


# Threshold the rivers from the opening image
eotasks['task_threshold'] = morphological.ThresholdStdDevTask('opening_icemask_desmask', 
	'threshold', 
	float(config.get('thresholding_stddev', 'number_std_dev')),
	remove_previous=remove_previous,
	min_abs_value=float(config.get('thresholding_stddev', 'min_abs_value', fallback=0)))


# Close holes to provide a bit of uniformity
ndilate = int(config.get('postprocess', 'ndilate'))
nerode = int(config.get('postprocess', 'nerode'))
eotasks['task_close_holes'] = morphological.MultiDilateErodeTask((FeatureType.DATA, 'threshold'),
	(FeatureType.DATA, 'closed'), ndilate, nerode, remove_previous=remove_previous,
	no_data_value=-1)


# Label regions
eotasks['task_label_main'] = morphological.LabelImageTask((FeatureType.DATA, 'closed'),
	(FeatureType.DATA, 'closed_label'))

eotasks['task_props_main'] = morphological.RegionPropertiesTask((FeatureType.DATA, 'closed_label'),
	'regions', json.loads(config.get('postprocess', 'label_properties')),
	feature_intensity=feature_intensity,
	no_data_value=-1)

eotasks['task_save_labels_im'] = ExportToTiff((FeatureType.DATA, 'closed_label'),
	folder=job_path, image_dtype=np.int16, no_data_value=-1, compress='lzw')

eotasks['task_save_labels_df'] = io_dataframes.ExportDFTask(save_path=job_path, 
	meta_feature_name='regions')


# Undertake same process but with skeletonised regions - this allows for later calculation
# of a proxy for channel length
eotasks['task_skel'] = morphological.SkeletonizeTask((FeatureType.DATA, 'closed'), 
	(FeatureType.DATA, 'skel'), remove_previous=False,
	no_data_value=-1)

# Extract the same labels as those allocated to the regions - important!
eotasks['task_label_skel'] = morphological.ApplyLabelsTask((FeatureType.DATA, 'skel'),
	(FeatureType.DATA, 'skel_label'), (FeatureType.DATA, 'closed_label'))

eotasks['task_props_skel'] = morphological.RegionPropertiesTask((FeatureType.DATA, 'skel_label'),
	'skel_regions', json.loads(config.get('postprocess', 'label_properties')),
	feature_intensity=(FeatureType.DATA, 'dem'),
	no_data_value=-1)

eotasks['task_save_skel_labels_im'] = ExportToTiff((FeatureType.DATA, 'skel_label'),
	folder=job_path, image_dtype=np.int16, no_data_value=-1, compress='lzw')

eotasks['task_save_skel_labels_df'] = io_dataframes.ExportDFTask(save_path=job_path, 
	meta_feature_name='skel_regions')


workflow = LinearWorkflow(*[eotasks[t] for t in eotasks])


################################################################################
# Query data and run pipeline.

products_df = run_landsat_query.execute_query(config, 
	run_landsat_query.create_query(config))


if args.restart:
	import glob
	already_processed = glob.glob(os.path.join(job_path, '*/*_*_*_*_*_*_*SKEL_LABELS.tif'))
	already_processed = [i.split('/')[-2] for i in already_processed]
	prev_n = len(products_df)
	products_df = products_df.drop(labels=already_processed, axis=0)
	print('\tRestart mode. Products for which there is already a processed folder will not be reprocessed.')
	print(('\tWere {prev_n} products, now processing {new_n}').format(
		prev_n=prev_n, new_n=len(products_df)))
	print('-'*80)


# Set up input arguments	
workflow_args = []
execution_names = []
for ix, product in products_df.iterrows():

	patch_args = {}

	# INPUTS
	# Path to input image
	image_path_opening = product.PRODUCT_ID + '_S%s_mask_cloud_EPSG%s.tif' %(available_outputs['mask_cloud'], epsg)
	image_path_opening = os.path.join(product.PRODUCT_ID, image_path_opening)
	patch_args[eotasks['task_load_opening']] = {'filename':image_path_opening}

	# Path to requested intensity image
	if feature_intensity is not None:
		image_path_intensity = product.PRODUCT_ID + '_%s_EPSG%s.tif' %(intensity_fn_label, epsg)
		image_path_intensity = os.path.join(product.PRODUCT_ID, image_path_intensity)
		patch_args[eotasks['task_load_intensity']] = {'filename':image_path_intensity}

	# If intensity image is a band, then MTL load will have been requested
	if 'task_load_mtl' in eotasks:
		image_path_mtl = product.PRODUCT_ID + '_MTL.txt' 
		image_path_mtl = os.path.join(product.PRODUCT_ID, image_path_mtl)
		patch_args[eotasks['task_load_mtl']] = {'filename':image_path_mtl}

	# Path to input designated fill mask
	image_path_desfill = product.PRODUCT_ID + '_qa_fill_EPSG%s.tif' %epsg
	image_path_desfill = os.path.join(product.PRODUCT_ID, image_path_desfill)
	patch_args[eotasks['task_load_mask_desfill']] = {'filename':image_path_desfill}


	## OUTPUTS
	# Path to output labels tif
	output_path_labels_im = product.PRODUCT_ID + '_EPSG%s_LABELS.tif' %epsg
	output_path_labels_im = os.path.join(product.PRODUCT_ID, output_path_labels_im)
	patch_args[eotasks['task_save_labels_im']] = {'filename':output_path_labels_im}

	# Path to output labels csv
	output_path_labels_df = output_path_labels_im[:-3] + 'csv'
	patch_args[eotasks['task_save_labels_df']] = {'filename':output_path_labels_df}

	# Path to output skel labels tif
	output_path_skel_labels_im = product.PRODUCT_ID + '_EPSG%s_SKEL_LABELS.tif' %epsg
	output_path_skel_labels_im = os.path.join(product.PRODUCT_ID, output_path_skel_labels_im)
	patch_args[eotasks['task_save_skel_labels_im']] = {'filename':output_path_skel_labels_im}

	# Path to output skel labels csv
	output_path_skel_labels_df = output_path_skel_labels_im[:-3] + 'csv'
	patch_args[eotasks['task_save_skel_labels_df']] = {'filename':output_path_skel_labels_df}

	workflow_args.append(patch_args)
	execution_names.append(product.PRODUCT_ID)


# Run chain
if args.test_first_product:
	print('TEST MODE')
	print(('Product ID: {id}').format(id=execution_names[0]))
	result = workflow.execute(workflow_args[0])
	first_pp = list(result.values())[0]

else:
	executor = EOExecutor(workflow, workflow_args, 
		save_logs=True, logs_folder=job_path)

	try:
		nworkers = int(os.environ['EOLEARN_WORKERS'])
	except KeyError:
		print('WARNING: EOLEARN_WORKERS env-var not set, defaulting to one worker.')
		nworkers = 1

	results = executor.run(workers=nworkers, multiprocess=True)
	executor.make_report()
	print('Report saved to:\n' + executor.get_report_filename())

