"""
Detect runoff limit from labelled hydrological features

Implemented as an EO Workflow pipeline

Relies upon having a geotiff of aspect already generated in the same
EPSG of the job being worked with. The area does not however need to be pre-cropped
as eolearn will subset from a larger tiff.  

Job file key: postprocess->aspect_in, full file path.

$ gdaldem aspect <dem_in_postprocess_epsg> <output_name>

-----------------------

To Do:
- Choose label type to run on based on .job file
- look again at rotated coordinate transformation maths

Andrew Tedstone, March-April 2020

"""

import pandas as pd
import numpy as np
import georaster
import richdem as rd
from scipy.ndimage import rotate
import rasterio
import rasterio.mask
from affine import Affine
import subprocess
import datetime as dt
import fiona

import argparse
import os
import configparser
import json

from eolearn.core import LinearWorkflow, EOExecutor, EOTask
from eolearn.core.constants import FeatureType
from eolearn.core.eodata import EOPatch
from eolearn.io.local_io import ImportFromTiff, ExportToTiff
from eolearn.mask import MaskFeature

import io_dataframes

import run_landsat_query

parser = argparse.ArgumentParser(description='Runoff limit detection')
parser.add_argument('job_id', type=str, 
	help='str, job ID, as stored in $OUTPUT_DIR')
parser.add_argument('-chunk', dest='chunk', type=int, default=None,
	help=('Control the workflow in chunks of n products. When this option is '
		'used, n products at a time are copied to local scratch for processing'
		'then the outputs are batch-copied back at the end. Proceeds for '
		'as many chunks as needed to finish the product query. Provide '
		'a value of n which will prevent local scratch from filling up. '
		'Best if value is a multiple of cores requested.').format())
parser.add_argument('-test', dest='test_first_product', default=False, 
	action='store_true', 
	help='Run for first product in query only.')
args = parser.parse_args()

config = configparser.ConfigParser()

if (args.chunk is not None) & (os.environ['WRKDIR'] == os.environ['OUTPUT_DIR']):
	raise RuntimeError('Chunking requested but $WORKDIR is the same as $OUTPUT_DIR!')

# Identify job path
job_path = os.path.join(os.environ['OUTPUT_DIR'], args.job_id)

# Work path
work_path = os.path.join(os.environ['WRKDIR'], args.job_id)

# Parse job file
job_file = os.path.join(job_path, args.job_id)
config.read_file(open(job_file, 'r'))

# This pipeline is run exclusively on warped imagery, so find out the epsg
epsg = config.get('warp', 'to_epsg')


#################################
FILTER_DESIGN_FILE = os.path.join(job_path, 'FilterFeaturesCriteria.py')
if not os.path.exists(FILTER_DESIGN_FILE):
	raise OSError('Filter design file not found! (Expected %s)' %FILTER_DESIGN_FILE)


################################################################################
print(('Started at {d}').format(d=dt.datetime.now()))

def px_to_km2(px):
	return px*(30**2) / 1e6

def km2_to_px(km2):
	return km2*1e6 / (30**2)


class FilterFeaturesTask(EOTask):

	def __init__(self, feature_in, feature_out, filter_design_file):
		self.feature_in = self._parse_features(feature_in)
		self.feature_out = self._parse_features(feature_out)
		self.filter_design_file = filter_design_file
		self._labels_to_drop = []
		return


	def _queue_labels(self, labels):
		labels = labels.label.values.tolist()
		self._labels_to_drop.extend(labels)
		self._labels_to_drop = np.unique(self._labels_to_drop).tolist()
		return


	def _filter_image(self, eopatch):

		# Load image
		feature_type, feature_name = next(self.feature_in())
		im = np.squeeze(eopatch[feature_type][feature_name])

		# Filter
		im = np.where(np.isin(im, self._labels_to_drop), 0, im)

		# Save
		im = np.reshape(im, (1, im.shape[0], im.shape[1], 1))
		new_feature_type, new_feature_name = next(self.feature_out())
		eopatch.add_feature(new_feature_type, new_feature_name, im)

		return eopatch


	def execute(self, eopatch):

		regions = eopatch.meta_info['regions']
		regions.index = regions.label
		skel_regions = eopatch.meta_info['skel_regions']
		skel_regions.index = skel_regions.label
		
		# Call 'arbitrary' code which designs filter - specific to this job
		with open(self.filter_design_file, 'r') as ff:
			exec(ff.read(), )

		## Now filter image to keep only retained labels
		eopatch = self._filter_image(eopatch)
		eopatch.meta_info['removed_labels'] = self._labels_to_drop

		return eopatch



class ScanRunoffLimitTask(EOTask):

	def __init__(self, labels_in, clds_in, icemsk_in, feature_out, 
		aspect_key=None, angle_l_r=90):
		self.labels_in = self._parse_features(labels_in)
		self.clds_in = self._parse_features(clds_in)
		self.icemsk_in = self._parse_features(icemsk_in)
		self.feature_out = self._parse_features(feature_out)
		self.aspect_key = aspect_key
		self.ANGLE_L_R = angle_l_r
		return


	def execute(self, eopatch, aspect=None):
		"""
		In principle it should be possible to transform coordinates directly
		from rotated coordinates, without needing to rotate the entire array.
		BUT, I haven't figured out the maths for this...

		"""

		__debug = False

		if aspect is not None:
			asp = aspect
		elif self.aspect_key is not None:
			asp = eopatch.meta_info[self.aspect_key]
		else:
			raise ValueError('No aspect provided.')

		rot_by = (360 - asp + self.ANGLE_L_R) * -1
		
		# Rotate labelled image.
		labels_type, labels_name = next(self.labels_in())
		im_rot = rotate(np.squeeze(eopatch[labels_type][labels_name]), rot_by)

		# Clouds also need rotating in order to calculate per row stats later.
		clds_type, clds_name = next(self.clds_in())
		cld_rot = rotate(np.squeeze(eopatch[clds_type][clds_name]), 
			rot_by,
			order=0)

		# Rotate ice mask
		icemsk_type, icemsk_name = next(self.icemsk_in())
		icemsk_rot = rotate(np.squeeze(eopatch[icemsk_type][icemsk_name]), 
			rot_by,
			order=0, mode='constant', cval=1) #!! This could cause problems in areas of narrow penisula

		h,w = np.squeeze(eopatch[labels_type][labels_name]).shape
		transform = rasterio.transform.from_bounds(*eopatch.bbox, 
			width=w, height=h)

		# Mask ice off-ice-sheet
		# This works under the presumption that the eopatch is fed with a full-size
		# ice mask, for example GIMP, which extends to all of eopatch - i.e. is 
		# not masked to the 'diamond' shape of the Landsat imagery.
		beyond = (icemsk_rot!=1).argmax(axis=1)
		# In rows that are completely ice-covered, set ice mask location to right hand edge
		beyond[np.sum(icemsk_rot, axis=1) == icemsk_rot.shape[1]] = icemsk_rot.shape[1] 

		# Now do L-->R scanning for every row of data in rotated image
		cols = im_rot.shape[1]
		indices = (np.where(im_rot > 0, 1, 0)!=0).argmax(axis=1)

		pr_indices = (np.where(im_rot > -1, 1, 0)!=0).argmax(axis=1)

		# Set up store to hold runoff position (in rotated coordinate space).
		store = {}
		store_arr = np.zeros(im_rot.shape)

		# Define column position to save to in event that no runoff limit is
		# detected. This is so that a -ve version of the row number can 
		# still be saved there without being clipped off when the store is
		# unrotated later. It will then be detected by virtue of being -ve.
		default_col = int(store_arr.shape[1] / 2)

		row = 0
		for col, prcol, b in zip(indices, pr_indices, beyond):

			if col > 0:
				## Check how near we are to image edge (based on 'designated fill' mask)
				# There'll be designated fill pixels along both sides of image, so just look at the left-hand-side following rotation
				pr_area = im_rot[row, :col]
				# Reverse so that we can find the designated fill pixel nearest the identified runoff limit
				pr_area = pr_area[::-1]
				# Look for first occurrence of designated fill
				pr_area = (np.where(pr_area == -1, 1, 0)!= 0).argmax()
				# Calculate index of this pixel in un-reversed pixel coordinates
				pr_area = col - pr_area - 1
				# Calculate pixel distance away from identified runoff limit
				px_to_pr = np.abs(col - pr_area)
				# Value saved to data frame and can be used by user for culling purposes later on.
			else:
				px_to_pr = np.nan


			if col > b:
				# Nothing detected within ice sheet bounds
				store_arr[row, default_col] = -row * 1e-4
			elif (col != 0) & (col != cols): # Ensure matrix edges not detected.
				# Runoff limit is detected.
				# Save with row id so we can link other row-indexed info back to it later
				store_arr[row, col] = row
			else:
				# No runoff limit is detected.
				# Maybe there isn't one, or maybe it's cloudy.
				# Denote 'failure' with -value of row number.
				store_arr[row, default_col] = -row

			# Find how many cloudy pixels there are in this row
			# - this will help establish whether lack of detectable runoff limit 
			# is due to clouds.
			tot_clouds_row = np.sum(np.where(cld_rot[row,:] > 0, 1, 0))

			# In this case there are clouds above the 'runoff limit', so find 
			# out more about them
			if (tot_clouds_row > 0) & (col > 0):
				# Find the index of the cloudy pixel nearest to the runoff limit.
				# Array is reversed so as to search upglacier from the runoff limit.
				cld_nearest_roff = np.where(cld_rot[row,:col] > 0, 1, 0)[::-1].argmax()
				# Calculate the column index in which this nearest cloud is found.
				cld_nearest_roff = int(np.maximum(0, col - cld_nearest_roff - 1))
				# How far away is the cloud ?
				px_to_cld = col - cld_nearest_roff 
				# Calculate total number of pixels above runoff limit which are cloudy.
				cld_total_roff = np.sum(cld_rot[row,:cld_nearest_roff])

				# Check also for cloud proximity below identified runoff limit.
				# This can be important for identifying false runoff limits later
				# on which occur by a cloud mask failure such that cloud is first item detected
				cld_nearest_roff_down = np.where(cld_rot[row,col:] > 0, 1, 0).argmax()
				cld_nearest_roff_down = int(np.maximum(0, col + cld_nearest_roff_down))
				px_to_cld_down = cld_nearest_roff_down - col
			else:
				cld_nearest_roff = np.nan	
				cld_total_roff = np.nan
				px_to_cld = np.nan
				px_to_cld_down = np.nan
			
			store[row] = dict(tot_cldpx_above_roff=cld_total_roff,
				tot_clouds_row=tot_clouds_row, nearest_cld=cld_nearest_roff, 
				px_to_cld_up=px_to_cld, px_to_cld_down=px_to_cld_down, px_to_pr=px_to_pr)

			row += 1

		# 'Un-rotate' the located runoff limit matrix.
		locs_orig = rotate(store_arr, 
			rot_by*-1, 
			order=0,
			reshape=True)

		# Crop back to original pre-rotated extent
		r, c = np.squeeze(eopatch[labels_type][labels_name]).shape
		rlim = int((locs_orig.shape[0] - r) / 2)
		clim = int((locs_orig.shape[1] - c) / 2)
		locs_crop = locs_orig[rlim:r+rlim, clim:c+clim]

		# Rotation can create non-zero artefacts. Get rid of them.
		# However, this shouldn't be a problem with rotation order=0.
		locs = np.where((locs_crop < 1) & (locs_crop > -1), 0, locs_crop).astype(np.int)
		
		# Save cloud info
		cloud_info = pd.DataFrame.from_dict(store, orient='index')

		# Row, column coordinates of runoff limit in 'north-up' projection
		# Returns indices of the elements that are non-zero
		locs_r, locs_c = np.nonzero(locs)
		
		# Retrieve the value at each runoff limit location
		# This correponds to the row number in the rotated image
		# And is therefore used to match up with corresponding cloud info
		vals = locs[locs_r,locs_c]
		
		# Transform north-up row,col coordinates to projected X,Y.
		# Derived from eolearn.io.local_io.ExportToTiff
		x,y = rasterio.transform.xy(transform, locs_r, locs_c)

		# Transform all data to data frame.
		df = pd.DataFrame({'x':x, 'y':y, 'row_northup':locs_r, 
			'col_northup':locs_c, 'row_rot':vals})
		# Join runoff limit with cloud info using 'rotated' row coordinate.
		df_join = pd.merge(df, cloud_info, left_on='row_rot', 
			right_index=True, how='left')
		# Save designation of whether given x/y coordinate is for 
		# a runoff limit (True) or just a failure, e.g. due to clouds (False)
		df_join['rlim'] = np.where(df_join.row_rot > 0, True, False)

		# Save located runoff limit raster.
		new_feat_type, new_feat_name = next(self.feature_out())
		locs = np.reshape(locs, (1, locs.shape[0], locs.shape[1], 1))
		eopatch.add_feature(new_feat_type, new_feat_name, locs)

		# Save runoff limit dataframe.
		eopatch.meta_info[new_feat_name] = df_join

		if __debug:
			eopatch.meta_info['rot'] = im_rot
			eopatch.meta_info['rot_by'] = rot_by
			eopatch.meta_info['store'] = store_arr
			eopatch.meta_info['locs'] = locs
			eopatch.meta_info['locs_orig'] = locs_orig
			eopatch.meta_info['locs_crop'] = locs_crop
		
		return eopatch



class MedianAspectTask(EOTask):

	def __init__(self, feature_in, meta_key_out):
		self.feature_in = self._parse_features(feature_in)
		self.meta_key_out = meta_key_out
		return

	def execute(self, eopatch):
		feature_type, feature_name = next(self.feature_in())
		asp = np.squeeze(eopatch[feature_type][feature_name])
		asp = np.nanmedian(np.where(asp > 0, asp, np.nan))
		eopatch.meta_info[self.meta_key_out] = asp
		return eopatch



class ExtractElevationsTask(EOTask):
	""" 
	Extract elevation at each x,y location of the DataFrame located at 
	eopatch.meta_info[meta_key]. Requires the DataFrame to have x and y columns.
	A new column 'elev' is added in-place.
	"""

	def __init__(self, meta_key):
		self.meta_key = meta_key

	def execute(self, eopatch):

		items = tuple(zip(
			eopatch.meta_info[self.meta_key].x,
			eopatch.meta_info[self.meta_key].y))

		with rasterio.open(config.get('postprocess', 'dem_in')) as dem:
			elevs = []
			for val in dem.sample(items):
				elevs.append(val[0])

		elevs = pd.Series(elevs, name='elev', 
			index=eopatch.meta_info[self.meta_key].index)
		eopatch.meta_info[self.meta_key] = pd.concat(
			(eopatch.meta_info[self.meta_key], elevs),
			axis='columns')

		return eopatch



class LoadMaskedLabelsTask(EOTask):

	def __init__(self, feature_out, shapes, work_path=None):
		self.feature_out = self._parse_features(feature_out)
		self.shapes = shapes
		self.work_path = work_path

	def execute(self, eopatch, filename=None):

		if self.work_path is not None:
			filename = os.path.join(self.work_path, filename)

		with rasterio.open(filename) as src:
			out_image, out_transform = rasterio.mask.mask(src, self.shapes, invert=True)
			out_meta = src.meta

		feature_type, feature_name = next(self.feature_out())
		eopatch.add_feature(feature_type, feature_name, 
			np.reshape(out_image, (1,out_image.shape[1],out_image.shape[2],1)))

		return eopatch



################################################################################
# Set up tasks


# Mask loading - do this first so that we get a proper EOPatch back (we don't from LoadMaskedLabelsTask)
task_load_clouds = ImportFromTiff((FeatureType.MASK, 'clouds'), work_path)

task_load_desfill = ImportFromTiff((FeatureType.MASK, 'des_fill_msk'), work_path)


# Main loading
# Decide whether to use master map mask
mmm = config.get('postprocess', 'master_map_mask', fallback=None)
if mmm is not None:
	print('Running in master_map_mask mode.')
	with fiona.open(mmm, "r") as shapefile:
		shapes = [feature["geometry"] for feature in shapefile]	
		# remove NoneTypes which may have snuck in 
		shapes = [i for i in shapes if i]

	task_load_labels = LoadMaskedLabelsTask((FeatureType.DATA, 'labels'), 
		shapes, work_path)

	task_load_skel_labels = LoadMaskedLabelsTask((FeatureType.DATA, 'skel_labels'), 
		shapes, work_path)

else:
	print('No master_map_mask path found, continuing without masking')
	task_load_skel_labels = ImportFromTiff((FeatureType.DATA, 'skel_labels'), work_path)
	task_load_labels = ImportFromTiff((FeatureType.DATA, 'labels'), work_path)

task_load_labels_df = io_dataframes.ImportDFTask(work_path, 'regions')
task_load_skel_labels_df = io_dataframes.ImportDFTask(work_path, 'skel_regions')


# Tasks for identifying slope aspect
task_load_icemsk = ImportFromTiff((FeatureType.MASK, 'icemsk'), 
	config.get('postprocess', 'mask_in'))

task_load_aspect = ImportFromTiff((FeatureType.DATA, 'aspect'),
	config.get('postprocess', 'aspect_in'))

task_mask_aspect = MaskFeature(
	(FeatureType.DATA, 'aspect'),
	(FeatureType.MASK, 'icemsk'), [0,],
	no_data_value=0)

task_calc_aspect = MedianAspectTask((FeatureType.DATA, 'aspect'), 'medasp')


# Filtering and scanning
task_filter_features = FilterFeaturesTask((FeatureType.DATA, 'labels'),
	(FeatureType.DATA, 'labels_filt'),
	FILTER_DESIGN_FILE)

# Set areas of labelled image that are outside the designated fill
task_mask_desfill = MaskFeature(
	(FeatureType.DATA, 'labels_filt', 'labels_filt_mskd'),
	(FeatureType.MASK, 'des_fill_msk'), [1,-1],
	no_data_value=-1)

task_scan = ScanRunoffLimitTask((FeatureType.DATA, 'labels_filt_mskd'), 
	(FeatureType.MASK, 'clouds'),
	(FeatureType.MASK, 'icemsk'),
	(FeatureType.DATA, 'RLIM'),
	aspect_key='medasp')

task_elev = ExtractElevationsTask('RLIM')

task_export_rlim = io_dataframes.ExportDFTask(
	meta_feature_name='RLIM',
	save_path=work_path)

task_export_filtered_labels = ExportToTiff((FeatureType.DATA, 'labels_filt_mskd'),
	folder=work_path, compress='lzw')



workflow = LinearWorkflow(
	task_load_clouds, task_load_desfill,
	task_load_labels, task_load_labels_df, 
	task_load_skel_labels, task_load_skel_labels_df, 
	task_load_icemsk, task_load_aspect, task_mask_aspect, task_calc_aspect,
	task_filter_features, 
	task_mask_desfill,
	task_scan, task_elev,
	task_export_rlim,
	task_export_filtered_labels)


################################################################################
# Query data and run pipeline.

products_df = run_landsat_query.execute_query(config, 
	run_landsat_query.create_query(config))


# Set up input arguments	
workflow_args = []
execution_names = []
for ix, product in products_df.iterrows():

	patch_args = {}

	# Path to input image
	image_path_labels = product.PRODUCT_ID + '_EPSG%s_LABELS.tif' %(epsg)
	image_path_labels = os.path.join(product.PRODUCT_ID, image_path_labels)
	patch_args[task_load_labels] = {'filename':image_path_labels}

	# Path to input df
	df_path_labels = product.PRODUCT_ID + '_EPSG%s_LABELS.csv' %(epsg)
	df_path_labels = os.path.join(product.PRODUCT_ID, df_path_labels)
	patch_args[task_load_labels_df] = {'filename':df_path_labels}

	# Path to input skel image
	image_path_labels = product.PRODUCT_ID + '_EPSG%s_SKEL_LABELS.tif' %(epsg)
	image_path_labels = os.path.join(product.PRODUCT_ID, image_path_labels)
	patch_args[task_load_skel_labels] = {'filename':image_path_labels}

	# Path to input skel df
	df_path_labels = product.PRODUCT_ID + '_EPSG%s_SKEL_LABELS.csv' %(epsg)
	df_path_labels = os.path.join(product.PRODUCT_ID, df_path_labels)
	patch_args[task_load_skel_labels_df] = {'filename':df_path_labels}

	# Path to clouds
	image_path_clouds = product.PRODUCT_ID + '_mask_cloud_EPSG%s.tif' %(epsg)
	image_path_clouds = os.path.join(product.PRODUCT_ID, image_path_clouds)
	patch_args[task_load_clouds] = {'filename':image_path_clouds}

	# Path to des. fill
	image_path_desfill = product.PRODUCT_ID + '_qa_fill_EPSG%s.tif' %(epsg)
	image_path_desfill = os.path.join(product.PRODUCT_ID, image_path_desfill)
	patch_args[task_load_desfill] = {'filename':image_path_desfill}

	# Export dataframe path
	df_path_rlim = product.PRODUCT_ID + '_RLIM.csv'
	df_path_rlim = os.path.join(product.PRODUCT_ID, df_path_rlim)
	patch_args[task_export_rlim] = {'filename':df_path_rlim}

	# Export image containing only filtered labels.
	image_path_export_labels = product.PRODUCT_ID + '_LABELS_FILT.tif'
	image_path_export_labels = os.path.join(product.PRODUCT_ID, image_path_export_labels)
	patch_args[task_export_filtered_labels] = {'filename':image_path_export_labels}	

	workflow_args.append(patch_args)
	execution_names.append(product.PRODUCT_ID)


################################################################################

def run_eoexecute(this_workflow_args, this_workflow_names):

	executor = EOExecutor(workflow, this_workflow_args, 
		save_logs=True, logs_folder=job_path)

	try:
		nworkers = int(os.environ['EOLEARN_WORKERS'])
	except KeyError:
		print('WARNING: EOLEARN_WORKERS env-var not set, defaulting to one worker.')
		nworkers = 1

	results = executor.run(workers=nworkers, multiprocess=True)
	executor.make_report()
	report_fn = executor.get_report_filename()
	print('Report saved to:\n' + report_fn)

	failures = executor.get_failed_executions()
	failed_products = []
	if (len(failures) > 0):
			for fail_id in failures:
				pid = this_workflow_names[fail_id]
				failed_products.append(pid)

	return (results, report_fn, failed_products)


################################################################################
# Set up execution strategy
tar = False


if args.test_first_product:
	print('TEST MODE')
	print(('Product ID: {id}').format(id=execution_names[0]))
	result = workflow.execute(workflow_args[0], execution_names[0])
	first_rl = list(result.values())[0]


elif args.chunk:

	chunks_cumsum = 0

	tot_chunks = int(np.ceil(len(workflow_args) / int(args.chunk)))
	eoreports = []

	for chunk_no in range(1, tot_chunks+1):

		print(('{t} | Executing chunk {n}/{total}').format(t=dt.datetime.now(), 
			n=chunk_no, total=tot_chunks))
		end_ix = chunks_cumsum+int(args.chunk)
		if end_ix > len(workflow_args):
			end_ix = -1

		chunk_workflow_args = workflow_args[chunks_cumsum:end_ix]
		chunk_execution_names = execution_names[chunks_cumsum:end_ix]

		if not os.path.exists(work_path):
			os.makedirs(work_path)

		#### Zip up on main storage
		# We want to preserve relative paths, so change to directory
		os.chdir(job_path)
		if tar:
			chunk_fn = 'chunk.tar'
			chunk_fn_array = os.path.join(job_path, chunk_fn)
			cmd = 'tar -cf {fn} '.format(fn=chunk_fn)
		else:
			# Could consider converting this to rsync with --include-files --exclude-files
			cmd = 'cp -r '
		for item in chunk_execution_names:
			cmd += ('{i} ').format(i=item)
		if not tar:
			cmd += (' {dest}/').format(dest=work_path)
		subprocess.check_output(cmd, shell=True)

		# Move to local scratch
		if tar:
			cmd = ('mv {fn_a} {dest}').format(fn_a=chunk_fn_array, dest=work_path)
			subprocess.check_output(cmd, shell=True)

			# Extract
			chunk_fn_scratch = os.path.join(work_path, chunk_fn)
			cmd = ('tar -xf {fn}').format(fn=chunk_fn_scratch)
			print(cmd)
			subprocess.check_output(cmd, shell=True)


		### Execute pipeline
		print(('\t{t} | Copy to scratch complete. Starting pipeline.').format(t=dt.datetime.now()))
		results, report, fails = run_eoexecute(chunk_workflow_args, chunk_execution_names)
		eoreports.append(report)


		#### Return results to main storage
		# Zip up on local
		print(('\t{t} | Pipeline finished. Starting copy.').format(t=dt.datetime.now()))
		os.chdir(work_path)
		if tar:
			chunk_res_fn = 'chunk_results.tar'
			chunk_res_path = os.path.join(work_path, chunk_res_fn)
			cmd = ('tar -cf {f} ').format(f=chunk_res_fn)
			for item in chunk_execution_names:
				fn = os.path.join(item, item + '_RLIM.csv')
				cmd += fn + ' ' 
			# Move to destination
			cmd = ('mv {fn} {dest}/').format(fn=chunk_res_path, dest=job_path)
			subprocess.check_output(cmd, shell=True)

			# Unzip
			os.chdir(job_path)
			cmd = ('tar -xf {f}').format(f=chunk_res_fn)
			subprocess.check_output(cmd, shell=True)
			os.remove(chunk_res_fn)
		else:
			cmd = 'cp '
			for item in chunk_execution_names:
				if item not in fails:
					dest = os.path.join(job_path, item, '')
					cmd_item = ('{c} {i}/{i}_RLIM.csv {d}').format(c=cmd, i=item, d=dest)
					subprocess.check_output(cmd_item, shell=True)		

		# Clear up scratch
		# System call avoids needing to import shutil.rmtree
		cmd = ('rm -rf {p}').format(p=work_path)
		subprocess.check_output(cmd, shell=True)

		print(('\t{t} | Copying and clear up complete.').format(t=dt.datetime.now()))

		# Increment chunking pointer
		chunks_cumsum = end_ix

	print('EOExecution reports associated with this job:')
	print(*eoreports, sep='\n')


# Run all products without chunking
else:
	results, report, fails = run_eoexecute(workflow_args, execution_names)

################################################################################

print(('Finished at {d}').format(d=dt.datetime.now()))
