"""

Post-Process-Job script: Reproject GeoTiff outputs of river detection workflow.

Andrew Tedstone, March 2020.

"""

import argparse
import os
import subprocess
import configparser
import json
import sys
from multiprocessing import Pool

import georaster

import run_landsat_query


parser = argparse.ArgumentParser(description='Reproject outputs of workflow')
parser.add_argument('job_id', type=str, 
	help='str, job ID, as stored in $PROCESS_DIR')
parser.add_argument('-outputs', dest='warp_outputs', default=False, 
	action='store_true',
	help='Reproject outputs')
parser.add_argument('-masks', dest='warp_masks', default=False, 
	action='store_true',
	help='Reproject masks')
parser.add_argument('-bands', dest='warp_bands', default=False, 
	action='store_true',
	help='Also reproject all bands specified as workflow inputs (by main->bands)')
parser.add_argument('-test', dest='test_first_product', default=False, 
	action='store_true', 
	help='Run for first product in query only.')
args = parser.parse_args()

config = configparser.ConfigParser()

# Identify job path
job_path = os.path.join(os.environ['OUTPUT_DIR'], args.job_id)

# Parse job file
job_file = os.path.join(job_path, args.job_id)
config.read_file(open(job_file, 'r'))


################################################################################


def warp(res, epsg, fn_old, fn_new, compress=True, cut_wrs='', cleanup=False,
	overwrite=False):

	if not overwrite:
		if os.path.exists(fn_new):
			return False

	if compress:
		compress_opt = '-co "COMPRESS=LZW"'
	else:
		compress_opt = ''
	# the -tap option is essential, IN CONCERT with equivalent in ppj_create_dems.py, to get matching extents.
	cmd = ('gdalwarp {a} -tap -overwrite -of GTiff -tr {b} {b} '
		'-t_srs EPSG:{c} {cut} {d} {e}').format(
			a=compress_opt, b=res, c=epsg, d=fn_old, e=fn_new,
			cut=cut_wrs)
	try:
		subprocess.check_output(cmd, shell=True)
	except subprocess.CalledProcessError:
		print('FAILURE: ' + cmd, file=sys.stderr)
		return

	if cleanup:
		cmd = ('rm {old}').format(old=fn_old)
		subprocess.check_output(cmd, shell=True)

	return


################################################################################

# Retrieve save tasks - all these will get reprojected.
save_tasks = json.loads(config.get('chain', 'save'))
mask_save_tasks = json.loads(config.get('chain', 'save_masks'))

# Retrieve bands
warp_bands = json.loads(config.get('main', 'bands'))

# Run query to find products to process.
products_df = run_landsat_query.execute_query(config, 
	run_landsat_query.create_query(config))

# Get EPSG and spatial resolution to project to
epsg = config.get('warp', 'to_epsg')
res = config.get('warp', 'to_xyres')

cleanup = config.getboolean('postprocess', 'cleanup')

# Set up info for printing progress
n = 1
ntot = len(products_df)

for_map = []
# Run through all products
for ix,product in products_df.iterrows():

	print('%s/%s %s' %(n, ntot, product.PRODUCT_ID))

	if config.getboolean('postprocess','cut_wrs', fallback=None) is not None:
		# Get details of polygon cutline using ogrinfo
		# Supply cutline to gdalwarp - here set cut_wrs
		print('stuff')
		cut_wrs = ''
	else:
		cut_wrs = ''

	if args.warp_outputs:
		tn = 1
		for output in save_tasks:

			fn_utm = product.PRODUCT_ID + '_S%s_%s.tif' %(tn, output)
			fnp_utm = os.path.join(job_path, product.PRODUCT_ID, fn_utm)

			fnp_new = fnp_utm[:-4] + '_EPSG%s.tif' %epsg

			if os.path.exists(fnp_new):
				tn += 1
				continue

			for_map.append((res, epsg, fnp_utm, fnp_new))

			# warp(res, epsg, fnp_utm, fnp_new, compress=True, cut_wrs=cut_wrs,
			# 	cleanup=cleanup)

			tn += 1

	if args.warp_masks:
		for output in mask_save_tasks:

			fn_utm = product.PRODUCT_ID + '_%s.tif' %(output)
			fnp_utm = os.path.join(job_path, product.PRODUCT_ID, fn_utm)

			fnp_new = fnp_utm[:-4] + '_EPSG%s.tif' %epsg

			if os.path.exists(fnp_new):
				continue

			for_map.append((res, epsg, fnp_utm, fnp_new))
			# warp(res, epsg, fnp_utm, fnp_new, compress=True, cut_wrs=cut_wrs,
			# 	cleanup=cleanup)


	if args.warp_bands:
		for band in warp_bands:
			fn_utm = product.PRODUCT_ID + '_B%s.TIF' %band
			fnp_utm = os.path.join(os.environ['L0data'], 
				product.SENSOR_ID, product.PRODUCT_ID, fn_utm)

			fnp_new = os.path.join(job_path, product.PRODUCT_ID, 
				(fn_utm[:-4] + '_EPSG%s.tif' %epsg))

			if os.path.exists(fnp_new):
				continue

			for_map.append((res, epsg, fnp_utm, fnp_new))
			# warp(res, epsg, fnp_utm, fnp_new, compress=True, cut_wrs=cut_wrs,
			# 	cleanup=False)


	if args.test_first_product:
		break

	n += 1


with Pool(processes=8) as pool:
	results = pool.starmap(warp, for_map)

print('Finished.')



# probably need some fail safes/logging in case of failed tiles. Keep an eye on this.