"""
Pre-processing steps for supraglacial river detection implemented as EO-Learn
EOTasks.

Andrew Tedstone, November 2019

"""

from eolearn import io
from eolearn.core.constants import *
from eolearn.core import EOPatch, EOTask, LinearWorkflow
from eolearn.features.bands_extraction import NormalizedDifferenceIndexTask
import numpy as np

import cv2
from skimage.filters import rank
from skimage.morphology import square
from skimage.filters import median

from skimage import img_as_float
from skimage.restoration import denoise_nl_means



class ConvertToTypeTask(EOTask):
	""" Convert to data type and apply min/max thresholds.

	Values above the max threshold are set to the minimum value.

	minval and maxval are to be provided as tuples:
	(value_for_threshold, replace_with_this_value)

	"""

	def __init__(self, band, dtype, minval, maxval):
		self.band = band
		self.dtype = dtype
		self.minval = minval
		self.maxval = maxval

	def execute(self, eopatch):
		eopatch.data[self.band] = eopatch.data[self.band].astype(self.dtype)
		eopatch.data[self.band] = np.where(eopatch.data[self.band] > self.minval[0], eopatch.data[self.band], self.minval[1])
		eopatch.data[self.band] = np.where(eopatch.data[self.band] < self.maxval[0], eopatch.data[self.band], self.maxval[1])

		return eopatch




class ApplyScaleTask(EOTask):
	""" Apply specified scaling factor to specified DATA FeatureTypes of EOPatch.

	Should this be generalised to DATA_TIMELESS?

	"""

	def __init__(self, scalers):
		""" 

		:param scalers: dict of {DATA_FEATURE_NAME:SCALING_FACTOR}.

		"""
		self.scalers = scalers


	def execute(self, eopatch):
		""" Apply scaling factors to eopatch.DATA attributes.

		"""
		for scaler in self.scalers:
			eopatch.data[scaler] = eopatch.data[scaler] * self.scalers[scaler]

		return eopatch




class PCATask(EOTask):

	def __init__(self):
		return NotImplementedError

	def execute(self):
		return NotImplementedError




class NDITask(EOTask):
	""" Normalised Difference Index task. 
	"""

	def __init__(self, feature_name, band_a, band_b):
		self.feature_name = feature_name
		self.band_a_feature_name = band_a
		self.band_b_feature_name = band_b

	def execute(self, eopatch):
		ndi = (eopatch.data[self.band_a_feature_name] - eopatch.data[self.band_b_feature_name]) / \
		      (eopatch.data[self.band_a_feature_name] + eopatch.data[self.band_b_feature_name])
		
		eopatch.add_feature(FeatureType.DATA, self.feature_name, ndi)
		return eopatch




class MeanFilterTask(EOTask):

	def __init__(self, neighborhood_size, band, feature_name):
		self.neighborhood_size = neighborhood_size
		self.band = band
		self.feature_name = feature_name

	def execute(self, eopatch):
		
		#scale_by = 1e2
		# rank.mean requires 8-bit or 16-bit unsigned integers
		to_filter = np.squeeze(eopatch.data[self.band])
		#to_filter = (to_filter * scale_by).astype(np.uint16)
		filtered = rank.mean(to_filter, selem=square(self.neighborhood_size))
		#filtered = filtered.astype(np.float32) / scale_by
		filtered = np.reshape(filtered, (1, filtered.shape[0], filtered.shape[1], 1))
		eopatch.add_feature(FeatureType.DATA, self.feature_name, filtered)

		return eopatch



class MedianFilterTask(EOTask):

	def __init__(self, neighborhood_size, band, feature_name):
		self.neighborhood_size = neighborhood_size
		self.band = band
		self.feature_name = feature_name

	def execute(self, eopatch):
		
		
		to_filter = np.squeeze(eopatch.data[self.band])
		filtered = median(to_filter, selem=square(self.neighborhood_size))
		filtered = np.reshape(filtered, (1, filtered.shape[0], filtered.shape[1], 1))
		eopatch.add_feature(FeatureType.DATA, self.feature_name, filtered)

		return eopatch



class CLAHETask(EOTask):
	""" Contrast Limited Adaptive Histogram Equalisation """

	def __init__(self, band, clipLimit, tileGridSize):
		# https://docs.opencv.org/master/d6/dc7/group__imgproc__hist.html#gad689d2607b7b3889453804f414ab1018
		self.band = band
		self.operator = cv2.createCLAHE(clipLimit=clipLimit, tileGridSize=tileGridSize)
		self.feature_name = 'CLAHE'

	def execute(self, eopatch):

		scale_by = 1e2
		# CLAHE requires 8-bit or 16-bit unsigned integers
		to_filter = np.squeeze(eopatch.data[self.band])
		to_filter = (to_filter * scale_by).astype(np.uint16)
		filtered = self.operator.apply(to_filter)
		filtered = filtered.astype(np.float32) / scale_by
		# reshape to dimensions required by FeatureType.DATA
		filtered = np.reshape(filtered, (1, filtered.shape[0], filtered.shape[1], 1))

		eopatch.add_feature(FeatureType.DATA, self.feature_name, filtered)

		return eopatch
		



class OBNLMTask(EOTask):

	def __init__(self, band, feature_name, 
		search_window, similarity_window, filtering_degree,
		remove_previous=False):

		self.band = band			
		self.feature_name = feature_name
		self.search_window = search_window
		self.similarity_window = similarity_window
		self.filtering_degree = filtering_degree
		self.remove_previous = remove_previous


	def execute(self, eopatch):		

		to_filter = img_as_float(np.squeeze(eopatch.data[self.band]))

		filtered = denoise_nl_means(to_filter, 
			patch_size=self.search_window,
			patch_distance=self.similarity_window,
			h=self.filtering_degree)
		filtered = np.reshape(filtered, (1, filtered.shape[0], filtered.shape[1], 1))

		eopatch.add_feature(FeatureType.DATA, self.feature_name, filtered)

		if self.remove_previous:
			eopatch.remove_feature(FeatureType.DATA, self.band)

		return eopatch




class ShadeCorrectTask(EOTask):

	def __init__(self, kernel, band, feature_name):
		self.kernel = kernel
		self.feature_name = feature_name
		self.band = band

	def execute(self, eopatch):
		meanf = MeanFilterTask(self.kernel, self.band, 'BG')
		eopatch_bg = meanf.execute(eopatch)
		detr = eopatch.data[self.band] - eopatch_bg.data['BG']
		eopatch.add_feature(FeatureType.DATA, self.feature_name, detr)
		return eopatch




class InvertTask(EOTask):

	def __init__(self):
		return

	def execute(self):
		return
