#!/usr/bin/env python
"""
Create a qsub shell script to execute post-processing scripts of rivers workflow
on beo05.unifr.ch.

Hard coded paths to my own directory - would need changing for other users, 
being lazy as this isn't really anticipated.

See also Grid Engine resource parameters in top parameter block - modify as 
necessary.

Andrew Tedstone, March 2020

"""

import argparse
import datetime as dt
import os
import configparser
import json

################################################################################


RESOURCES = {
	 'warp': [16,12],
	 'hydro_features':[16,12],
	 'runoff_limit':[16, 12],
	 'filter_rlims':[16,14] # doesn't fully work for this - as needs the merge template too
	 }

q = "big.q" #big.q
M = "andrew.tedstone@unifr.ch"

################################################################################

parser = argparse.ArgumentParser(description='Create qsub file to run post-processing chain')
parser.add_argument('process', type=str,
	help='PPJ process to execute. One of "warp", hydro_features" or "runoff_limit".')
parser.add_argument('job_id', type=str, 
	help='str, job ID')
parser.add_argument('-mt', dest='mt', type=str, default=None,
	help='str, absolute path of culling/merging template for the quadrant - required for filter_rlims')
args = parser.parse_args()

if (args.process == 'filter_rlims' ) and (args.mt is None):
	raise ValueError('No -mt provided.')

if args.mt is None:
	args.mt = ''

# folder of job
outputs_save_path = os.path.join(os.environ['OUTPUT_DIR'], args.job_id)

# config = configparser.ConfigParser()
# config.read_file(open(os.path.join(outputs_save_path, args.job_id), 'r'))
# satellite = config.get('main','satellite')

# try:
# 	v = RESOURCES[args.process][satellite]['mem']
# except KeyError:
# 	raise KeyError('Process or satellite not known to this script.')

ppj_name = args.job_id + '.' + args.process

# Designate a qsub file name
qsub_file = os.path.join(outputs_save_path, ('qsub_{a}.sh').format(a=ppj_name))

# Specify outputs
sge_out = os.path.join(outputs_save_path, '{a}.out'.format(a=ppj_name))
sge_err = os.path.join(outputs_save_path, '{a}.err'.format(a=ppj_name))

with open(qsub_file, 'w') as fh:

	header = ('#!/bin/bash\n'
	 		  '#$ -N {name}\n'
	 		  '#$ -S /bin/sh\n'
	 		  '#$ -pe smp {pe}\n'
	 		  '#$ -q {q}\n'
	 		  '#$ -o {sge_out}\n'
	 		  '#$ -e {sge_err}\n'
	 		  '#$ -m eas\n'
	 		  '#$ -M {mail}\n\n\n'
	).format(name=ppj_name,
		pe=RESOURCES[args.process][0], 
		q=q, 
		sge_out=sge_out, sge_err=sge_err,
		mail=M)

	fh.write(header)

	script_name = ('ppj_{proc}.py').format(proc=args.process)
	body = (
			'__conda_setup="$(\'/home/tedstona/miniconda3/bin/conda\' \'shell.bash\' \'hook\' 2> /dev/null)"\n'
			'if [ $? -eq 0 ]; then\n'
			'    eval "$__conda_setup"\n'
			'else\n'
			'    if [ -f "/home/tedstona/miniconda3/etc/profile.d/conda.sh" ]; then\n'
			'        . "/home/tedstona/miniconda3/etc/profile.d/conda.sh"\n'
			'    else\n'
			'        export PATH="/home/tedstona/miniconda3/bin:$PATH"\n'
			'    fi\n'
			'fi\n'
			'unset __conda_setup\n'
			'export EOLEARN_WORKERS={eow}\n'
			'# Setup river detection\n'
			'source /home/tedstona/setup_rivers.sh\n\n'
			'ionice -c 3 python /home/tedstona/scripts/river_detection/{ppj} {job_id} {mt}'
		).format(ppj=script_name, job_id=args.job_id, eow=RESOURCES[args.process][1], mt=args.mt)

	fh.write(body)


print(('\n\nSubmission script created. Run:\n\n\tqsub {f}\n\n').format(f=qsub_file))