#!/usr/bin/env python
"""
Create a qsub shell script to execute river_detection run_workflow.py
on beo05.unifr.ch.

Hard coded paths to my own directory - would need changing for other users, 
being lazy as this isn't really anticipated.

See also Grid Engine resource parameters in top parameter block - modify as 
necessary.

Andrew Tedstone, March 2020

"""

import argparse
import datetime as dt
import os
import configparser
import json

################################################################################

q = "big.q" # has 16 cores with 128 Gb ram, should allow 11 EOLearn workers
TOT_CORES = 16
N_WORKERS = 11
M = "andrew.tedstone@unifr.ch"

# Memory needed per beo05 processor
RESOURCES = {
		'LT5': dict(mem='10G', pe=10),
		'LE7': dict(mem='10G', pe=10),
		'LC8': dict(mem='19G', pe=6)
}

################################################################################

parser = argparse.ArgumentParser(description='Create qsub file to run processing chain')
parser.add_argument('template_file', type=str, 
	help='str, absolute path of template file')
args = parser.parse_args()

config = configparser.ConfigParser()
config.read_file(open(args.template_file, 'r'))

# Generate job ID
job_id = args.template_file.split('/')[-1]
job_id = job_id.split('.template')[0] + '.' + dt.datetime.strftime(dt.datetime.now(), '%Y%m%d-%H%M%S') + '.job'

# Create folder for job
outputs_save_path = os.path.join(os.environ['OUTPUT_DIR'], job_id)

if not os.path.exists(outputs_save_path):
	os.makedirs(outputs_save_path)

# Designate a qsub file name
qsub_file = os.path.join(outputs_save_path, ('qsub_{a}.sh').format(a=job_id))

# Specify outputs
sge_out = os.path.join(outputs_save_path, '{a}.out'.format(a=job_id[:-4]))
sge_err = os.path.join(outputs_save_path, '{a}.err'.format(a=job_id[:-4]))

# Check data source --> infer memory quantity
satellite = config.get('main', 'satellite')

with open(qsub_file, 'w') as fh:

	header = ('#!/bin/bash\n'
	 		  '#$ -N {job}\n'
	 		  '#$ -S /bin/sh\n'
	 		  '#$ -pe smp {pe}\n'
	 		  '#$ -q {q}\n'
	 		  '#$ -o {sge_out}\n'
	 		  '#$ -e {sge_err}\n'
	 		  '#$ -m eas\n'
	 		  '#$ -M {mail}\n\n\n'
	).format(job=job_id,
		pe=TOT_CORES, q=q, 
		sge_out=sge_out, sge_err=sge_err,
		mail=M)

	fh.write(header)

	body = (
			'__conda_setup="$(\'/home/tedstona/miniconda3/bin/conda\' \'shell.bash\' \'hook\' 2> /dev/null)"\n'
			'if [ $? -eq 0 ]; then\n'
			'    eval "$__conda_setup"\n'
			'else\n'
			'    if [ -f "/home/tedstona/miniconda3/etc/profile.d/conda.sh" ]; then\n'
			'        . "/home/tedstona/miniconda3/etc/profile.d/conda.sh"\n'
			'    else\n'
			'        export PATH="/home/tedstona/miniconda3/bin:$PATH"\n'
			'    fi\n'
			'fi\n'
			'unset __conda_setup\n'
			'export EOLEARN_WORKERS={eow}\n'
			'# Setup river detection\n'
			'source /home/tedstona/setup_rivers.sh\n\n'
			'ionice -c 3 python /home/tedstona/scripts/river_detection/run_workflow.py -j {job_id}'
		).format(job_id=job_id, eow=N_WORKERS)

	fh.write(body)


print(('\n\nSubmission script created. Run:\n\n\tqsub {f}\n\n').format(f=qsub_file))