#!/usr/bin/env python
"""
Execute a search for Landsat imagery kept by Google Cloud Store,
based on parameters provided in .template file used to run this script.

Optionally, 
*check* that data are available on disk.
*download* all products returned by specified query from Google Cloud Storage.

This script uses the parameters set in the query and download sections of the
template file.


Andrew Tedstone, 2019/2020

"""

import argparse
import configparser
import os
import pandas as pd
import json
import subprocess

import gs_landsat


def open_config_file(config_path):
	config = configparser.ConfigParser()
	config.read_file(open(config_path, 'r'))
	
	return config



def create_query(config, #Hex(ST_AsBinary(geom)) ', AsText(geom) as geom'
	sql_basic="select DATE_ACQUIRED, SENSOR_ID, DATA_TYPE, SCENE_ID, PRODUCT_ID, BASE_URL, WRS_PATH, WRS_ROW, CLOUD_COVER, COLLECTION_NUMBER, COLLECTION_CATEGORY "):
	""" Given a template file parsed by ConfigParser, create full SpatialSQL
	query string.

	"""

	if config is type(str):
		config = open_config_file(config)

	query_args = config.items('query')

	# Construct regular query
	regular_query = ' WHERE '
	for item in query_args:
		name = item[0]
		value = item[1]
		if name[0] == 'q':
			regular_query += value
			regular_query += ' AND '
	# Remove trailing AND
	regular_query = regular_query[:-4]

	spatial = None
	spatial = config.get('query', 'spatial', fallback=None)
	if spatial is not None:
		sql_spatial = "AND (rowid IN (SELECT rowid FROM SpatialIndex WHERE ((f_table_name = 'landsat') and (f_geometry_column = 'geom') AND " 
		sql_spatial += "( " + config.get('query','spatial') + " ) )))"	
	else:
		sql_spatial = ''

	sql_full = sql_basic + ' FROM "%s" ' %config.get('query','table_name') + regular_query + sql_spatial

	return sql_full



def execute_query(config, sql_statement, remove_duplicates=True):
	"""

	Execute specified query

	Includes logic to filter out duplicated scenes where C1 and PRE available

	"""

	if type(config) is str:
		config = open_config_file(config)

	# os.environ['L0data'] + 'landsat_gs.sqlite'
	db = gs_landsat.open_database(config.get('query', 'db'))

	df_query = gs_landsat.execute_query(db, sql_statement)

	if remove_duplicates:

		# Collection 2020
		df_2020 = gs_landsat.filter_collection_2020(df_query)

		# Collection 1
		df_c1 = gs_landsat.filter_collection_1(df_query)

		# PRE
		df_PRE = gs_landsat.filter_collection_pre(df_query)

		# Combine
		df = pd.concat((df_2020, df_c1, df_PRE), axis=0)

		db.close()

		return df
	else:
		return df_query



def stage_to_L0data(df, quick=False):

	check_pth = os.path.join(os.environ['L0data'],df.iloc[0].SENSOR_ID)
	if not os.path.exists(check_pth):
		os.makedirs(check_pth)

	n = 0
	ntot = len(df)
	for ix, row in df.iterrows():
		n += 1
		print('%s/%s %s' %(n, ntot, row.PRODUCT_ID))
		cluster_path = gs_landsat.get_product_save_path(row.PRODUCT_ID, row.SENSOR_ID)
		stage_path = os.path.join(os.environ['L0data'],row.SENSOR_ID,row.PRODUCT_ID)
		
		if quick:
			if os.path.exists(stage_path):
				continue

		cmd = ('cp -r {src} {dest}').format(src=cluster_path, 
			dest=stage_path)
		subprocess.check_output(cmd, shell=True)

	return


def delete_products(df, delete_from):
	"""
	delete_from = path of containing folder to remove from
	"""

	check_pth = os.path.join(os.environ['L0data'],df.iloc[0].SENSOR_ID)
	if not os.path.exists(check_pth):
		raise IOError('Storage path corresponding to query does not exist ({p}).'.format(p=check_pth))

	n = 0
	ntot = len(df)
	for ix, row in df.iterrows():
		n += 1
		print('%s/%s %s' %(n, ntot, row.PRODUCT_ID))

		if delete_from == 'lib':
			pth = gs_landsat.get_product_save_path(row.PRODUCT_ID, row.SENSOR_ID)
		elif delete_from == 'stage':
			pth = os.path.join(os.environ['L0data'],row.SENSOR_ID,row.PRODUCT_ID)
		else:
			raise ValueError('delete_from value unknown.')
		
		cmd = ('rm -r {p}').format(p=pth)
		try:
			subprocess.check_output(cmd, shell=True)
		except subprocess.CalledProcessError:
			print('RM non-zero exit')
			continue

	return




if __name__ == '__main__':

	parser = argparse.ArgumentParser(description='Query/download Landsat products')
	parser.add_argument('template_file', type=str, help='str, absolute path of template file')
	
	parser.add_argument('-check', dest='check', default=False, action='store_true', help='If provided, run check to see which files available in $L0lib')
	
	parser.add_argument('-download', dest='download', default=False, action='store_true', help='If provided, download all data returned by query')
	parser.add_argument('-v', dest='verbose', default=False, action='store_true', help='Verbose outputs. Only has effect if -download specified.')
	
	parser.add_argument('-stage', dest='stage', default=False, action='store_true', help='If provided, stage (copy) the products returned by the query from $L0lib to $L0data')
	parser.add_argument('-quick', dest='quick', default=False, action='store_true', help='For staging. If provided, products with a directory already on $L0data are skipped (but contents are not checked).')
	
	parser.add_argument('-unstage', dest='unstage', default=False, action='store_true', help='Remove the products corresponding to this query from staging storage.')
	parser.add_argument('-remove', dest='remove', default=False, action='store_true', help='Remove the products corresponding to this query from the L0lib.')
	args = parser.parse_args()

	config = open_config_file(args.template_file)
	
	sql_full = create_query(config)
	print('Executing SQL statement . . . :')
	print(sql_full)

	df_query = execute_query(config, sql_full, remove_duplicates=True)
	
	print('\n\n\n--------------------\nProducts returned by query: ')
	print(df_query.PRODUCT_ID)

	print('Total products: %s' %(len(df_query)))

	bands = json.loads(config.get('download', 'bands'))

	if args.check:
		statuses = gs_landsat.check_products_available(df_query, bands)
		print('\n\nAvailability in $L0lib directory:')
		print(statuses)
		print(statuses.sum())

	if args.download:
		gs_landsat.download_products(df_query, bands=bands, 
			verbose=args.verbose)

	if args.stage:
		stage_to_L0data(df_query, quick=args.quick)
	elif args.unstage:
		delete_products(df_query, 'stage')

	if args.remove:
		delete_products(df_query, 'lib')



#extra options needed:
#unstage
#delete


	# sql_basic = "select DATE_ACQUIRED, SENSOR_ID, DATA_TYPE, SCENE_ID, PRODUCT_ID, BASE_URL, WRS_PATH, WRS_ROW, CLOUD_COVER, COLLECTION_NUMBER, Hex(ST_AsBinary(geom)) as geom from landsat"
	# sql_where = " where EAST_LON > WEST_LON and \
	# SPACECRAFT_ID='LANDSAT_5' and \
	# COLLECTION_NUMBER='1'"