#!/usr/bin/env python
"""
Run the river detection workflow using options specified in template file.

	$ run_workflow.py -t <filename.template>
	$ run_workflow.py -j <job_id>
	$ run_workflow.py -t <filename.template> -j <desired job_id>

run with -test flag to only run the first image returned by the query, simply
through an EOWorkflow rather than an EOExecutor.

Andrew Tedstone, UniFR, 2019/2020

"""

# System libraries
import numpy as np
import argparse
import configparser
import json
import os
import shutil
import datetime as dt
import pandas as pd
import subprocess

# EOLearn libraries
from eolearn.core import LinearWorkflow, EOExecutor, SaveToDisk, OverwritePermission
from eolearn.core.constants import FeatureType
from eolearn.core.eodata import EOPatch
from eolearn.core.eotask import CompositeTask
from eolearn.io.local_io import ExportToTiff, ImportFromTiff
from eolearn.mask import masking
from sentinelhub import BBox, CRS

# Special libraries
import geopandas as gpd

# Landsat querying
import run_landsat_query
import gs_landsat

# Custom EOLearn libraries
import landsat
import preprocessing
import morphological
import dfft
from exportandwarp import ExportAndWarpTask

# For testing in interactive name space: force reloading.
from importlib import reload
reload(landsat)
reload(preprocessing)
reload(dfft)
reload(morphological)
reload(masking)


################################################################################

parser = argparse.ArgumentParser(description='Run processing chain')
parser.add_argument('-t', dest='template_file', type=str, default=None,
	help='str, absolute path of template file')
parser.add_argument('-j', dest='job_id', default=None,
	help=('Specify a Job ID. One will be auto-generated if this is not provided.' 
	'If the Job ID already has an associated .job file then this will be used in '
	'preference to a template file.').format())
parser.add_argument('-test', dest='test_first_product', default=False, 
	action='store_true', 
	help='Run for first product in query, as workflow, provide result to namespace as first_result.')
parser.add_argument('-retry_failures', dest='retry_failures', default=False,
	action='store_true',
	help=('Try to execute the products which failed in a previous execution (i.e. ' 
	'all those products listed in the .failures file). Requires job_id. Do not use '
	'with -restart.').format())
parser.add_argument('-restart', dest='restart', default=False,
	action='store_true',
	help=('Restart the processing chain to process remaining products which were '
		' not processed previously. Requires job_id.Do not use with '
		'-retry_failures.').format())
parser.add_argument('-chunk', dest='chunk', type=int, default=None,
	help=('Control the workflow in chunks of n products. When this option is '
		'used, n products at a time are copied to local scratch for processing'
		'then the outputs are batch-copied back at the end. Proceeds for '
		'as many chunks as needed to finish the product query. Provide '
		'a value of n which will prevent local scratch from filling up.').format())
parser.add_argument('-dry', dest='dry', default=False,
	action='store_true',
	help=('Dry run: initialise entire workflow but do not run it. Only useful when run interactively!').format())

args = parser.parse_args()

if (args.template_file is None) & (args.job_id is None):
	raise RuntimeError('Error. Provide at least one of -t or -j.')

if (args.chunk is not None) & (args.test_first_product is True):
	raise RuntimeError('Error. -chunk and -test are not compatible - provide only one.')

if (args.chunk is not None) & (os.environ['WRKDIR'] == os.environ['OUTPUT_DIR']):
	raise RuntimeError('Chunking requested but $WORKDIR is the same as $OUTPUT_DIR!')


################################################################################



def add_task(task_name, task_func, inputs=None):
	global eo_tasks
	eo_tasks[task_name.lower()] = task_func



## Set up the workflow 

# Generate job ID
if args.job_id is None:
	job_id = args.template_file.split('/')[-1]
	job_id = job_id.split('.template')[0] + '.' + dt.datetime.strftime(dt.datetime.now(), '%Y%m%d-%H%M%S') + '.job'
	print('\n\nInitialising workflow with new job ID.')
else:
	job_id = args.job_id
	print('\n\nInitialising workflow with provided job ID.')

print('\n\n\t%s' %job_id)

# Create folder for job, if it doesn't already exist
job_path = os.path.join(os.environ['OUTPUT_DIR'], job_id)
if not os.path.exists(job_path):
	if args.job_id is None:
		os.makedirs(job_path)
	else:
		raise RuntimeError('Existing job specified but it has not been found in OUTPUT_DIR.')


# Chunking requires an interim save location before the outputs are moved to 
# job_path at the end of the job.
if args.chunk is not None:
	work_path = os.path.join(os.environ['WRKDIR'], job_id)
else:
	work_path = job_path
if not os.path.exists(work_path):
	os.makedirs(work_path)


# Identify the .job file (generated from .template file)
job_file_path = os.path.join(job_path, job_id)
if os.path.exists(job_file_path):
	print('\tProceeding with the existing .job file associated with this job_id')
else:
	if args.template_file is not None:
		template_file = args.template_file
	else:
		template_file = os.path.join(os.environ['TEMPLATES'], job_id[:-20] + '.template')
	# Copy the template file to the jobs folder.
	shutil.copyfile(template_file, job_file_path)
	print('\tInitialising .job file')

print('\n\n\n')

# Open job file containing workflow configuration
config = configparser.ConfigParser()
config.read_file(open(job_file_path, 'r'))


# ------------------------------------------------------------------------------
## Build workflow

# Main tasks
eo_tasks = {}
tasks = json.loads(config.get('chain', 'tasks'))

# Save tasks
save_tasks = json.loads(config.get('chain', 'save'))
mask_save_tasks = json.loads(config.get('chain', 'save_masks'))

# If true, the output of the previous step is removed by the next step, saving memory
remove_previous = config.getboolean('chain', 'remove_previous')
# If true, run gdalwarp on the exported tiffs - uses custom EOTask
warp = config.getboolean('chain', 'warp')

output_name = None
previous_output_name = None
restart_from_step_number = None
invert = True # By default, invert image going into hipass filter

for task in tasks:

	save_dtype = np.int16

	# Non-chained tasks
	if task.lower() == 'load_single_band':
		band = json.loads(config.get('main', 'bands'))
		assert len(band) == 1
		output_name = 'B%s' %band[0]
		eo_task = landsat.ImportLandsatTask(band)
		previous_output_name = 'B%s' %band[0]
	elif task.lower() == 'load_bands':
		# Here, we don't set an output_name, as further processing will be 
		# needed before entering main workflow (e.g. NDWI)
		bands = json.loads(config.get('main', 'bands'))
		if 'ndwi' in tasks:
			to_refl = True
		else:
			to_refl = False
		eo_task = landsat.ImportLandsatTask(bands, convert_to_refl=to_refl)
	elif task.lower() == 'qa_cloud':
		# Deprecated - use task 'mask_cloud' instead
		eo_task = landsat.ImportLandsatQATask('CLOUD', clean=True)

	elif task.lower() == 'qa_cloudshadow':
		# Deprecated - use task 'mask_cloud' instead
		eo_task = landsat.ImportLandsatQATask('CLOUD_SHADOW', clean=True)

	elif task.lower() == 'qa_fill':

		eo_task1 = landsat.ImportLandsatQATask('FILL', clean=True)
		
		if 'slc_fill' in tasks:	
			eo_task2 = landsat.SLC_FillFillTask()
		else:
			eo_task2 = None

		eo_task3 = morphological.MultiDilateErodeTask(
			(FeatureType.MASK, 'FILL'),
			(FeatureType.MASK, 'FILL'), 
			50, 0, 
			remove_previous=False, #Overwrite in-place instead
			no_data_value=-1)

		if eo_task2 is not None:
			eo_task = CompositeTask(eo_task1, CompositeTask(eo_task2, eo_task3))
		else:
			eo_task = CompositeTask(eo_task1, eo_task3)


	elif task.lower() == 'slc_fill':
		eo_task = landsat.SLC_FillDataTask(json.loads(config.get('main', 'bands')))


	## SPECIAL CASE - to restart from a known last good step 
	elif task.lower() == 'restart_from':
		restart_from_step = config.get('chain', 'restart_from_step')
		restart_from_step_number = int(config.get('chain', 'restart_from_number'))
		eo_task = ImportFromTiff((FeatureType.DATA, restart_from_step.upper()),
			work_path)
		output_name = restart_from_step.upper()
	## END SPECIAL CASE


	# Chained tasks
	elif task.lower() == 'ndwi':
		output_name = task.upper()
		eo_task = preprocessing.NDITask(output_name, 
			config.get('chain', 'ndwi_blue'), config.get('chain', 'ndwi_red'))
		save_dtype = np.float32
		invert = False

	elif task.lower() == 'to_real':
		eo_task = preprocessing.ConvertToTypeTask(previous_output_name, 
			np.float16, (0, 0), (255, 255))

	elif task.lower() == 'filt_med':
		output_name = task.upper()
		neighborhood_size = int(config.get('filt_med', 'size'))
		eo_task = preprocessing.MedianFilterTask(neighborhood_size, 
			previous_output_name, output_name)

	elif task.lower() == 'hipass':
		output_name = task.upper()
		filter_design = np.array([float(config.get('hdfft','f1')), 
								  float(config.get('hdfft','f2'))])
		eo_task = dfft.TwoDFiltTask(previous_output_name, output_name, 
			'highpass', 
			filter_design,
			remove_previous=remove_previous,
			invert=invert)

	elif task.lower() == 'obnlm':
		output_name = task.upper()
		search = float(config.get('obnlm', 'm1'))
		patch = float(config.get('obnlm', 'm2'))
		eo_task = preprocessing.OBNLMTask(previous_output_name, output_name, 
			search, patch, float(config.get('obnlm', 'smooth')),
			remove_previous=remove_previous)

	elif task.lower() == 'gabor':
		output_name = task.upper()
		# Minimum width in pixel space of rivers to identify (1 or greater)
		w = (float(config.get('gabor', 'W')) - 1) / 2
		# Frequency of the modulating sinusoid
		f0 = 1. / w
		sigma = w / (2 * np.sqrt(2 * np.log(2))) 
		eo_task = morphological.GaborTask(previous_output_name, output_name,
			int(config.get('gabor', 'ndirs')), f0, sigma_x=sigma, sigma_y=sigma,
			remove_previous=remove_previous,
			mask_borders=int(config.get('gabor','mask_borders')), jobs=1) 

	elif task.lower() == 'opening':
		output_name = task.upper()
		eo_task = morphological.PathOpeningTask(previous_output_name, 
			output_name, int(config.get('pathopen','length')),
			remove_previous=remove_previous)

	elif task.lower() == 'mask_cloud':
		cloud_mask_types = json.loads(config.get('clouds', 'mask_types'))
		task_options = None
		if 'cloud_swir' in cloud_mask_types:
			task_options = dict(cloud_swir=dict(
				band=config.get('clouds','swir_band'),
				threshold=config.get('clouds','swir_threshold'),
				dilate=config.get('clouds','dilate_by'), 
				keep_band=False))

		output_name = task.upper()

		eo_task = landsat.MultiCloudMaskTask(task.upper(), 
			cloud_mask_types,
			previous_feature_name=previous_output_name,
			feature_name=output_name,
			task_options=task_options,
			remove_previous=remove_previous,
			mask_with=(FeatureType.MASK, 'FILL')) # option to apply a mask, in this case the designated fill mask

	elif task.lower() == 'mask_pathrow':
		wrs_shapefile = gpd.read_file(config.get('main', 'wrs2'))
		output_name = task.upper()
		eo_task = landsat.MaskByWRSPathRow(previous_output_name, 
			output_name, wrs_shapefile, remove_previous=remove_previous)

	elif task.lower() == 'threshold_stddev':
		output_name = task.upper()
		eo_task = morphological.ThresholdStdDevTask(previous_output_name, 
			output_name, 
			float(config.get('thresholding_stddev', 'number_std_dev')),
			remove_previous=remove_previous)

	else:
		raise ValueError('Specified task is not a known option in workflow.')

	# Append main task
	add_task(task.lower(), eo_task)

	# If there is a save task associated with this task, add it here
	for save_task in save_tasks:
		if save_task.lower() == task.lower():

			if warp:
				eo_save_task = ExportAndWarpTask((FeatureType.DATA, save_task.upper()),
					work_path, 
					config.get('warp', 'to_epsg'),
					int(config.get('warp', 'to_xyres')),
					dtype=save_dtype, no_data_value=-1, 
					compress=True, cleanup=True
					)
			else:
				eo_save_task = ExportToTiff((FeatureType.DATA, save_task.upper()), 
					folder=work_path, image_dtype=save_dtype,
					no_data_value=-1, compress='lzw')                                               # DTYPE INT16 WILL NOT ALWAYS BE APPROPRIATE
				
			add_task('save_' + save_task.lower(), eo_save_task)	


	# If this task has a mask associated with it then save that too
	for mask_save_task in mask_save_tasks:
		if mask_save_task.lower() == task.lower():
			if task.lower()[0:3] == 'qa_':
				mask_key = task.lower()[3:]
			else:
				mask_key = task.lower()

			if mask_key == 'mask_pathrow':
				ft = FeatureType.MASK_TIMELESS
			else:
				ft = FeatureType.MASK

			if warp:
				eo_save_task = ExportAndWarpTask((ft, mask_key.upper()),
					work_path, 
					config.get('warp', 'to_epsg'),
					int(config.get('warp', 'to_xyres')),
					dtype=save_dtype, no_data_value=-1, 
					compress=True, cleanup=True
					)
			else:
				eo_save_task = ExportToTiff((ft, mask_key.upper()), 
					folder=work_path, image_dtype=save_dtype,
					no_data_value=-1, compress='lzw') 
                                            # DTYPE INT16 WILL NOT ALWAYS BE APPROPRIATE
			
			add_task('save_' + mask_save_task.lower() + '_mask', eo_save_task)		

	previous_output_name = output_name
		


## Set up workflow with tasks identified above.
workflow = LinearWorkflow(*[eo_tasks[t] for t in eo_tasks])


################################################################################
# Run query to establish pipeline to run.

# Run query to find products to process.
products_df = run_landsat_query.execute_query(config, 
	run_landsat_query.create_query(config))

# If this flag is set, we only want to reprocess failed products - filter out 
# the successful ones from the query.
if args.retry_failures:
	failures = pd.read_csv(os.path.join(job_path, job_id[:-4] + '.failures'),
		header=None,
		names=['PRODUCT_ID'])
	products_df = products_df.filter(items=failures.PRODUCT_ID, axis=0)
	print('Running only the products which failed during last execution:')
	print(products_df)
	print('-'*80)

if args.restart:
	import glob
	already_processed = glob.glob(os.path.join(job_path, '*_*_*_*_*_*_*'))
	already_processed = [i.split('/')[-1] for i in already_processed]
	prev_n = len(products_df)
	products_df = products_df.drop(labels=already_processed, axis=0)
	print('\tRestart mode. Products for which there is already a processed folder will not be reprocessed.')
	print(('\tWere {prev_n} products, now processing {new_n}').format(
		prev_n=prev_n, new_n=len(products_df)))
	print('-'*80)


## In test mode when only one image is processed: option to restrict processing
# to a smaller bounding box.
# min_x, min_y, max_x, max_y
test_patch = None
if args.test_first_product:
	latlon = config.get('testing', 'bbox', fallback=None)
	if latlon is not None:
		latlon = json.loads(latlon)
		product = products_df.iloc[0]
		bands = json.loads(config.get('main', 'bands'))
		import georaster
		imref_path = gs_landsat.get_product_save_path(product.PRODUCT_ID, 
			product.SENSOR_ID)
		imref_path = os.path.join(imref_path, 
			(product.PRODUCT_ID + '_B%s.TIF') %(bands[0]))
		imref = georaster.SingleBandRaster(imref_path, load_data=False)
		epsg_target = int(imref.srs.GetAttrValue('AUTHORITY', 1))
		imref = None
		
		import pyproj
		target = pyproj.Proj('EPSG:%s' %epsg_target)
		(min_x, min_y) = target(latlon[0], latlon[1])
		(max_x, max_y) = target(latlon[2], latlon[3])
		roi = BBox(bbox=[min_x, min_y, max_x, max_y], crs=CRS(epsg_target))

		test_patch = EOPatch(bbox=roi)



## Set up arguments for each patch (i.e. Landsat product)


workflow_args = []
execution_names = []
for ix,product in products_df.iterrows():
	
	patch_args = {}

	image_path = gs_landsat.get_product_save_path(product.PRODUCT_ID, 
	product.SENSOR_ID, path=os.environ['L0data'])
	image_path = os.path.join(image_path, product.PRODUCT_ID)
	
	if (args.test_first_product) & (test_patch is not None):
		# Use the bounding box created above
		opts = {'image_path':image_path, 'eopatch':test_patch}
	else:
		# No bounding box
		opts = {'image_path':image_path}
	
	if 'load_single_band' in tasks:
		patch_args[eo_tasks['load_single_band']] = opts
	elif 'load_bands' in tasks:
		patch_args[eo_tasks['load_bands']] = opts
	else:
		if 'restart_from' not in tasks:
			raise('No input task specified.')
	
	if 'restart_from' in tasks:
		fn = product.PRODUCT_ID + '_S%s' %restart_from_step_number + '_' + restart_from_step + '.tif'
		r_opts = {'filename': os.path.join(work_path, product.PRODUCT_ID, fn)}
		patch_args[eo_tasks['restart_from']] = r_opts

	# Create filenames for outputs
	if restart_from_step_number is not None:
		s_num = restart_from_step_number + 1
	else:
		s_num = 1
	for save_task in save_tasks:
		fn = os.path.join(product.PRODUCT_ID, 
				'%s_S%s_%s.tif' %(product.PRODUCT_ID, s_num, save_task))
		patch_args[eo_tasks['save_' + save_task]] = {'filename': fn}
		s_num += 1
	for mask_save_task in mask_save_tasks:
		fn = os.path.join(product.PRODUCT_ID, 
				'%s_%s.tif' %(product.PRODUCT_ID, mask_save_task))
		patch_args[eo_tasks['save_' + mask_save_task.lower() + '_mask']] = {'filename': fn}

	# Save options for this patch to the workflow arguments
	workflow_args.append(patch_args)
	execution_names.append(product.PRODUCT_ID)


if args.dry:
	import sys
	print('-'*80)
	print('Dry run set up. Examine workflow_args, execution_names, workflow, products_df...')
	print('-'*80)
	sys.exit(0)

################################################################################

def run_eoexecute(this_workflow_args, this_workflow_names):
	executor = EOExecutor(workflow, this_workflow_args, 
		save_logs=True, logs_folder=job_path)
	
	try:
		nworkers = int(os.environ['EOLEARN_WORKERS'])
	except KeyError:
		print('WARNING: EOLEARN_WORKERS env-var not set, defaulting to one worker.')
		nworkers = 1

	results = executor.run(workers=nworkers, 
		multiprocess=True)
	executor.make_report()
	report_fn = executor.get_report_filename()
	print('Report saved to:\n' + report_fn)

	# Log any failures so that they can be reprocessed.
	failures = executor.get_failed_executions()
	failed_products = []
	if (len(failures) > 0):
		with open(os.path.join(job_path, job_id[:-4] + '.failures'), 'a') as fh:
			for fail_id in failures:
				pid = this_workflow_names[fail_id]
				failed_products.append(pid)
				fh.write(pid + '\n')

		print('Some products failed. See EOExecution report and .failures file for more info.')

	return (results, report_fn, failed_products)


################################################################################






################################################################################
# Run specified execution strategy - regular, test or chunk ...

# Delete any previous .failures file
fail_file = os.path.join(job_path, job_id[:-4] + '.failures')
if os.path.exists(fail_file):
	os.remove(fail_file)


# If test mode requested, just execute first scene returned by query
if args.test_first_product:
	result = workflow.execute(workflow_args[0])
	first_result = list(result.values())[0]

# Run in chunked mode
elif args.chunk is not None:
	chunks_cumsum = 0

	tot_chunks = int(np.ceil(len(workflow_args) / int(args.chunk)))
	eoreports = []

	L0_reqs = json.loads(config.get('download', 'bands'))

	for chunk_no in range(1, tot_chunks+1):

		print(('{t} | Executing chunk {n}/{total}').format(t=dt.datetime.now(), 
			n=chunk_no, total=tot_chunks))
		end_ix = chunks_cumsum+int(args.chunk)
		if end_ix > len(workflow_args):
			end_ix = len(workflow_args)

		chunk_workflow_args = workflow_args[chunks_cumsum:end_ix]
		chunk_execution_names = execution_names[chunks_cumsum:end_ix]

		# Make sure nothing left over from previous attempts
		if os.path.exists(work_path):
			cmd = ('rm -rf {p}').format(p=work_path)
			subprocess.check_output(cmd, shell=True)
		
		# Make work path , and --> L0data
		os.makedirs(L0data)

		# Copy data to work location
		os.chdir(job_path)	
		cmd = 'cp -r '
		for item in chunk_execution_names:
			# Create L0 directory
			dest = os.path.join(L0data, item, '')
			os.makedirs(dest)

			src = gs_landsat.get_product_save_path(item, config.get('main', 'sensor'))
			src = os.path.join(src, item)
			
			# Copy required data only (based on config::download)
			for p in L0_reqs:			
				fe = 'TIF'
				if p == 'MTL':
					fe = 'txt'
					name = 'MTL'
				elif p == 'BQA':
					name = 'BQA'
				else:
					name = 'B%s' %p

				src_f = ('{src}_{p}.{fe}').format(src=src, p=name, fe=fe)
				cmd = ('cp {src} {dest}').format(src=src_f, dest=dest)
				subprocess.check_output(cmd, shell=True)	


		# Deal with data needed for -restart mode
		if args.restart:
			# Zip up relevant materials from OUTPUT_DIR and move over
			raise NotImplementedError('Chunked with restart mode not yet implemented!')


		# Execute pipeline
		print(('\t{t} | Copy to scratch complete. Starting pipeline.').format(t=dt.datetime.now()))
		results, report, fails = run_eoexecute(chunk_workflow_args, chunk_execution_names)
		eoreports.append(report)


		# Copy output data to destination
		print(('\t{t} | Pipeline finished. Starting copy.').format(t=dt.datetime.now()))
		os.chdir(work_path)
		cmd = 'cp -r '
		for item in chunk_execution_names:
			if item not in fails:
				# We want to copy contents of entire product directory
				dest = os.path.join(job_path, '')
				cmd_item = ('{c} {i} {d}').format(c=cmd, i=item, d=dest)
				subprocess.check_output(cmd_item, shell=True)		


		# Clear up scratch
		# System call avoids needing to import shutil.rmtree
		cmd = ('rm -rf {p}').format(p=work_path)
		subprocess.check_output(cmd, shell=True)


		print(('\t{t} | Copying and clear up complete.').format(t=dt.datetime.now()))

		# Increment chunking pointer
		chunks_cumsum = end_ix

	print('EOExecution reports associated with this job:')
	print(*eoreports, sep='\n')


# Otherwise process all scenes in query without chunking
else:
	results, report, fails = run_eoexecute(workflow_args, execution_names)



print('\n\n\nJob ID %s finished.\n\n\n' %job_id)
